/*
    base and operators defining the Hilbert space
    base = spin-configurations,
    p.e. get the index from config or reverse, get symetric conf.
    - look at block_no
  
 */
#include "hilbert.h"
#include "mem.h"
#include "tbase.h"
#include "ansatz.h"
#include "array_m.h"
#include "block.h"       /* mpi_* declarations */
/* open,lseek,close using off_t (64bit using LFS) */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>  /* fsync, fdatasync, ftruncate, sleep */
#include <stdio.h>  /* fileno */

#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif


/* 2016-01-18 */
double MaxLog2Tnorm2; /* log2 of maximum storable Tnorm2, overflow check */
/* O(NN^3) NN=128 16MB ToDo! remove, only pascal[] O(n^2) is needed fast */
/* ULong  npa[NN+1][NN+1][NN+1]; */ /* (o+u+d)!/(o!u!d!) stored as 3D table */
/* 2016-01-18 better size O(N^2) than old O(N^3) */
ULong       pascal[NN+1][NN+1];  /* ToDo2016: better use Tnorm2 __int128? */
double  dbl_pascal[NN+1][NN+1];

#if 0 /* unused, slower than pascal[][] */
static  /* not called from outside */
const double fak(const int i){
  double r=1.0;int j; for(j=2;j<=i;j++)r*=j;return(r);}
#endif
/* ------------------------------------------------------
 *  calculate the binominal coefficient (use pascal triangle(?))
 *  over(n,k) = n!/((n-k)!k!) = over(n-1,k-1)*n/k = over(n-1,k)+over(n-1,r-1)
 * --------------------------------------------------------- */
/* ToDo: pascal and paini do nearly the same, use pascal or paini only! */
static inline /* not called from outside, np()=paini() + b_factor_lm() only? */
const ULong over(const int n, const int k){ /* ToDo: test overrun ULong! */
  // 2016-01-17 n >= k max nn check!
  ULong rc = pascal[n][k];
#if 1 /* ToDo2016: debug only!? */
  if (rc==0 || n>NN || k>n) {
    E2("over(%d,%d) ULong overflow, continue?",n,k);
  //  ERR(9999);
  }
  /* check 2016-01 JJ N=10 2s=8 (80 33)=3e19 (68 31)=2.2e19 */
#endif
  return rc; /* only 0<=k<=n<=NN */    
}


#if   CM(1,0,0)  
/* Perm. mit Wdhlg (3 Elemente)=(o+u+d)!/(o!u!d!)=over(n-u,d)*over(n,u) */
//const ULong np(int o,int u,int d){ ULong r1=over(o+u+d,u), r2=over(o+d,d);
const ULong np(int n,int u,int d){ ULong r1=over(n,u), r2=1;
/* ToDo:                           better    over(n,u)*over(n-u,d) */
#elif CM(0,1,0)  
//const ULong np(int o,int u,int d){ ULong r1=over(o+u+d,u), r2=over(o+d,d);
const ULong np(int n,int u,int d){ ULong r1=over(n,u), r2=over(n-u,d);
/* ToDo:                           better    over(n,u)*over(n-u,d) */
#else	/* Produkt 2er Kombinat. (2 Elemente)=(n over u)(n over d) */
const ULong np(int n,int u,int d){ ULong r1=over(n,u), r2=over(n,d);
#endif
 /* 2016-01-18 check for overflow, ToDo: call dbl_over() */
 if (r1*(double)r2 >= (double)((~(ULong)3)>>1) || r1*r2==0)
   printf("\nERROR: np(%d,%d,%d) OVL %.1e\n",n,u,d,r1*(double)r2);
 return r1*r2;
}

/* until 2^1024 = 1e308 but only 52 bit mantisse, ToDo2016: long double? */
/* stackdepth NN*32B? ulimit -s ... 32kB until N=1000 ?
   recursion is too slow, use array too */
static const double dbl_over(const int n, const int u){ /* short? */
  double rc;
  if (n<0 || n>NN || u>n || u<0) return 0.0; /* over(0 0) must give 1 */
  // if (n<=1 || u==0 || u==n)      return 1.0;
  // rc = pascal[n][u]; if (rc) return ((double)rc);
  rc = dbl_pascal[n][u];
  if (rc != pascal[n][u])
    printf("\nDBG over(%2d %2d)= %.2e != %.2e",n,u,1.*pascal[n][u],rc);
  return (rc);
//  return (dbl_over(n-1,u)
//         +dbl_over(n-1,u-1));
}

/* 2016-01-19 works until 1e308 */
const double slow_dbl_np(const int n,const int u,const int d){
 double r1=dbl_over(n,u),
        r2=CM(1,dbl_over(n-u,d),dbl_over(n,d));
 return r1*r2;
}


/* ToDo: not needed as fast version!
 *   old version stored over(n,u)*over(n,d) or over(o+u+d,u)*over(o+d,d)
 *   new stores over(n,u) less space O(n^2) instead O(n^3)=16MB for N=128!
 */
void   paini(){
  int i,j, err, gerr=0;
  /* 2016-01-18 add MaxLog2Tnorm2 */
/* -- output in spins.c already --
  if (!mpi_i) printf(" Tnorm2=%3s%d",((((Tnorm2)1)/2==0)?"int":"flt"),
                                     (int)sizeof(Tnorm2)*8);
*/    
  for(i=0; i<=NN; i++) for(j=0; j<= 0; j++) pascal[i][j]=(ULong)1;
  for(i=0; i<=NN; i++) for(j=1; j<=NN; j++) pascal[i][j]=0;
  for(i=1; i<=NN; i++) for(j=1; j<=i;  j++)
    pascal[i][j]=pascal[i-1][j]+pascal[i-1][j-1];	/* ToDo: OVL ? */
  for(i=0; i<=NN; i++) for(j=0; j<= 0; j++) dbl_pascal[i][j]=1.0;
  for(i=0; i<=NN; i++) for(j=1; j<=NN; j++) dbl_pascal[i][j]=0.0;
  for(i=1; i<=NN; i++) for(j=1; j<=i;  j++)
    dbl_pascal[i][j]=dbl_pascal[i-1][j]+dbl_pascal[i-1][j-1];
  for(      i=0; i<=NN; i++)
  for(err=0,j=0; j<=i ; j++) {  /* test OVL 2016-01-18 */
    // if (i<=6) printf("DBG.paini %d %d = %.0f\n",i,j,1.*pascal[i][j]);
    /* float64 hast 52 bits, so float64 != uint64 triggers at 4e15 or 2^52 */
    if (i>0)
    if ((pascal[i-1][j]>>1)+(pascal[i-1][j-1]>>1) >= ((~(ULong)3)>>1)
        /* test above may fail on left side! */
        || (i>j && pascal[i-1][j]==0) || (j>0 && pascal[i-1][j-1]==0)
      ) {
      err++; if (err==1) gerr++;
      if ((!mpi_i) && err==1 && ((gerr<17 && nn>=i) || gerr<2))
        printf("\nWARNING: paini overflow (n d)=(%d %d)=n!/(d!(n-d)!)=%.1e %.1e",
               i, j, 2.*((pascal[i-1][j]>>1)+(pascal[i-1][j-1]>>1)),
                dbl_pascal[i][j]);
      if (err) pascal[i][j]=0; /* access to it should cause (div0) errors */
    }
    // if (i<=6) printf("DBG.paini %d %d = %.0f %0.f\n",i,j,1.*pascal[i][j],dbl_pascal[i][j]);
  }
  if (err && !mpi_i)
    printf("\nWARNING: "
    "Please ensure that nn,nu,nd do not trigger above overflows!\n");
  /* ToDo: replace ULong by Tnorm2!? ui64 (68 31), (69 21) > 2^64 = 1.8e19 */
  return;
}
    
/* number of spin configurations (nn,nu,nd), same as n1 without Sym */
static
const ULong numconf(const int nn, const int nu, const int nd){
//  return (np(nn-CM(1,1,0)*(nu+nd),nu,nd));
  ULong rc=np(nn,nu,nd);
  if (!rc) E0("numconf access to over-OVL, return 0");
  return (rc);
}
/* --------------------------------------------------------------------*
 *                    in-out-functions
 * call this function not more than 8 times within one printf() :) 
 *  v2.26:  .=empty-site, u=up, d=down, 3=up+down 
 *  v2.27:  .=empty-site, o=up, x=down, Q=up+down (looks better)
 *          think on pointers pointing to you "o" or away from you "x"
 * ToDo: for S>1/2 output for tJ,tU (s,m=-s...+s) and for JJ (m)=(-s...+s)
 * - not thread safe
 * ------------------------------------------------------------------- */
char *ccfg(const tbase ic){ /* spin configuration |543210> zB: |o.xoxo> */
  // 2016-01 S_tbase may be bigger than NN/8 (log2-based)
  static char s1[8*(S_tbase*2+3)+1], /* 8 output strings "0xHEX...HEX\0" */
              s2[4+16+1]=".oxQ0123456789abcdef"; /* spins + hex nibbles */
  char *s3; /* return pointer to outstring indexed by x */
  static int x=0; int i;
  s1[8*(S_tbase*2+3)]='\0'; /* safety first */
  x=(x+1)%8; s3=s1+x*(S_tbase*2+3);  /* NN+1 Bytes per string */
#if 0 // old site wise
  for (i=nn-1;i>=0;i--) s3[nn-1-i]=s2[ geti(ic,i) ];
  s3[nn]='\0';
#else
  /* 2016-01-08 test NN=200 nn=28 all bits for debugging */
  if ((ic>>(CM(1,2,2)*nn-1))>>1) { /* found set bits in zero area */
    s3[0]='0'; s3[1]='x';
    for (i=S_tbase*2-1;i>=0;i--)
      s3[2+S_tbase*2-1-i]= s2[(15&(int)(ic>>(4*i)))+4];
    s3[  2+S_tbase*2]='\0';
  } else {  /* 2012-06-27 relevant bits only */
    s3[0]='0'; s3[1]='x';
    for (i=(nn*CM(1,2,2)+3)/4-1;i>=0;i--)
      s3[2+(nn*CM(1,2,2)+3)/4-1-i]= s2[(15&(int)(ic>>(4*i)))+4];
    s3[  2+(nn*CM(1,2,2)+3)/4]='\0';
   }
#endif
  return(s3);
}


tbase m55=0;    /* dynamic initialisation only in C++ => initmasks */
tbase mAA=0;
tbase m0F=0;

/* initialize bitmasks m55,mAA,m0F */
int initmasks(int nn){
  int i; tbase r=0;
  for(i=0;(unsigned)i<sizeof(tbase)*8;i+=2) r|=((tbase)1)<<i;
  m55=r;
  mAA=r<<1;
  m0F=~(((~(tbase)0)<<(nn-1))<<1); /* (int32)<<32 won't work on x86 */
  /*  printf(" ini55: %s\n",ccfg(r)); */
  return 0;
}
    

/* --------------------------------------------------------------------	*
 * 		Spin-Konfigurationen ki und deren Index i		*
 * --------------------------------------------------------------------	*/
#if NN == 8*8 /* JS2012-03-06 fix 64bit overflow, but slower, x=0 never happens? */
#define MF(x)   (((x)==(NN))?~(tbase)0:~((~(tbase)  0)<<(x))) /* x=0..64 */
#else
#define MF(x)   ( ~((~(tbase)0)<<(x)))	/* << only int/long, not longlong */
#endif
#define MA(x)   (mAA & MF(x))		/* MF(x) = x * 1 bits */
#define M5(x)   (m55 & MF(x))
#define M1(x)   (  ((tbase)  1)<<(x) )  /* undefined for x>=64 */
#define M2(x)   (  ((tbase)  2)<<(x) )
#define M3(x)   (  ((tbase)  3)<<(x) )
#define ME(x)   (  ((tbase)0-2)<<(x) )
#define MC(x)   (  ((tbase)0-4)<<(x) )

// works until 64 bits, JS 2012-05-31 (faster than loop)
// count trailing zero bits using bintree search
static inline int ctz64(tbase v){ 
/* GNUC>=4 added 2013-05-23 for compatibility */
#if NN <= CM(32,16,16) && defined(__GNUCC__) && __GNUC__ >= 4 /* faster */
  return __builtin_ctz((unsigned int)v); /* undefined result for v=0 */
#else
  unsigned int c;
  if (1&(int)v) return  0;
  if (!v)       return 64;
  c=1;
#if    NN > CM(32,16,16)
  if ((0xffffffff & (int)v) == 0) { v >>= 32; c += 32; }
#endif
  if ((0xffff     & (int)v) == 0) { v >>= 16; c += 16; }
  if ((0xff       & (int)v) == 0) { v >>= 8;  c += 8;  }
  if ((0xf        & (int)v) == 0) { v >>= 4;  c += 4;  }
  if ((0x3        & (int)v) == 0) { v >>= 2;  c += 2;  }
  c -= 0x1 & (int)v;
  return c; // Warning 0 returns 31
#endif
}

#if CM(0,1,0)  /* --- tJ --- */
inline tbase next(tbase cfg){                                       /* ki++ */ 
  int i=0,d=0,u=0; 	/* Achtung: ki<<32 bei 32bit nicht definiert*/
  if(2&(int)cfg)d+=2; else if(1&(int)cfg)u+=2;
  while((i+=2)<2*nn)switch(3&(int)(cfg>>i)){
   case 0:       if(u){ u-=2;
     return((cfg & MC(i)) | M1(i) | M5(u)<<d | MA(d));} else if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 1: u+=2; if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 2: d+=2;}
  return(M5(u)<<d | MA(d));}

tbase  ll1(ULong ii){ /* getbasis from index 0..N1-1 */
  ULong i,t;tbase j;int io,iu,id;
  j=0l;io=nn-nu-nd;iu=nu;id=nd;i=ii+1l;
  for(;;){
   if(io>0){if(i<=(t=np(iu+id+io-1,iu,  id))){j=(j<<2);  io--;continue;}i-=t;}
   if(iu>0){if(i<=(t=np(iu-1+id+io,iu-1,id))){j=(j<<2)|1;iu--;continue;}i-=t;}
   if(id>0){if(i<=(t=np(iu+id-1+io,iu,id-1))){j=(j<<2)|2;id--;continue;}i-=t;}
   else break;}
  return(j);}
    
ULong ll2(tbase ii){ /* n-dimensionale Linkodierung */
/* Inp: Spinconfiguration, Out: Index des Basiszustandes */
    ULong j=0l;int io,iu,id,i,i1;
    io=nn-nu-nd;iu=nu;id=nd;for(i=nn-1;i>=0;i--){
      i1=3 & (int)(ii>>(2*i));
      if(i1>0 && io>0)j+=np(iu+id+io-1,iu,  id);
      if(i1>1 && iu>0)j+=np(iu-1+id+io,iu-1,id);
      if(i1==0)io--;
      if(i1==1)iu--;
      if(i1==2)id--;}
    return(j);}

#elif CM(1,0,0)

/* 01101 i=1 d=1 inp  */
/* 01110 i=2 d=1 !!!  */
/* 10011 i=2 d=2      */
/* 10101 i=1 d=1      */
/* 10110 i=2 d=1      */
/* 12233=123=LM
   00011 002 i=2 d=1
   00101 011
   01100 020
   10001 101
   10100 110
 */
/* last change: long long Anpassung 9.4.1998 JS */
/* MP-version Dec99, better readable */
/* improved next()-function, some easy symmetries included */
/* 2012-03-06 64bit trouble for NN=64 */
/* next_prepareLM() must be called before next() if cfg is not LM sym! */
static inline tbase next(tbase cfg){	/* JJ */ 
  int i,d;     /* i - position of first "01"-bit sequence from bit0 */
               /* d - number of 1-bits exclusive 01 pattern */
#if defined(CONFIG_S1SYM)
  // removed 2013-04-02 because algorithm does not work for higher S(ge 2)
  //    and the future will be use recursive variant (ToDo 2014)
#endif
  /* searching xxxxx0111...11100...0 */
  /*                ^ i=d+1        ToDo: __builtin_ctzll()  */
  // count number of trailing zero bits (i-d) and 1-bits (d) 
  // gcc-3.2 __builtin_ctz x86 (add fkt for vlint.h)  ctz=ffs-1
  // http://en.wikipedia.org/wiki/Find_first_set
  //  (12)ooxoxoOxxxoo(0)  i=5 d=3 5-loops (2 trailing zeros)
  //  (12)ooxoxoXOOOXX(0)   i=(d)?d:i+1...
#if NN<=64 && 1 // triang 6 0 0 9 8+46 0sym n1=1040e6 156s  46+8=3.63m
  i=ctz64(cfg); d=ctz64(~(cfg>>i)); i+=d; // count trailing zero bits
#else           // triang 6 0 0 9 8+46 0sym n1=1040e6 159s  46+8=3.63m
  i=d=0;for(;i!=nn;i++){ if(1&(int)(cfg>>i))d++;else if(d)break; }
#endif
  if (i==nn) return MF(d);  /* JS 2012-03-06 fix overrun */
  d--;                        /* one d-bit is shifted higher, d=rest */
  // clear i bits + set 1 bit + set d lowest bits
  return ( (cfg & ~MF(i)) | M1(i) | MF(d) ); /* d==0: x<<-1 ???  */
}

tbase  ll1(ULong ii){	/* getbasis from index 0..N1-1 (no-sym) */
  ULong i,t;tbase j;int iu,id;
  j=0l; iu=nu; id=nd; i=ii+1l; /* j overflow if ii>=nns && nn=64 */
  if(nn!=nu+nd)F0("nu+nd != nn");
  for(;;){
    if(iu>0){ if(i<=(t=np(iu-1+id,iu-1,id))){j=(j<<1)|0;iu--;continue;}i-=t;}
    /* else { j=(j<<id)|MF(id);break; F A S T E R ???? } */
    if(id>0){ if(i<=(t=np(iu+id-1,iu,id-1))){j=(j<<1)|1;id--;continue;}i-=t;}
    break;}
  return(j);
}
    
ULong ll2(tbase ii){	/* n-dimensional Lin-coding */
/* Inp: spin configuration, Out: index of base state */
  ULong j=0l; int iu,id,i,i1;
  iu=nu; id=nd; if(nn!=nu+nd) F0("nu+nd != NN")
  for(i=nn-1;i>=0;i--){
    i1= 1 & (int)(ii>>i);
    if (i1>0 && iu>0 && id>0) j+=np(iu-1+id,iu-1,id);
    if (i1==0) iu--;
    if (i1==1) id--;}
  return(j);
}

#elif CM(0,0,1)  /* now the coding for the Hubbard model */

inline tbase next(tbase cfg){  				/* ki++ */ 
  int i,d,u;if(1&(int)cfg)u=2;else u=0;if(2&(int)cfg)d=2;else d=0;  
  for(i=2;i<2*nn;i+=2)switch((3&(int)(cfg>>i))){      /* -1<<2=-4 */
   case 0:
     if(u){ return((cfg & MC(i)) | M1(i) | M5(u-2) | MA(d));} else 
     if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));}
     break;
   case 1:  u+=2; 
    if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));}
    break;
   case 2:
    if(u){ return((cfg & MC(i)) | M3(i) | M5(u-2) | MA(d));} d+=2; 
    break;
   case 3:  u+=2;d+=2;}
  return( M5(u) | MA(d) );}

tbase  ll1(ULong ii){	/* get configuration from index 0..N1-1 */
    tbase j;int in,iu,id;ULong i,t;i=ii;j=0l;iu=nu;id=nd;
    for(in=nn-1;in>=0;in--){
     if(in>=iu && in>=id){
         if(i<(t=np(in,iu  ,id  ))){continue;}i-=t;}
     if(iu>0 && in>=id){
         if(i<(t=np(in,iu-1,id  ))){j|=M1(2*in);iu--;continue;}i-=t;}
     if(id>0 && in>=iu){
         if(i<(t=np(in,iu  ,id-1))){j|=M2(2*in);id--;continue;}i-=t;}
     if(id>0 && iu>0)
         if(i<   np(in,iu-1,id-1) ){j|=M3(2*in);id--;iu--;continue;}
     break;}
    return(j);}
    
ULong ll2(tbase ii){                /* ki -> i */
    ULong j;int iu,id,i,i1;j=0l;iu=nu;id=nd;
    for(i=nn-1;i;i--){ if(i<iu)if(i<id)break;i1=3&(int)(ii>>(i<<1));
      if(i1>0){  if(i>=iu   && i>=id        )j+=np(i,iu  ,id);
       if(i1>1){ if(i>=iu-1 && i>=id && iu>0)j+=np(i,iu-1,id);id--;
        if(i1>2){if(i>=iu   && i>=id        )j+=np(i,iu  ,id);iu--;}}
       else iu--;}}
    return(j);}
#endif     /* tJ JJ tU */
#undef MF
#undef MA
#undef M5
#undef M1
#undef M2
#undef M3
#undef MC
#undef ME

/* suche entry in sorted tab[0..l-1], not found => 0 */
/* ToDo: it should be easier with uuu and ddd-tab if tU */
ULong findcfg(tbase entry, ULong l){
  ULong j,m;j=0l;  /* v203 tbase1=>ULong, 98: ULong=>tbase1 */
#if (CM(1,2,2)*NN-1)>8*S_long-1
  m=~(~(ULong)0u>>1);		/* set msb (highest bit) */
#else
  m=((ULong)1)<<(CM(1,2,2)*nn-1);
  if(m==0 || m==1)F0("OVL");
#endif
  while(m){
    if ((j|m)<l) { /* long unsigned == unsigned long ? */
     if (getl1(j|m)<=entry) j|=m; }
    m>>=1;
  }
  /* if(getl1(j)!=entry)j=0l;			 not found */
  return((ULong) j);			/* ULong/tbase1 ??? */
}

const char *outphase(const tphase r){  /* tphase: complex float or integer */
  static char s1[100];
#if VecType & 4
  double phase=arg(r)/M_PI; 
  if (phase<0.0) phase+=2.0; /* sometimes gcc_cplx tJ_lc phase=-0.000 (-1,-0) */
  if (fabs(phase)<1e-12) phase=+0.0; /* remove -0.000 2015-12-31 */
  /* sprintf(s1,"+(% g,% g)",Re(r),Im(r));  2015-12 bad output on fabs(r) */
  // sprintf(s1,"(1%+5.0e)*phase(%5.3f)", sqrt(Norm(r))-1, phase); // JS2012 1-r to r-1
  if (fabs(Im(r))<1e-9 && fabs(Re(r))<1e-9)
    sprintf(s1,"  0               "); // JS2016
  else if (fabs(Im(r))<1e-9 && fabs(Re(r))-1<1e-9)
    sprintf(s1," % 1.0f               ", Re(r)); // JS2016
  else
    sprintf(s1," %5.3f*phase(%4.2f)", sqrt(Norm(r)), phase); // JS2016
#else
  sprintf(s1,"%+d",r); 
#endif
  return(s1);}


/* ------------------------ symmetric base config (class Hbase) ----
 * inp: sym.k[], sym.i[] 
 * out: |rc)=sign(s)*|s|^(-1/2)*|rc],  s=<ib1|rc)
 * |] is sym-base, summe aller zu |ib1> symmetr. configs 
 * |) is normed sym-base, 
 * |> is spin-configuration (1st base) 
 * ------------------------------------------------------------- */
 
/*
 *  print symmetric base config for debugging purposes
 *  $len>0 limits the output to $len lines
 */
int b_out(struct fullsbase *b,int len){
  int i;if(len==0)len=b->num;
  printf("\n sign=%s norm2=%4g num=%4d o=%d scfg=%s",
    outphase(b->sign),(double)b->norm2,b->num,b->o,ccfg(b->scfg));
  for(i=0;i<len && i<b->num;i++) 
    printf("\n%3d %s %s",i,outphase(b->is[i]),ccfg(b->ib[i]));
  return 0;
}

#if VecType & 4		/* used in allsymcfg... */
#define ABS1(x) ((int)(sqrt(Norm(x))+.5))
#else
#define ABS1(x) abs(x)
#endif

/* --- ax|vx>+...+1|vx> => (ax+1)|vx>+...+0|vx> --- ib[num] already stored
 * replace 0-entries by (b.num-1)-entries and b.num-- (pack) ???? 
 * the same cfg-entries with different signs do interfere to 0! 
 * twice=0: the caller function knows the twice factor already
 *     we dont have to recalculate it (ToDo: check!)
 * called by allsymcfg()
 */
void b_twice(struct fullsbase *b,int twice){  
  int k,j,num/* ,t */;
  if(twice==0)
  for(num=0;num<b->num;num++) if (b->ib[num]==b->scfg) twice++;
  if(b->o!=2 && twice>1)
  for(num=0;num<b->num;num++) if (ABS1(b->is[num])) {
    for(k=num+1,j=twice;k<b->num && j>0;) if (b->ib[k]==b->ib[num]) {
      b->num--;j--;
      b->is[k]=b->is[b->num];		/* reduce table size */
      b->ib[k]=b->ib[b->num];
    } else k++; /* Jan03: "else" correct? new {} */
  }
  b->norm2=b->num;
  return;
}

/* ----------------------------------------------------------------------- * 
 * Sz-symmetry (spin inversion) u<->d,   attention: ss may point to s1 !!!
 *  ToDo: change type of tphase from complex value (k*i*2pi/l) to integer angle (k)
 *  ToDo: dont input s1, better return the sign
 *  used in xval.c wop_ud() if nu==nd obsoletem and in hilbert.c
 * ToDo-2015-12: do we need a negativ sign here for tJ + tU ???
 * ----------------------------------------------------------------------- */
static inline  // replace tphase by sig-bit (char/int)
void get_sudconfig(const tbase ib1,const tphase s1,tbase *ib2,tphase *ss){
  /* ib2(ss) can also point to ib1(s1) !!! */
  /* ToDo: remove s1, replase tphase by int {+1,-1,0=no_minsymcfg(tU=3,wud=+1)}
      see doc/sym_tU.txt */
  *ib2=ib1;*ss=s1;		/* default: 1-operator */
  if(nu!=nd || !sym.wud) return; /* ToDo: remove it, decide it outside! */
#if CM(1,0,0)
  (*ib2)=ib1 ^ m0F;
#else /* tJ + tU */
  (*ib2)=((mAA & ib1)>>1) | ((m55 & ib1)<<1); // ToDo: ((ib1>>1)&M55)|...
#if CM(0,0,1)  /* 3=(u+d) will change sign, ToDo: only if sign needed? slow */
  { tbase1 i3; // ToDo2012: remove if (bit hacks) + logcount ib1&(ib1/2)&m55
     // better rightshift (Ld_bits=Ld_tbase) ib1 until zero or 1<<i bigger? &3 
    for (i3=((tbase)3)<<(2*nn-2);i3!=TBASE0;i3>>=2) if((ib1&i3)==i3)(*ss)*=-1;
    /* only Tud(i)( |*3*>=cdi.cui|*.*> )=-( |*3*>=cui.cdi|*.*> ) */
  }
#endif /* tU only */
#endif /* tJ + tU */
  (*ss)*=sym.wud;	/* ToDo: can be done outside, return only +/-1 */
  return;
}

/* ---- get_lm_factor ---- (long64: 2^64=1.845e19)
 * N=6 s=14 s*N=84 sym=12 spin-chain, J1=J2=1
 *   u=18,d=66: ns=1e18 n1= 2888 n2max=2e16 E0=-15   2s,8s (fits to 64bit)
 *   u=20,d=64: ns=1e19 n1= 4440 n2max=1e17 E0=-62         (ERR9112 k=0)
 *   u=22,d=62: ns=9e19 n1= 6500 n2max=1e18 E0=-105  3s,19s
 *   u=24,d=60: ns=2e19 n1= 9097 n2max=9e18 E0=-144  4s,26s
 *   u=26,d=58: ns=1e19 n1=12208 n2max=2e19 E0=-179  5s,35s
 *   u=42,d=42: ns=2e19 n1=17052 n2max=2e22 E0=-309 i70=169s
 * Tnorm2 is ULong per default, but can be set to double for big-s systems
 *   see spins.h L170  try -DTnorm2=__int128 or -DTnorm2=double
 * const function: same output for same input for optimization 2016
 */
const Tnorm2 b_factor_lm(const tbase ib1){  /* ToDo: very slow! s=7/2 */
  int k, obj[NN][4]; /* 0,1,2,3 for empty,up,down,up+down sites */
  Tnorm2 n2=1, factor;     if(sym.num_lm==nn) return(1);
  /* compute factor (n1+n2+n3+n4)!/(n1!n2!n3!n4!) =
   *   = (n1+n2+n3+n4)!/n1! * (n2+n3+n4)!/n2! * (n3+n4)!/n3!
   *  where n1=number_of_empty_sites, n2=number_of_ups, ... per LM-spin */
  for(k=0;k<nn;k++){ obj[k][0]=obj[k][1]=obj[k][2]=obj[k][3]=0; } 
  /* count types of sites n1,..,n4 = .. .u d. ud */
  for(k=0;k<nn;k++){ obj[ sym.lm[k] ][ geti(ib1,k) ]++; }
if ( ((Tnorm2)1)/2==0) { /* detect integer */
  for(k=0;k<nn;k++) if (sym.lm[k]==k) { /* ToDo: is if() needed? vectorize! */
    /*                ..        .u        d.        dd    dd,d.,.u */
    /*     over(     arg1.max=NN                       ,arg2.max=NN <= arg1) */
    factor=over(obj[k][0]+obj[k][1]+obj[k][2]+obj[k][3],obj[k][3]) /* tU */
          *over(obj[k][0]+obj[k][1]+obj[k][2]          ,obj[k][2])
          *over(obj[k][0]+obj[k][1]                    ,obj[k][1]); /* tJ */
    /* ToDo: remove? insert for debug only? speedup?
        or extra test max at start
       N=12 JJ s=7/2 models/cubocta.def  -DTnorm2=double needed
       min samples for _i128 float double ToDo
     */
    if ( (/*(n2*(double)factor >= ((~(Tnorm2)65535)>>1)) ||*/ /* Tn2=dbl fail */
          (n2*(double)factor >= 1e19) || /* maxTnorm2 as double */
     factor==0)) 
            {E1("b_factor_lm OVL %.1e",n2*(double)factor);ERR(9110);}
    // if( (double)(n2*factor) != n2*(double)factor ) ERR(9110); /* Overflow? */
    n2 *= factor;
  } /* for + if {} */
} else {  /* Tnorm2 = float */
 for(k=0;k<nn;k++) if (sym.lm[k]==k) { /* ToDo: is if() needed? vectorize! */
    /*                ..        .u        d.        dd    dd,d.,.u */
    /*     over(     arg1.max=NN                       ,arg2.max=NN <= arg1) */
    factor=(Tnorm2)( /* 2016-04-08 to avoid flt-to-int-warnings */
           dbl_over(obj[k][0]+obj[k][1]+obj[k][2]+obj[k][3],obj[k][3]) /* tU */
          *dbl_over(obj[k][0]+obj[k][1]+obj[k][2]          ,obj[k][2])
          *dbl_over(obj[k][0]+obj[k][1]                    ,obj[k][1])); /* tJ */
    if (factor==0) {
//      printf("dbl_over=0 %d %d %d %d\n",obj[k][0],obj[k][1],obj[k][2],obj[k][3]);
//    printf(" %.0f\n",dbl_over(obj[k][0]+obj[k][1]+obj[k][2]+obj[k][3],obj[k][3]));
//    printf(" %.0f\n",dbl_over(obj[k][0]+obj[k][1]+obj[k][2]          ,obj[k][2]));
//    printf(" %.0f\n",dbl_over(obj[k][0]+obj[k][1]                    ,obj[k][1]));
//      E2("dbl_over(%d,%d)=0",obj[k][1]+obj[k][2],obj[k][2]);
      ERR(9110);
    }
    n2 *= factor;
  } /* for + if {} */
} /* Tnorm2 = float */
  return(n2);
}

/* ----------------------------------------------------------------------- */
#define Err17 {static int e0=1;if(e0){ E0("*_lm may be buggy, try NOS1SYM!"); e0=0;}}

/* 2015-12-31 moved from spins.h */
static inline int phase_eq(tphase p1,tphase p2){ // ToDo: replace by p_eqz(p1-p2)
#if VecType & 4 /* remove rounding errors! 1998-170 not exact!!! */
   /* 2015-12-31 tJ_lc8_s1 k=2/8 p1=i,p2=-i gcc472.fabs(cplx)=bad! ~cxx */
//   return( (fabs(p1-p2)<1e-6)? 1: 0);} /* fabs(cplx) gives bad results  */
   return( (Norm(p1-p2)<2e-12)? 1: 0);} /* Norm faster as fabs() ????? */
#else
   return( (p1==p2)? 1: 0);}
#endif

/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)              */
/* Feb04: rewritten
   ToDo: test S=5/2 N=6 16+14 nk=1.5e8 sym=4 n1=1091 33s until 2S=
       does not work for LM, tU u2d0 N=8 square
   oo=0: return full expanded base (all symmetric configs)
   oo=1: return smallest + sign + norm2 (dont extend base)
   oo=2: break, if sign==0 or smaller found (like ifsmallest=isminsymcfg)
   ToDo: better return was_smallest/got_smallest/0?
 * ToDo: remove return within loops (see vectorization)
 */
tbase allsymcfg_lm(struct fullsbase *b, const int oo){  
  tbase ib0,ib2; int j,b1,b2,i1,i2,n2; tphase s0,s1,s2;
  CM({},Err17,Err17)	/* signum ??? ToDo 2016 testing doc/sym_tU.txt */
  n2=0;	/* count same lm-scfgs */
  /* b->num must be greater than 0 (one argument or more) */
  b->sign=b->is[0]; 
  b->scfg=b->ib[0]; /* if we have n2==1 */
  if (b->num >= MaxSBase || b->num<0) F0("allsymcfg_lm.b.num out of range");
  for(j=0;(ULong)j<(ULong)b->num;j++){
    ib0=b->ib[j]; s0=s1=b->is[j];	/* check next config */
    if(s0==0) E0("Warning: sign[j]=0");
    if (ib0< b->scfg) { b->scfg=ib0; b->sign=s0; n2=1; }
    /* ib0 = original (smallest?), ib2 = new_lm_cfg */
    /* ToDo: use faster next indizes 2016 */
    for(i1=0;i1<nn-1;i1++)for(i2=i1+1;i2<nn;i2++)if(sym.lm[i1]==sym.lm[i2]) {
      b1=geti(ib0,i1); /* site 1 within LM-site */
      b2=geti(ib0,i2); /* site 2 within LM-site */
      if(b1!=b2){ /* exchange sym-sites */
        /* 2s=3: 3du -> pd3u + -pud3 + -3ud */
        ib2=puti(puti(ib0,i1,b2),i2,b1); // b1=b2; b2=geti(ib0,i2);
        s2=s1; /* new states sign init 2016-01-16 */
        /* tU,tJ fermionic sign change? call (2sites) permutation function? */
        /* tJ must give same sign as JJ without holes */
#if CM(0,1,1) /* see doc/sym_tU.txt ToDo2016 unsure */
        /* parity for odd number of sites */
        if (((b1^b2)^((b1^b2)>>1))&1) { /* examples: .xu ux3 (fermionic transfer) */
          s2 *= parity(ib0 & p_mask[i1][i2]); }  /* not coding independent */
        if (zizj(ib0,i1,i2)==0) s2 = -s2; /* ud (fermionic exchange) */
        /* two types of LM: JJ.maxS(S+1)  and  tU.maxJ(J+1) */
        /*                 0: prefers ud+du, 1: prefers ..,.3,33 */
        if (sym.type_lm) s2 = -s2;
#endif
        if(b->scfg==ib2){	/* same lm-symmetric config stored */
          if(!phase_eq(s2,b->sign)){
          s2=b->sign=0; b->num=0; b->norm2=0; return TBASE0;}
        }
        if(!(ib0<=ib2)){ b->ib[j]=ib0=ib2; b->is[j]=s0=s2;
          if(!(b->scfg<=ib0)){
            b->scfg=ib0; b->sign=s0; n2=0; if(oo==2) return ib2; }
          i1=i2=nn; j--; /* restart for this config */ break;
        }
      } /* b1!=b2 different particles */
#if CM(0,1,1)
      else {  /* b1==b2 */
        if (sym.type_lm==0) /* JJ.maxS(S+1): T(.X.)=-.X. 2016-01 */
        if(abs(zi(ib0,i1))==0){b->sign=0;b->num=0;b->norm2=0;return TBASE0;}
        if (sym.type_lm==1) /* tU.maxJ(J+1): T(uXu)=-uXu 2016-01 */
        if(abs(zi(ib0,i1))!=0){b->sign=0;b->num=0;b->norm2=0;return TBASE0;}
     } /* b1==b2 */
#endif
   } /* i1, i2 lm-loop */
   /* add factor_lm new lm-symmetric configs to table *slow!* */
   if (oo==0) { Err17; } /* not supported yet, we should not need it */
  }
  /* calculation of n2 is buggy, so we reat it here */
  for(n2=j=0;(ULong)j<(ULong)b->num;j++){
   if (b->ib[j]==b->scfg) n2++; /* how often is scfg stored within table */
  }
  if(oo<2){	/* (n1+n2+n3+n4)!/(n1!n2!n3!n4!) */
    Tnorm2 factor=b_factor_lm(b->scfg);
    b->norm2=b->num*factor/(unsigned)n2; /* ULong -> int ??? */
    if( (double)(b->num*factor/n2) != b->num*(double)factor/n2 ) ERR(9111);
  }
/* printf("\n dbg3: %s o=%d n2= %d norm2= %d bnum %d",ccfg(b->scfg),oo,(int)n2,(int)b->norm2,(int)b->num); */
  if(!b->norm2) b->sign=(tphase)0;
  return(b->scfg);
}

/* b = b + Tud(b), Tud is the inversion operator, b is the state 
 * called from allsymcfg() for nu==nd only
 * ToDo: remove return within loops (see vectorization)
 * ToDo-2015-12: do we need a negativ sign here for tJ + tU ???
 */
void allsymcfg_ud(struct fullsbase *b){  /* for Sz!=0 wasted CPU time! */
  tbase ib0,ib1,ib2;int i,j,nj; tphase s1;
  j=nj=b->num; ib0=b->scfg; ib1=b->ib[0];
  for(i=0;i<nj;i++)/* if(b->is[i]!=0) */{
    get_sudconfig(b->ib[i],b->is[i],&ib2,&s1);	/* s|ib2>=wud*T|ib1> */
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return; }
    if( (ib1==ib2)){
      if(!phase_eq(s1,b->is[0])){b->sign=0;b->num=0;return;}
      j=nj;break; /* because |symcfg> = Tud|symcfg> Jan00 */
    }
    if (j<MaxSBase /*2014-12*/) { b->ib[j]=ib2; b->is[j]=s1; }
    j++;
  }
  b->norm2=b->num=j; return;
}

/* ----------------------------------------------------------------------- * 
 * symmetry-OP   PermOp(ip)*ib1=s*ib2 , s = +1/-1  ip = permutation vector 
 * ----------------------------------------------------------------------- */
/*   example:
 P = [0 2 1 3] = (0) (1,2) (3)
 ib1= (.. du .u d.) = sites (3 2 1 0) = bits (76 54 32 10)
 m1=7=3d -
 m1=6=3u -						 
 m1=5=2d m2=3=1d i2=(.. .. d. ..) s=0       i3=(11 11 10 00)=(0000,1110)
 m1=4=2u m2=2=1u i2=(.. .. du ..) s^=i3_1=0 i3=(00 00 01 00)=(1110,1110)
 m1=3=1d -
 m1=2=1u m2=4=2u i2=(.. .u du ..) s^=i3_3=0 i3=(11 11 01 00)=(0010,1110)
 m1=1=0d m2=1=0d i2=(.. .u du d.) s^=i3_0=0 i3=(00 00 10 10)=(0010,0001)
 m1=0=0u -      
     see doc/sym_tU.txt
 ------------------------------------------------------------------------ */
/* ms!=0 multiply phase with k, ToDo: remove and simplify
 * later build in FPGA!
 * add special versions for speedtest (2i64 or 4i64 vectorized, benes)
 *  but keep it SIMPLE and fast !
 */
#if 0  /* (tbase)(1&int) is slower on AVX2-CPU 18.58ns vs. 6.93ns */
#define B_AND1(x) ((tbase)(1&(int)(x)))
#else
#define B_AND1(x) (((tbase)1)&(x))
#endif  /* B_AND version, replace by getbit() ? */
/* static inline */ /* check asm-code for inlining gcc -S -dA */
/* tphase is complex in complex mode, ToDo: always signed char? +/-1 */
/* 2015-02-24: renamed get_symconfig to symcfg_loopN (better perf overview) */
//static  /* better to find asm without static, icore2: 46.1clks vs. 46.4clks */
__attribute__ ((hot))  // 2016-01-09 add hot + const ib1
void symcfg_loopN( const tbase  ib1,             const int ip,
                   tbase *ib2, tphase *ss, const int ms){
  /* register (1% slower!) */ tbase i2, i3; int m1;
#if CM(0,1,1)
  int s; int m2;
#endif
  *ib2=ib1; *ss=+1; if(ip<0) return; /* 1-Operator (0,identity) */
#if CM(0,1,1)
  //  T01*ud = + .u d. => + .u .. => + .. .. 
  //           + .. .. => + d. .. => + d. .u
  //  T10*ud = + .u d. => - .. d. => - .. .. 
  //           + .. .. => + .. .u => - d. .u  sign=-1*-1=+1
  //  T01*.3 = + .. du => + .. d. => + .. ..
  //           + .. .. => + .. .u => - .. du  sign=-1
  // printf("\n symcfg_i= %s %d",ccfg(ib1),ip);
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
  // from:  uX.   dX.  3X.  dXu        3Xu   3Xd  see utils/fermion_sign.c
  // to:  +p.Xu +p.Xd +.X3 -uXd      +puX3 +pdX3   p=Parity(X)  old (reverse sign)
  // to:  -p.Xu -p.Xd -.X3 +uXd      -puX3 -pdX3   p=Parity(X)  ToDo
  //                  -p^2 (-p)^2
  // assumes X is staying, but here we Permite X using P(i)=i
  i2=(tbase)0; /* new config will build on this vacuum state |....> */
  i3=(tbase)0; /* each bit will show parity of lower bits in i2 */
  s=0;         /* product of parity for sign */	/* remove from NN-1 to 0 */
  /* for LM-type syms we need k=1  to get JJ k=0 comparable states */
  /*    s=1 and one k=0 others skipped works, but if use 2 k=0 it failes
   *  sym_tU.txt N=4   ToDo: not clear why this */
  /* P=1-sym should generate always sign = +1 */
#if 1  // code von 1995..2003-04 ? s95*.tgz? why start on site nn-1? sign?
  /* starts from n-1 to ensure Perm=1 does never change the sign */
  for(m1=2*nn-1;m1>=0;m1--)if(1&(int)(ib1>>m1)){ /* m1 = old index, start high */
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	 /* m2 = P(m1) = new index */
    i2 |= ((tbase)1) << m2;		 /* populate vacuum state at m2 */
#if 0  /* two ways of getting parity of lower fermions O(n) vs. O(n^2) */
    /* 36ns n=4 2u+2d store xor for higher bits, avoid extra loop 2015-- gcc492-O1 */
    if (m2>0) s^=(1&(int)(i3>>(m2-1)));  /* get xor of lower i2-bits */
    i3 ^= (~(tbase)0)<<m2;  /* change lower-bit-parity for higher bits */
#elif 1 /* above O(n) code improved */
    /* 33ns n=4 2u+2d remove if-condition for speed 2016-02 */
    s^=(1&(int)(i3>>m2));  /* get xor of lower i2-bits */
    i3 ^= ((~(tbase)0)<<m2)/*<<1*/;  /* change lower-bit-parity for higher bits */
#else
    /* 54ns n=4 2u+2d better readable, but slow, 2016-01, O(nn^2) */
    {int m3; for (m3=0;m3<m2;m3++) s^=(i2>>m3);} /* xor of lower spins placed */
#endif
  }
#else
  //  printf(" we are here at symcfg_loopN");
  for(m1=0;m1<2*nn-2;m1++)if(1&(int)(ib1>>m1)){ // -m1 +m2=+P[m1]
    int m3;
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	    /* m2 = new index */
    i2 |= ((tbase)1) << m2;		    /* move spin to site m2 */
    // for (m3=0;m3<m1-1;m3++) s^=(ib1>>m3); /* all lower spins removed */
    for (m3=0;m3<m2;m3++) s^=(i2>>m3);    /* lower spins placed */
//    s  ^= 1&(int)(i3>>m2); /* get xor of lower i2-bits */
//    i3 ^= (~(tbase)0)<<m2; /* change lower-bit-parity for higher bits */
  }
#endif
  // printf("\n symcfg_o= %s %d",ccfg(i2),s);
  // 2016-01old: (*ss)= 1-2*(s&1);  /* ToDo: return s in units of I*2Pi/2 */
  (*ss) = 1-2*(s&1);   /* 2016-01: see sym_tU.txt, ToDo: units of I*2Pi/2 */
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  i2=0l;
  /* const may be a hint, that cache-coherence is not needed, 
   *  but it is not working on MIPS4 multithread, still slower, const pointer?
   * const int * a;        // pointer to constant variable
   * int * const a = &b;   // unmovable pointer to variable
   * const int * const a = &b; // unmovable pointer to const variable  */
  const t_sites *a = sym.i[ip]; /* constant values, writable pointer */
  // simple code is for auto micro parallelization SIMD/SSE2
  //   or just to compare the advantage of hand optimization
  /* autounrolled gcc4 -funroll-loops -O3 -S -dA */
  /* worst=scalarCPU: NN*4OPs=40*4= 160 OPs/sym (+mem+loop)
   *       3depOPs+1independendOP = 120 clks 
   *  minimum_4*64bitsuperscalar:
    (4*8+4*4+4*2+4*1+2*1+1*1)=17clksDepOR + 10clk*4U*(AND+2SHFT)=30 = 47clks
    in princip AND+OR+SHIFT could be done 10*parallel on 2*256bit+2*64bit-ALUs
      = 10clksOR+4clks(AND+2SHFT)=22clks
      = 11+15=26clks (8*64bit-superscalar)
   *  minimum_NNsuperscalar: (ln2(NN)=6)*or + 3*dependOPs = 9 clks
   haswell+loop5*unroll8: 5*4depOps=20clks(8ALUs) or 40clks(4ALUs)!!!
   */
  /* proci         tuned8or    simple_code    10+30 clks/sym 
    MIPS-500MHz  112.2 clks   118.4 clks # pathcc32 no_ib1brk 2015-06 IPC=1.43
    AMD2400MHz    55.6 clks    51.2 clks # gcc481 ssse3+3dnow (==sse2) 
    iCore2-1.6Ghz 46.1 clks    45.0 clks # gcc492 ssse3 2015-06 unroll8 r64 4ALUs 
    haswell2.4GHz 36.0 clks    36.9 clks # gcc49  avx2  2015-06 unroll8 r64 2*64+2*256bitALUs 
                                         # IPC=4.4 lNv32=14.6clk,IPC=11
  */
  /* NN*b|=((a>>i)&1)<<pi; or-s depend (40clk), rest parallel 42clk*(2SHFT+ALU) */
  /* NN*b|=(a&(1<<i))<<(pi-i) 1<<i=hardcoded,stored_pi-i = 1ROT+2ALU */
  /* pipelined 43 stages 4ALUs */
#if 1 || defined(CFG_SIMPLE_CODE) || (NN % 4 !=0) /* ToDo: round up NN! */
  i3=ib1;
  for (m1=0;m1<NN;m1++) { /* faster if nd<=nu if break loop on !ib1 */
    /* unrolled and parallizable 3 step pipe, should be done by compiler */
    /* auto parallelize a|b|c|d to (a|b)|(c|d) !?, vlint class needs cast */
    i2 |= B_AND1(i3) << a[m1]; /* asm: jmp if NN>sizeof(int) */
    i3 >>= 1; /* hard coded shift may be faster than variable shift x>>m1 */
    /* only dependency is i2|= but a good compiler should take 5/8 time */
    // if (!(ib1>>m1) ) break; // this is bad for unroll + SIMD
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
#else /* tuned8 8+32: 35.0 vs 42.3ns, 2015-06 haswell no diff!, bad for big NN */
  // better (simple+flexible) using auto-unroll without extra breaks! 2015-06
  // MIPS-5kf 120clk (slower than simple loop 112clk) a<<1 == a+a, ToDo: sym.i?
  //       no rot-unit (rot pseudo code replaced by sl+sr+or)
  // core2:    42clk = fully pipelined simple code (2SHIFTs+2ALU)/clk=4IPC noSpeedUp
  // haswell:  34clk = simple-loop  (3shifts+3ALUs)/clk? pipe4-version same
  //  --- slower for NN>64 !!! ---
  tbase r1=0,r2=0,r3=0,r4=0; // temporary registers to help piping
  i3=ib1;
  for (m1=0;m1<NN;m1+=4) { /* faster if nd<=nu if break loop on !ib1 */
    /* unrolled and parallizable 3 step pipe, should be done by compiler */
    /* auto parallelize a|b|c|d to (a|b)|(c|d) or inCPU uCode reordering */
    /*   but if NN>64 it may be slowdown !!! better use VS */
    r1|=B_AND1(i3   ) << *(a  ); 
    r2|=B_AND1(i3>>1) << *(a+1); // replacy by B_AND2(ib1)>>1; ?  or ib1/2
    r3|=B_AND1(i3>>2) << *(a+2);
    r4|=B_AND1(i3>>3) << *(a+3);  // fully paralizable for 2SHIFTS+1AGU+2ALU
    a+=4; i3>>=4;
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
  i2=(r1|r2)|(r3|r4); // 2 extra clocks * 2ALU
#endif /* simple code */
#endif /* JJ */
  *ib2=i2;
  if(ms){
#if CM(1,0,0)
    (*ss)=sym.w[ip];   /* on JJ no multiply needed */
#else
    (*ss)*=sym.w[ip];   /* better add fractional angles (integer) outside */
#endif
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    if(fabs(Im(*ss))<1e-9) (*ss)=C(Re(*ss),0);
#endif
  }  
  return;
}
/* vectorized version JJ N=64  by SSE2 128bit=2i64, AVX2 by 256bit=4i64
   2015-02-05 for vectorization benchmarks
 */
// static  /* bad to find in asm, not much faster on haswell */
//#pragma GCC push_options
//#pragma GCC reset_options
//#pragma GCC optimize ("O2","unroll-loops","tree-vectorize","unsafe-loop-optimizations")
// icore2 O3: 28.70ns*1.6GHz
// icore2 O2: 19.30ns*1.6GHz 
__attribute__ ((hot))
void symcfg_loopNv( const tbase ib1[VS], const int ip,
                          tbase ib2[VS], tphase ss[VS], const int ms){
  register int m1; int i; /* unroll!? */
  /* ULong  m3[4] __attribute__ ((aligned (32))); */ // expanded bytes for better vectorization?
  tbase  i2[VS] __attribute__ ((aligned (64)));
  tbase tb1[VS] __attribute__ ((aligned (64)));
#if CM(0,1,1)
  tbase  i3[VS] __attribute__ ((aligned (64)));
  int     s[VS] __attribute__ ((aligned (64)));
  register int m2;
#endif
  for (i=0;i<VS;i++) ib2[i]=ib1[i]; 
  for (i=0;i<VS;i++) ss[i]=+1;
  if(ip<0) return; /* 1-Operator (0,identity) ToDo: do this outside */
  for (i=0;i<VS;i++) tb1[i]=ib1[i]; /* work copy */
#if CM(0,1,1)
  // printf("\n symcfg_i= %s %d",ccfg(ib1),ip);
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
  // from:  uX.   dX.  3X.  dXu        3Xu   3Xd  see utils/fermion_sign.c
  // to:  +p.Xu +p.Xd +.X3 -uXd      +puX3 +pdX3   p=Parity(X)
  for (i=0;i<VS;i++) i3[i]=i2[i]=TBASE0; /* =0l */
  for (i=0;i<VS;i++) s[i]=0;  /* sign */
  for (i=0;i<VS;i++) /* does probably does not help */
  for(m1=2*nn-1;m1>=0;m1--,tb1[i]>>=1)if(1&(int)tb1[i]){ /* 1100ns tU24.0+8 */
//for(m1=2*nn-1;m1>=0;m1--)if(1&(int)(tb1[i]>>m1)){ /* 1100ns tU24.0+8 */
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	    /* m2 = new index */
    i2[i] |= ((tbase)1) << m2;		    /* move spin to site m2 */
    if (m2>0) s[i]^=(1&(int)(i3[i]>>(m2-1)));/* number of spins jumped over */ 
    i3[i] ^= (~(tbase)0)<<m2;		/* derived fermionic sign, 64bit?? */
  }
  // ----- tJ,tU not implemented !!! --- missing ib1[1..3]
  for (i=0;i<VS;i++) ss[i]= 2*(s[i]&1)-1; /* 2016-01 ToDo: ss= units of I*2Pi/2 */
  for (i=0;i<VS;i++) ib2[i]=i2[i]; /* unrolled by gcc -funroll-loops */
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  const t_sites *a=sym.i[ip];
  // const array not scalable!?
  //  const tbase ibu[4] __attribute__ ((aligned (32))) = {1,1,1,1};
  for (i=0;i<VS;i++) i2[i]=TBASE0;
  /* 2014-05-14 micro parallelization (SIMD), NN must be even -40% */
  /* worst: MaxSym*NN*4OPs=160*40*4=25600 OPs */
  /* quanta1-2.2GHz+gcc-4.8-O2: 33.6ns/0.45ns=74clks(40*36ops/4) noAVXused */
  /*
    not vectorized: relevant stmt not supported: _90 = pretmp_118 >> m1_111;
    not vectorized: relevant stmt not supported: _92 = 1 << m1_115;
  aligned ?  
  __m128i v = _mm_load_si128(&a[i]); # a[4] __attribute__ ((aligned (16)));
  __m128i    _mm_and_si128 (__m128i,__m128i)
  __m128i    _mm_srli_epi64(__m128i,int) = SSE2.PSRLQ  = shift right quad words
  __m256i _mm256_srli_epi64(__m256i,int) = AVX2.VPSRLQ = shift right quad words
  __m256i _mm256_and_si256(__m256i,__m256i) = AVX2.VPAND
      ai64[0..1]<<=const. but no OP:ai64[0..1]<<=bi32[0..1]
      _mm_mul_epu32 + _mm_add_epi64, but no _mm_mul_epu64
      missing instructions for perm-loop!!! but enough for bfly/bn-version
   * ToDo: what is the better order of loops, i=0..3 inside or outside?
   */
  for (m1=0;m1<NN;m1++) { /* faster if nd<=nu if break loop on !ib1 */
    // could be done outside via m3[NN][4]
    /* does not vectorize SSE2 or AVX2, missing va<<vi */
    for (i=0;i<VS;i++) i2[i] |= B_AND1(tb1[i]) << a[m1];  /* <<a_i does not vectorize!? */
    for (i=0;i<VS;i++) tb1[i]>>=1; // autovectorization
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
  for (i=0;i<VS;i++) ib2[i]=i2[i];
#endif /* JJ */
  if(ms){  /* do this outside! */
#if CM(1,0,0)
    for (i=0;i<VS;i++) ss[i]=sym.w[ip];   /* on JJ no multiply needed */
#else
    for (i=0;i<VS;i++) ss[i]*=sym.w[ip]; /* ToDo: angles as integer outside */
#endif
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    for (i=0;i<VS;i++) if(fabs(Im(ss[i]))<1e-9) ss[i]=C(Re(ss[i]),0);
#endif
  }  
  return;
}
//#pragma GCC pop_options


/* ---- long=+uudd+uudd -> short=2uudd+0uudd ----
 * faster, if symconfig and sudconfig called one after another?
 * kuerzen als extra-funktion ??? redundanter code .._lm
 * better save index for smallest? (temporarly) SORTING???
 * create cache??? or create only different configs
 * n1_sym is ca. n1/sym, d.h. ib1==ib2 | twice>1 is rare -> faster?
 *  ToDo: same as minsymcfg_loopN()?
 * called by xval.c: modeloverlap stringorder (ToDo: remove?)
 *           spins.c: nhm_err1 storeh2 outbasis
 * in:  o==2:ismincfg=isminsymcfg needed?
 * out: b->scfg=ret=mincfg; 
 *      b->ib ??? (unused? remove! ToDo)
 * ToDo: remove return within loops (see vectorization)
 */
tbase allsymcfg(const tbase ib1,struct fullsbase *b){  
  tbase ib0,ib2; int j,n2,nj=sym.numall+1; tphase s1; static int err=4;
  b->ib[0]=b->scfg=ib0=ib1; b->is[0]=b->sign=1; n2=1;
  if(nj*((nu==nd)?2:1)>MaxSBase){
    E0("b-getbase: MaxSBase to small, skip it");
    if (!--err) F0("stop after 4 errors");
    return(b->scfg);
  }
  for(j=1;j<nj;){ s1=1;
    symcfg_loopN(ib1,((unsigned)j)-1,&ib2,&s1,1);
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return ib2;}
    if( (ib1==ib2)){
      if(!phase_eq(s1,1)){b->sign=0;b->num=0;return ib2;} n2++; }
    /* storing to ib[syms] may be memory intensive, want to remove that */
    b->ib[j]=ib2; b->is[j]=s1; /* ToDo: better call insert_sorted here */
    j++;
  }
  b->num=nj; b->norm2=nj*n2; if(n2>1) b_twice(b,n2); /* obsolete if insert_sorted */
  if (nu==nd) allsymcfg_ud(b);
#if defined(CONFIG_S1SYM)    /* ToDo: does not work at the moment for tU JJ? */
  /* b_out(b,b->num); */
  if(sym.num_lm!=nn) allsymcfg_lm(b,2);
  /* b_out(b,b->num); */
#endif
  return(b->scfg);
}

/* reference function for fast minsymcfg_test3, slow, only debugging
 * ATTENTION if is[smallestcfg]=0 !!! sign of ib1
 * 1=true 0=false(smaller symcfg), -1=false(destructiv sign)
 *  ud-test only usefull if 1st spin=d!!!faster since Jan00
 * ToDo: rename b_ifsmallest to isminsymcfg 
 * 2016: does not return useful norm2 (S1SYM can couse OVLs for rekursiv.ns)
 */
#if 0
int isminsymcfg_slow(const tbase ib1, struct fullsbase *b){ 
  int old_o;
  old_o=b->o;b->o=2; allsymcfg(ib1,b); b->o=old_o;
  /* printf("\ncfg=%s: ",ccfg(ib1));b_out(b,0); */
  if (!(ib1==b->scfg)) return 0;
  if (b->norm2==0 || phase_eq(b->sign,0)) return -1;
  b->sign=1; b->num=0;				/* not given here */
  return 1;
}
#endif

/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)
 *     see doc/sym_tU.txt
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 20s(43%) => 3s(9%) PM600MHz-pg
 *  return 1=ok 0=not_min_by_cfg -1=not_min_by_destructiv_sign (for recursion)
 * 2016: does not return useful norm2 (S1SYM can couse OVLs for rekursiv.ns)
 */
__attribute__ ((hot)) /* gcc-4.3 and better, hotspot function, if S1SYM */
int isminsymcfg_lm(const tbase ib0){  /* max_time=O(nn) */
  int i1, i2; 
  for(i1=0;i1<nn-1;i1++) {
    i2=sym.ln[i1]; /* i2=next_inner_LM_group_site, loop needed? */
    if(i2<nn) {  /* compare member of composed spin with next member */  
      if (geti(ib0,i1)< geti(ib0,i2)) return 0;
#if CM(0,1,1) 
      if (geti(ib0,i1)==geti(ib0,i2)) {
        if (sym.type_lm==0) /* JJ.maxS(S+1): T(.X.)=-.X. 2016-01 */
          if (abs(zi(ib0,i1))==0) return -1;
        if (sym.type_lm==1) /* tU.maxJ(J+1): T(uXu)=-uXu 2016-01 */
          if (abs(zi(ib0,i1))!=0) return -1;
      }
#endif
    }
  }
  return 1;
}

/* calculate smallest lm! (enough for smallest)
   ToDo: hard code this part or examine only lm arround changed spins
      ns*(ns+1)/2 => ns, or lm*H as optimized C++ Operator
      or lm sorted Index (proof only idx-neighbours) or lm-table
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 7s(25%) => 4s(17%) PM600MHz-pg
 * 2015-02: rename b_smallest_lm to minsymcfg_lm
 */
__attribute__ ((hot))
tbase minsymcfg_lm(const tbase ib0, tphase *s1){  
  int b1,b2,i1,i2; tbase ib1;   // c=creation a=anihilation
#if CM(0,1,1)
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
// see utils/fermion_sign.c   Lieb-Mattis-model .-3 + u-d-equivalenz
//      JJ  uXu  dXu                                  
//      tJ  uXu  dXu   uX.   dX.  .X.                 
//   from:  uXu  dXu   uX.   dX.  .X.  3X3  3X.   3Xu      p=Parity(X)
// to LM+: +uXu +uXd -p.Xu -p.Xd  -0   -0  -.X3 -puX3  LM+ maxSS -J JJ k=0
// to LM-:  -0  -uXd +p.Xu +p.Xd +.X. +3X3 +.X3 +puX3  LM- maxJJ -U tU6.0+3 k=1
//   from:                            =.X.       =dX.  --equivalence--
//   beide Varianten erlauben und auswaehlen per k_LM=0 or 1 
// translation von 2s=2 oder 2s=4:  33X33 => + or 0, ..Xuu => +uuX..!
  // table for 16 tU-possibilities (4*site1+site0)  J=t*t/U=ci.ai.cj.aj
  // see loopN, take away from highest bit + parity(+1) to higher bits(=vacuum)
  //            put to P(i) to vacuumstate and parity to lower (new) set bits
  //    JJ      __ __ __ __ __ uu ud __ __ du dd __ __ __ __ __   =  4
  //    tJ      .. .u .d __ u. uu ud __ d. du dd __ __ __ __ __   =  9
  //    tU      .. .u .d .3 u. uu ud u3 d. du dd d3 3. 3u 3d 33   = 16
  // zero_tU                    0              0
//int fsig[16]={+1,+1,+1,+1,+1,-1,-1,+1,+1,-1,-1,+1,+1,+1,+1,+1}; // sign 4+12?
//int fsig[16]={+1,+1,+1,+1,+1,-1,-1,-1,+1,-1,-1,-1,+1,-1,-1,+1}; // sign 4+12?
//int pari[16]={ 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0}; // parity 8
#endif
  //                           ??  ?        ? ??
//  CM({},Err17,Err17)	/* signum ??? */
//  Slow but working Fallback (tJ_s1 2015-12):
#if 0  /* alternative test path */
  { struct fullsbase b; b.is[0]=*s1; b.ib[0]=ib0; b.num=1;
    ib0=allsymcfg_lm(&b, 1); *s1=b.sign; return ib0; }
#endif
  if (s1) ;             /* dummy access, avoiding cc warnings for JJ */
//  printf("\n cfg_sm_lm_i= %+g %s", Re(*s1), ccfg(ib0));
  ib1=ib0;
  for(i1=0;i1<nn-1;i1++) { /* l1=sym.lm[i1]; */
    b1=geti(ib1,i1);
    // if (b1<CM(2,2,3)) { /* b1={CM(u,.ud,.ud3)} removed 2015-12 */
    for(i2=sym.ln[i1];i2<nn;i2=sym.ln[i2]) /* i2>i1, if(l1==sym.lm[i2]) */ {
      b2=geti(ib1,i2);
      if (b1<b2) {  /* wrong order, replace minimum */
        ib1=puti(puti(ib1,i1,b2),i2,b1); b1=b2; /* exchange sym-sites */
        b2=geti(ib1,i2);
#if CM(0,1,1) // 2015-12-31 recoded (similar to allsymcfg_lm), bad before
        /* parity for odd number of sites */
        if (((b1^b2)^((b1^b2)>>1))&1) { /* examples: .xu ux3 (fermionic transfer) */
          (*s1) *= parity(ib1 & p_mask[i1][i2]); }  /* not coding independent */
        if (zizj(ib1,i1,i2)==0) (*s1) = -(*s1); /* ud (fermionic exchange) */
        /* two types of LM: JJ.maxS(S+1)  and  tU.maxJ(J+1) */
        /*                 0: prefers ud+du, 1: prefers ..,.3,33 */
        if (sym.type_lm) (*s1) = -(*s1);
        // *s1 *= fsig[b2*4+b1];    /* ud (fermionic exchange) */
      }
      if (b1==b2) {
        if (sym.type_lm==0) /* JJ.maxS(S+1): T(.X.)=-.X. 2016-01 */
          if (abs(zi(ib1,i1))==0) { (*s1)=0;return TBASE0; }
        if (sym.type_lm==1) /* tU.maxJ(J+1): T(uXu)=-uXu 2016-01 */
          if (abs(zi(ib1,i1))!=0) { (*s1)=0;return TBASE0; }
#endif
      }
//      } /* skip max site cfgs */
    } /* i2=ln[i1]... > i1 */
  } /* i1=0..nn-2 */
//  printf("\n cfg_sm_lm_o= %+g %s", Re(*s1), ccfg(ib0));
  return(ib1);
}

/* is=0 dont store or at end, is!=0 to the top, .num < nj */

/* b->o: 0=getbase     sign,norm2,scfg,num,ib[],is[] ok
 *       1=minsymcfg   sign,norm2,scfg ok ???
 *       2=isminsymcfg  if(ret=1){ret sign,scfg,norm2} ok, num=0
 */
/* ToDo: replace by minsymcfg/allsymcfg + b->o=2, very similar fkt.
 * what about costs? min/mean/max clocks per call? see minsymcfg() 
 * minsymcfg is very similar, here we have only performance adaptions
 *  1=true 0=false(smaller permutation found) -1=false(destructive sign)
 *  -1 is returned after all ret=0 tests, else we have additional
 * return 0, not -1, for cfgs like (3.3333)
 *   this reduces compute time in new_numsymconf(v2.27)
 *   . means empty, 3 means up+down for the tU-model
 * called by ini_norm2 ns_thread
 * on S1SYM isminsymcfg may fail, if other syms mix S1 subsites
 *   ToDo: Permutations must not mix sites (its bitweight) within S1 groups?
 *         do it for all commuting perms!? -> example
 *   |1234> P1=(1-3 2-4) P2=(1-2) P3=(3-4) P4=P1*P3=(1-4 2-3)=not_allowed
 * ToDo: store biggest (max) and/or smallest (min) symcfg???
 * ToDo: better use ... static tbase &b.cfg, tbase &b.scfg ... for -O2 2016
 * 2016-01: does not return useful norm2 (S1SYM can couse OVLs for rekursiv.ns)
 */
__attribute__ ((hot)) /* gcc-4.3 and better, hotspot function, if S1SYM */
int isminsymcfg_lNbrk(const tbase ib1, struct sbase *b){ 
  t_sites k1[NN],k2[NN],*k3,i2; int n2;
  register int i,j,d; tphase s1,s2,s3;
  /* move k1,k2,ks to sym.buff for L1-caching? NN*MaxSym=40*410 */
  tbase ib2=ib1;
  b->sign=1; n2=1; s2=+1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM) /* return isminsymcfg(ib1,b); */
  if (sym.num_lm!=nn) { /* check for LM-symmetric sites */
    int rc;
    if ((rc=isminsymcfg_lm(ib1))!=1) return rc; /* never hits!? tU? 2015-12 */
    if (nu==nd && sym.wud) {
       // tU 0x36ffff (LM2) generates minus sign from odd double sets
       //   this can lead to bad sign later  2016-01-12
       tbase ib3=ib2;s3=s2;
       ib2=minsymcfg_lm(ib2,&s2); /* smallest inverse cfg */
       if (s2==0) { /* destructive sign, tU+tJ only, never hits? */
         E1(" %s",ccfg(ib1));
         E2(" %s %+.0f",ccfg(ib3),1.*D(s3));
         E1(" %s",ccfg(ib2));
         ERR(9999); return -1;
       }
    }
  }
#endif
  if (nu==nd && sym.wud) {
    if (!(ib1<=ib2)) return 0; /* inverse config is smaller */
    if ( (ib1==ib2)) { // would also be done in the permutation loop?
      if (s2!=1) { b->sign=0; return -1; } /* destructive phase, tU only? (.333) */
      n2=2;
    }
    for(i=0;i<NN;i++) k2[i]=geti(ib2,i);
  }
  // NN makes optimization easier for some compiler
  for(i=0;i<NN;i++) k1[i]=geti(ib1,i);  /* 30% CPU */
  /* _lm may generate -1 for tU, so old s3=1 fails! */
  for(i2=0,k3=k1,s3=1; ; i2++,k3=k2,s3=s2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){
      // sym.i[nn..NN-1] must be filled with unit operator [k]=k
      i=nn; do { i--;	/* Tj, high -> low, loop unroll2? */
        /* backward loop can be replaced by forward loop easily */
        //  4 byte = 1 int aus ram lesen und unloop 4
        /* permutation, 40%..73% CPU */
        d= (           k1[i  ])
          -( k3[ sym.i[j][i  ] ]);
      } while (d==0 && i!=0);       /* permute as long as bits are equal */
      if (d< 0) continue;           /* permutation is greater */
      if (d!=0) return 0;           /* permutation is smaller */
#if CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[j];                  /* T|x>=t|x> same config */
#elif CM(0,1,1)  /* signum is cfg-dependend */
      tbase ib4; s1=1;
      symcfg_loopN(((i2)?ib2:ib1),j,&ib4,&s1,0); /* get sign */
      s1*=sym.w[j];
#endif
      if(!phase_eq(s1*s3,1)) {s2=b->sign=0;break;} /* return (3.3333)=0 not -1, ToDo! */
      n2++;
    } /* for(j=0;j<sym.numall;j++) */
    if(nu!=nd || i2>=1 || (!sym.wud)) break; /* Sep00 */
    IFww( break; )
  }
  b->scfg=ib1; // b->num=0; /* not given here */
  b->norm2=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* OK ??? */
#ifdef CONFIG_S1SYM
#if 0 /* ToDo2016 try to work without norm2, it triggers OVL on rekursiv.ns */
  /* 2015-12-31 sgn!=0-speedup?=no ToDo: skip if norm2 not needed */
  if (sym.num_lm!=nn && !phase_eq(b->sign,0)) 
  { Tnorm2 factor=b_factor_lm(ib1); /* check for 64bit overflow */
    /* we do not care about low digits not stored in double!
     * integer-only-test 2016-01-18 */
    if ( ((Tnorm2)1)/2==0 /* integer only test */
     && b->norm2*(double)factor >= (~(Tnorm2)1)<<(sizeof(Tnorm2)*8-2))
      {E1("isminsymcfg_lNbrk n2 OVL %.1e",b->norm2*(double)factor);ERR(9112);}
    b->norm2 *= factor;	/* Ok ??? */
  }
#endif
  b->norm2=0; /* ismin... gives no return value, dont use 2016-01-16 */
#endif
  if (b->sign==0) return -1; /* Oct07 2.27 */
  /* b_out(b,0); */
  return 1;
}

/* BN(n) = Benes network of stage n = BFLY(n) + iBFLY(n)
 *  hierarchical:
 *   removing the first and last stage of BN(n)
 *   results in two BN(n/2)
 *  see utils/benes[1-3].c for more info
 * ToDo: use tbase + S_tbase=8,16 Bytes etc. ??
 */
#if   CM(1,2,2)*NN<=64
#define Ld_bits 6   /* log-dualis of 64 = 6 */
#elif CM(1,2,2)*NN<=128
#define Ld_bits 7   /* log-dualis of 128 = 7 */
#elif  CM(1,2,2)*NN<=256
#define Ld_bits 8   /* log-dualis of 256 = 8 */
//#error "CM(1,2,2)*NN>128 not implemented yet"
#elif  CM(1,2,2)*NN<=512
#define Ld_bits 9   /* log-dualis of 512 = 9 */
//#error "CM(1,2,2)*NN>256 not implemented yet"
#else
#error "please define Ld_bits for NN>512"
#endif

#define Bits (1 << Ld_bits)         // 1<<6 == 64

/* 2015-05 64bit defined, ToDo: tbase only to reduce test-cases */
#if   CM(1,2,2)*NN <= 32 /* 32bit test 2015-09-23 */
# define Bits_t unsigned int
#elif CM(1,2,2)*NN <= 64 /* minimum 64bit, not smaller!? 2015-09-23 */
# if   __WORDSIZE >= Bits
#   define Bits_t ULong
# elif __WORDSIZE >= 64
#   define Bits_t uint64    /* 2015-09-23 not available on gcc -m32 */
# elif  S_tbase*8 >= 64     /* S_tbase is in Bytes! see spins.h */
#   define Bits_t  tbase    /* __int128 or c++.vlint */
# else  // NN <= 64 but no int64
#   error "panic, no int64"
# endif
#elif CM(1,2,2)*NN <= S_tbase*8 /* NN>64 */
#  define Bits_t  tbase
#  if (NN&(NN-1))!=0 /* 2015-09-23/28 test NN=400, 2016-01 fail NN=128+LM */
//#    error "NN must be a power of 2 for NN>64"
// test m_1d 28 N=64 N=100 N=128 ok 2016-01-08 gcc492+__i128
#  endif
#else /* never happens? tbase always defined or error define tbase */
# error _no_matching_wordsize_found
#endif

/*  masks for benes network bit permutation (one for each symmetry)
 *  size = numsym * 2log2(NN=64)
 *  during minsymcfg_blk_*() sym_bn_m should be in the L1 cache
 *  size = (sym=160)*11*8B = 14kB
 * ToDo: replace 11 by 2*LD_NN-1, LogDualis LD(33..64)=6 LD(65-128)=7
 *    getconf LEVEL1_DCACHE_LINESIZE  # 64
 * cat /sys/devices/system/cpu/cpu0/cache/index0/size # 32K haswell
 */
tbase sym_bn_m[MaxSym][2*Ld_bits-1] __attribute__ ((aligned (64)));

typedef struct {
  Bits_t b1[Ld_bits], b2[Ld_bits]; // butterfly configuration
 } tr_benes;
#define no_index ((int)(-1))
#define lo_bit   ((Bits_t)1)  /* 2016-01 outer braces () fix lo_bit << 80 */
void invert_perm(const int src[Bits], int tgt[Bits]) {
  int i;
  for (i = 0; i <= Bits-1; ++i) tgt[i] = no_index;
  for (i = 0; i <= Bits-1; ++i) if (src[i] != no_index) tgt[src[i]] = i;
}
static void exchange_bit_index(int* a, int* b) { // INLINE
  int q; q = *a; *a = *b; *b = q;
}
static void gen_benes_ex(tr_benes* self, const int c_tgt[Bits]) {
// Generate a configuration for the Benes network with variable stage order.
// Use benes_fwd_ex and benes_bwd_ex.
// Algorithm as sketched by Donal E. Knuth,
//   The art of computer programming, vol. 4, pre-fascicle 1a
// Implemented 2011-12-02 by Jasper L. Neumann
// Modified 2012-08-31 to allow for "don't care" entries.
// ToDo: modify for 128bit++
//  const unsigned int a_stage[Ld_bits] = {5,4,3,2,1,0}; // perm_b64.h
  int src[Bits], inv_src[Bits];
  int tgt[Bits], inv_tgt[Bits];
  int stage;
  int mask;
  Bits_t cfg_src, cfg_tgt;
  Bits_t src_set;
  int main_idx, aux_idx, src_idx, tgt_idx, idx2;
  int stage_idx;
  int s;
  for (s = 0; s <= Bits-1; ++s) { src[s] = no_index; tgt[s] = no_index; }
  for (s = 0; s <= Bits-1; ++s) {
    if (c_tgt[s] != no_index) { tgt[s] = s; src[c_tgt[s]] = s; }
  }
  invert_perm(src,inv_src);
  invert_perm(tgt,inv_tgt);
  for (stage_idx = 0; stage_idx <= Ld_bits-1; ++stage_idx) {
    stage = Ld_bits-1-stage_idx;  src_set = 0;  /* stage= {5,4,3,2,1,0} */
    mask = lo_bit << stage;       cfg_src = 0;    cfg_tgt = 0;
    for (main_idx = 0; main_idx <= Bits-1; ++main_idx) // This order to meet
    if ((main_idx & mask) == 0)   // low only
    for (aux_idx = 0; aux_idx <= 1; ++aux_idx) {
      src_idx = main_idx+(aux_idx << stage);
      if (((lo_bit << src_idx) & src_set) == (Bits_t)0)   // yet unhandled
      if (src[src_idx] != no_index)  // not open
      do {
        src_set = src_set | (lo_bit << src_idx);
        tgt_idx = inv_tgt[src[src_idx]];
        if (tgt[tgt_idx] == no_index) break;  // open end

        if (((src_idx ^ tgt_idx) & mask) == 0) // straight
          tgt_idx = tgt_idx ^ mask;
        else {
          // cross
          cfg_tgt = cfg_tgt | (lo_bit << (tgt_idx & ~mask));
          idx2 = tgt_idx ^ mask;
          exchange_bit_index(&tgt[tgt_idx], &tgt[idx2]);
          inv_tgt[tgt[idx2]] = idx2;
          if (tgt[tgt_idx] != no_index) inv_tgt[tgt[tgt_idx]] = tgt_idx;
          }  
        if (tgt[tgt_idx] == no_index) break;  // open end
        src_idx = inv_src[tgt[tgt_idx]];
        if (((src_idx ^ tgt_idx) & mask) == 0) { // straight
          src_set = src_set | (lo_bit << src_idx);
          src_idx = src_idx ^ mask;
        } else { // cross
          cfg_src = cfg_src | (lo_bit << (src_idx & ~mask));
          idx2 = src_idx ^ mask;
          src_set = src_set | (lo_bit << idx2);
          exchange_bit_index(&src[src_idx], &src[idx2]);
          inv_src[src[idx2]] = idx2;
          if (src[src_idx] != no_index) inv_src[src[src_idx]] = src_idx;
          }  
        if (src[src_idx] == no_index) break;  // open end
        if (((lo_bit << src_idx) & src_set) != (Bits_t)0) break; 
        } while (1);
      } // if       
    self->b1[stage] = cfg_src;
    self->b2[stage] = cfg_tgt;
    }
}
// 2015-09-22 called by spins.c */
int sym_bn_m_ini_flag=1; // needs initialization
void sym_bn_m_ini(int pt){  // ToDo2015: what about odd N ???
  int i1, i2, i3, p1[Bits]; tr_benes benes;
  if (!sym_bn_m_ini_flag) return; /* once! 2015-12-31 */
  sym_bn_m_ini_flag=0;  // reset needs_ini_flag
  i1= (int)((size_t)(&(sym_bn_m[0][0])) & 4095); i1=i1-(i1&(i1-1));
  i2= (int)((size_t)(&(   sym.i[0][0])) & 4095); i2=i2-(i2&(i2-1));
  if (vvv && (!mpi_i) && !pt){ /* task 0 && thread 0 */
    printf("DBG: Ld_bits= %d tbase= %d (2^n)\n",Ld_bits,(int)sizeof(tbase)*8);
    printf("DBG: sym_bn_m size=%8d alignment=%5d MaxSym=%d (*%d*%dB)\n",
      (int)sizeof(sym_bn_m), i1, MaxSym, 2*Ld_bits-1, (int)sizeof(tbase1));
    printf("DBG: sym_i    size=%8d alignment=%5d MaxSym=%d (*%d*%dB)\n",
      (int)sizeof(sym.i), i2, MaxSym, NN, (int)sizeof(t_sites));
  }
  for (i1=0;i1<sym.numall;i1++){
    for (i2=0;i2<nn  ;i2++) p1[i2]=(int)sym.i[sym.inverse[i1]][i2];
    for (    ;i2<Bits;i2++) p1[i2]=i2;
//    if (i1<4 && !mpi_i){ printf("P%d:",i1+1);
//      for (i2=0;i2<Bits;i2++) printf(" %d",p1[i2]);printf("\n"); }
    // bn_get_mask(bitsIn, bitsOut, NN, bn_m); /* n a power of 2 only ??? */
    gen_benes_ex(&benes,p1); /* copy + compress masks */
    for (i3=0;i3<Ld_bits;i3++) { sym_bn_m[i1][i3]= benes.b1[Ld_bits-1-i3]; }
    for (i3=0;i3<Ld_bits;i3++) { sym_bn_m[i1][Ld_bits-1+i3]= benes.b2[i3]; }
//  if(i1<4 && !mpi_i) for (i3=0;i3<2*Ld_bits-1;i3++) 
//    printf("BN%02d %016lx\n",i3,sym_bn_m[i1][i3]);
  } // i1=sym.numall-loop
}
/* 2015-02-06 IBFLY =  8IOPs = 1ANDN+2AND+3SHIFT+2OR = bad
 * 2015-04-28 IBFLY =  9IOPs = 1ANDN+2AND+3SHIFT+3OR = OK 
   11bfly 9i+1read or 6i+2read? read2=~(msk|(msk<<s))
   superscalar: ~(msk|(msk<<s)) = 3 clks independent from others 1ALU
                rest_and_shift  2 clocks in 2ALUs + 2 ors in 1 ALU 
                = 4 clocks * 11 BFLYs + 1..3clocks pipefill using 3 ALU64s
       but bn1= 25ns*1.6GHz= 40 clocks / 11bfly = 3.64 clocks/bfly !!!
       // -mdump-tune-features
    // gcc generiert VSHIFT Befehle psllq+psrlq fuer core2,sandybridge
    // atomN455.pineview64bit  2ALUs(SIMD=2i64) alu^shift+alu
    // MIPS-5kf  evl. not enough register? L1-latency or 2ALU,1SHFT  67clk
    //               2ALUs V=3 3*4clks + parallel next mask = 11*4+3=47clk
    //                     V=3 best = 49clks, (using nor V=2 shoul best 46clk)
    // k8.SSE2_64b   2VALUs+1ALU           SSE-path is 64bit! 3ALUs
    // k10             2VALUs?+3ALUs
    // core2=SSE2=   3ALUs(SIMD=2i64)      (penryn?)        (psllq+psrlq+notq) 
    // nehalem:      2(alu^shift)+alu      2valu+1shift noVSHIFT?
    // sandyBridge=  3ALUs(SIMD=2i64)+3mem vshift+valu+valu (psllq+psrlq) 13clk
    // haswell=AVX2= 4ALUs(SIMD=4i64)+4mem vshift^vlog+2vlog+alu^shft(1vshift only)
    //    does vect-logical includes shift? ToDo
    1data64-version:
 ##  ALU1         ALU2        ALU3 (ToDo: explore by synthetic tests)
 -3: mask=msk<<s;            // independent = 1 clock           +1ALU
 -2: mask|=msk;              // depends-3 = 1 clock +wait -3.
 -1: imsk=~mask;             // depends-2 = 1 clock +wait -2.
  0: i2b=i2&msk;  i2c=i2>>s;  +ndat( 0) // depends+3 = 1 clock +wait pre3,-1. *2ALUs
  1: i2b<<=s;     i2c&=msk;   +next(-3)
  2: i2a=i2&imsk; i2=i2b|i2c; +next(-2)
  3: i2|=i2a;     +next(-1)   +ndat( 0)  // go step 0 for next BFLY (11*)
     // 1data:  11BFLY * (4clocks * 3ALUs) + 3waits * 1ALU  =  47   clk/sym
     //    or:  11BFLY * (6clocks * 2ALUs) + 3waits * 1ALU  =  69   clk/sym
     //    or:  11BFLY * (10clocks * 1ALU) + 0waits         = 110   clk/sym
     // 2data+unroll2: 2nd data needs     + 3clocks * 2ALUs =  80/2 clk/sym
     //    or:         2nd data needs     + 3clocks * 2ALUs = 102/2 clk/sym
     //    or:         2nd data needs     + 7clocks * 1ALU  = 187/2 clk/sym
     // 4data+unroll4: 3th+4th data needs + 4clocks * 3ALUs = 124/4 = 31.0
     //    or:         3th+4th data needs + 6clocks * 2ALUs = 168/4 = 42.0, size=4*168*4B?
     // SSE2-SIMD-vect2:                                       47/2 = 23.5
     // vect2+unroll4:                                        124/8 = 15.5
     // AVX2-SIMD-vect4:                                       47/4 = 11.8
     // vect4+unroll4:                                        124/16=  7.8
     // MIPS-nor: d=~(s|t); reduces 1 wait unroll2:
     // MIPS: 1shift+2ALUs(incl.NOR): 1data=11*4clk(1shft+2alu(nor))+2wait=46
     //       unroll4: (4*2+1)shifts+4*3and+4*2or+1nor  (8reg/data)
     //                = 9shifts + 21ALs/2ALUs = 11*11/4 = 30.3 clk/sym
     //                                   vs. pathcc = 50 clocks (*2!)
     // MMX  pandn(dst,src): dst=(~dst)&src; = 1clk (reduces 3waits to 2 waits)
     //      wins 1/2 clock per unroll (3ALUs)
     // AVX512 can do -1,2,3 together = 3clk+2wait, 2nd data needs 
     // AVX512: replaces 2 dependend ors (2clk) and ~(1wait) to one clk
     //         or VPERMT2PD ??? 1clk/permute not vectorizable 2015 = 1 clk
     // ToDo: compare with real systems!
 */
//  after 2015-10-08 factor 2 correction fits better to theory above!
// JS-PC-1.6GHz bnV8 MACRO1=16.06ns=25.7clks MACRO2=14.82ns=23.72clks
//#define IBFLY(i2,msk,s) {i2=(i2&~(msk|(msk<<(s))))
//                 |(((i2&msk)<<(s)) |((i2&(msk<<(s)))>>(s)));}
#define IBFLY(i2,msk,s) i2=((i2)&~((msk)|((msk)<<(s))))\
                        |((((i2)&(msk))<<(s))|(((i2)>>(s))&(msk)))
/* benchmarks/fastest-speed: (ToDo: perf available? bench i99)
bad quanta1+2 2*16*2t-core-XeonE5-4620-2.2GHz-16MB 768GB gcc-4.9.2 AVX
bad   bn16: 3.6ns/loop*2200MHz = 8 clks/loop  =  11 OPs/clk  = 24.6 GIOPs
    comp2 22nodes*256GB 16*AMD-Opteron6134-512KB-2300MHz gcc-4.9.2 3dNOW
      bn8:  5.6ns/loop*2300MHz = 12.9clks/loop = 7.7 OPs/clk = 17.7 GIOPs
      bn12: 5.4ns SH=12.3ns*160
    J-PC-DualCore2-E5400-2.70GHz L2=2MB 1*2GB gcc-4.9.2 -march=core2 SSE2
      bn8:  7.4ns/loop*1600MHz = 11.8clks/loop  = 8.4 OPs/clk = 13.4 GIOPs
    Atom-N455: bn16              22.12clks/1GHz = 4.5 OPs/clk=IPC
    AVX512: ~i1 | i2 | i3 made in one step (bitwise ternary logic, 8bit, 3reg)
 */
__attribute__ ((hot))  /* put to section .text.hot (dflt: .text.unlikely) */
// static 
void symcfg_bn(const tbase  ib1, const int ip,
                     tbase *ib2, tphase *ss, const int ms){
  tbase i2;
#if CM(0,1,1)
  tbase i3; int m1, m2, s;
#endif
  *ib2=ib1; *ss=+1; if(ip<0) return; /* 1-Operator (0,identity) */
#if CM(0,1,1)
  // printf("\n symcfg_i= %s %d",ccfg(ib1),ip);
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
  // from:  uX.   dX.  3X.  dXu        3Xu   3Xd  see utils/fermion_sign.c
  // to:  +p.Xu +p.Xd +.X3 -uXd      +puX3 +pdX3   p=Parity(X)
  i3=i2=0l; s=0;			    /* remove from NN-1 to 0 */ 
  for(m1=2*nn-1;m1>=0;m1--)if(1&(int)(ib1>>m1)){ /* m1 = old index */
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	    /* m2 = new index */
    i2 |= ((tbase)1) << m2;		    /* move spin to site m2 */
    s^=(1&(int)(i3>>m2));     /* number of spins jumped over */ 
    i3 ^= (~(tbase)0)<<m2;		/* derived fermionic sign, 64bit?? */
  }
  // printf("\n symcfg_o= %s %d",ccfg(i2),s);
  (*ss) *= 1-2*(s&1);   /* 2016-01 ToDo: return s in units of I*2Pi/2 */
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  const tbase *m= sym_bn_m[ip]; // bn = bfly + ibfly mask
  i2=ib1;
  // t_sites *a=sym.i[ip];
  // printf("DBG BFLY00 cfg=%s %016lx m=%016lx\n",ccfg(i2),i2,m[0]);
/* (x&~(m|(m<<s))) | ((x>>s)&m) | (x&(m<<s)) # 2shift+3or+3and+1not
   mips-pathccO3=67clk uses nor as not     # 2shift+3and+3or+nor=11*6*2ALUs
   mips-pathccO2=84clk 22*dsll+11dsrl+   O1=120clk=8clks(SHFT+ALU+2mem)
              Os=83clk pipeline-stalls?  max(1ALU)=10*11=110clk
              best-V=3: 49clk=11*4++
   gcc492 -O2 core2: 25ns*1.6GHz=40clks = 3.64clks/BFLY (noSSE,faster theory!?)
   gcc491 -m64 AVX2: 11ns*2.4GHz=26clks = 2.40clks/BFLY (arch=core2) boost?
   gcc491 -m64 AVX2: 11ns*2.4GHz=25clks = 2.31clks/BFLY (noSSE,AVX used) boost?
   core2-copy  AVX2: 11ns*2.4GHz=27clks = 2.48clks/BFLY
   core2-copy  atom: 57ns*1.0GHz=57clks = 5.15clks/BFLY (in-order) SH=109ns*160
   gcc472 -O2  atom: 64ns*1.0GHz=64clks = 5.84clks/BFLY worse but  SH= 99ns*160
   gcc472 -O3  atom: 51ns*1.0GHz=51clks = 4.64clks/BFLY
    atom: 1*shiftALU+1ALU (2*64bit-ALUs == 1*128bit-ALU)  10OPs as 5+5OPs = 5clks
          but bn16=fastest!     24.4clks/11BFLY= 2.2clks/BFLY(128bit) :: 4.4*64bit
           = 128bit*unroll8 psrlq+pand+psllq+por  (-pragmas) +prepare-64bit-mask
   AVX2: trotz arch=core2 slightly different code! sall + shldl +andn (no salq +not) unroll?
     Todo: proof!
   AVX512: 2clk(shifts/and)+1clk(ternary~||) 4parallel pipelined = 1clk VS=8*4
   3shift+3and+3or=9OPs
        movq    8(%r9), %r11           # = m1           start BFLY16
        movq    %r11, %rax
	salq    $16, %rax              # m1 << 16    overlaps BFLY32 clk-3
        andq    %r11, %r10             # b1&m1                       clk+0 dep
        orq     %r11, %rax             # m1|(m1<<16) overlaps BFLY32 clk-2
        salq    $16, %r10              # (b1&m1)<<16                 clk+1
        notq    %rax                   # ~mask       overlaps BFLY32 clk-1
        andq    %rdi, %rax             # b1&mask                     clk+0
        shrq    $16, %rdi              # b1>>16                      clk+0
        # next3 dependend steps could be done in one ternary OP = 1 clk
        andq    %r11, %rdi             # (b1>>16)&m1                 clk+1
        orq     %r10, %rdi             # ... | ((b1&m1)<<16)         clk+2
        orq     %rax, %rdi             # ... | (b1&mask)             clk+3
                                              SUM:    3*ALUs64bit * 4clocks
        (a|b|c) could be in theory done in 1 clock, is that implemented?
        atom has same commands slightly reordered
 */
#define TEST_BFLY_128 1 // slower(?) but code for CM(1,2,2)*NN>128
/*
 4+36  0=7.36ns 1=11.32ns gcc492 jj-chain NN=64        2015-09-23 ToDo!
 19+21 0=38.6ns 1=62.00ns gcc492 jj-chain NN=100 lNbrk=17ns!
 49+51 0=38.5ns 1=62.00ns gcc492 jj-chain NN=100 lNbrk=18ns!
 m_1d 28 N=64 J1.6GHz g492-O2+ur+vec-k8  bn8=43ns  loop_bn8=54ns SH=0.34m 2016
 m_1d 28 N=64 J1.6GHz g492-O2+ur+vec-k10 bn8=45ns  loop_bn8=55ns SH=0.35m 
 m_1d 28 N=128 1.6GHz g492-O2+ur+vec-k8  bn8=161ns loop_bn8=267ns lNbrk=87ns  
 m_1d 28 N=100 OK  cxx0_0+ n1=361270  E0= -12.43764754 -11.82997466
*/
#if  TEST_BFLY_128 || Ld_bits>7   
  /* gcc492-funroll-loops is slower than manual unroll! */
  { int mi; 
    // unroll should be at same speed as below code, but is -10% 2015-09
    for(mi=0;mi<Ld_bits  ;mi++)IBFLY(i2,m[        mi],1<<(Ld_bits-1-mi));
    for(mi=0;mi<Ld_bits-1;mi++)IBFLY(i2,m[Ld_bits+mi],2<<mi);
  }
#elif Ld_bits <= 6    // faster than the upper code, but should same! 2015-09
#if NN>CM(32,16,16)
  IBFLY(i2, m[0],32); // 11* 9 instr. = bad (11*10!)
#endif
  IBFLY(i2, m[1],16); //  bn1=12/1i64=40/44clks = ok
  IBFLY(i2, m[2], 8); //  avx 7/2i=20/57clks = bad
  IBFLY(i2, m[3], 4);
  IBFLY(i2, m[4], 2);
  IBFLY(i2, m[5], 1); /* Ld_bits-1 */
  IBFLY(i2, m[6], 2);
  IBFLY(i2, m[7], 4);
  IBFLY(i2, m[8], 8);
  IBFLY(i2, m[9],16);
#if NN>CM(32,16,16)
  IBFLY(i2,m[10],32);
#endif
#elif Ld_bits <= 7     /* 128 bits */
  IBFLY(i2, m[0],64); // 13* 9 instr.
  IBFLY(i2, m[1],32); //  PC  
  IBFLY(i2, m[2],16); //  avx 
  IBFLY(i2, m[3], 8); // avx2=
  IBFLY(i2, m[4], 4);
  IBFLY(i2, m[5], 2);
  IBFLY(i2, m[6], 1);
  IBFLY(i2, m[7], 2);
  IBFLY(i2, m[8], 4);
  IBFLY(i2, m[9], 8);
  IBFLY(i2,m[10],16);
  IBFLY(i2,m[11],32);
  IBFLY(i2,m[12],64);
#else
#error panic
#endif
  // printf("DBG BFLY11 cfg=%s %016lx\n",ccfg(i2),i2);
//  for (m1=0;m1<NN;m1++) /* loopN to compare speedup */
//    i2 |= ((tbase)(1& (ib1>>m1))) << a[m1];  /* 40* 4 instr. 44clks */
#endif /* JJ */
  *ib2=i2;
  if(ms){
#if CM(1,0,0)
    (*ss)=sym.w[ip];   /* better add fractional angles (integer) outside */
#else
    (*ss)*=sym.w[ip];   /* better add fractional angles (integer) outside */
#endif
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    if(fabs(Im(*ss))<1e-9) (*ss)=C(Re(*ss),0);
#endif
  }  
  return;
}
//#pragma GCC push_options
//#pragma GCC reset_options
//#pragma GCC optimize ("O2","unroll-loops","tree-vectorize","unsafe-loop-optimizations")
//  #pragma GCC optimize ("tree-vectorize","unroll-loops")
// check assembler result gcc -S -dA
// J-PC-E5400 minsymcfg_bn    t[ns]/cfs=   24.71 = 39.5 clks
// J-PC-E5400 minsymcfg_bnV8  t[ns]/cfs=    7.48 = 12.0 clks fastest!?
//   
// 2015-04-28 use blocked data (see minsymcfg_blk_bnVS SIMD version)
__attribute__ ((hot))
void minsymcfg_blk_bn(thbuf *hbuf){ 
  tbase ib1, ib2, ib3, ib4, scfg; // struct sbase *b, tb;
  int ii,j,i2,n2,norm2; tphase s1,s2,s3,sgn;
//  b=&tb; /* for minimum source changes against old code, test code */
  for (ii=0;ii<hbuf->n;ii++) {  ib1=hbuf->acj[ii];
    /* start old code minsymcfg_blk_bn(ib1, b); */
    sgn=1;  scfg=ib1; norm2=1; n2=1; ib2=ib1; s2=1;
    if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
    if(sym.num_lm!=nn){ scfg=       ib1=minsymcfg_lm(ib1,&(sgn));
             if (nu==nd && sym.wud) ib2=minsymcfg_lm(ib2,&s2);
             if (s2==0 || sgn==0){ s2=sgn=0; }
             else norm2 *= b_factor_lm(ib1); } /* fermionic sign? */
#endif
    if (nu==nd && sym.wud){ /* better do ib2==ib1 later, because its rarely */
      if( (ib2==ib1)){ n2=2; if(s2!=1) s2=sgn=0; /* s1!=1 ToDo: JJ only? */}
      if(!(ib1<=ib2)){ ib3=ib1;scfg=ib1=ib2;ib2=ib3; sgn=s2;s2=1;}/* newmin */
    }
    /* search for minimum symmetric config as representator */
    /* if !sgn break added 2015-12-31 */
    for(i2=0,s3=sgn,ib3=ib1; ; i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
      /* ToDo: vectorize! j+=4 + rest (+bn replacing loopN) */
      for(j=0;j<sym.numall;j++){	/* take a SymOp=PermOp(j)=ps */
        /* better call/inline symcfg_loopN ? */
     // symcfg_loopN(ib3,j,&ib4,&s1,0);
        symcfg_bn( ib3,j,&ib4,&s1,0);
        s1 *= sym.w[j];
        // if (ib4> scfg) continue; /* its greater, forget it (how often?) */
        //s1=sym.w[ j ]; 
        if (ib4==scfg) {    /* same (how often?) phase -> int */
          n2++; if(!phase_eq(s1*s3,sgn)){s2=sgn=0;/*norm2=0;s1=0;*/}
          // continue;
        } /* else: found smaller cfg */
        if (ib4< scfg){scfg=ib4; n2=1; sgn=s1*s3;} /* better use k(2PIi/l) */
      }   /* test next sym */
      if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
      IFww( break );
    }     /* test ud-sym */
    norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
    /* if(norm2) norm2=(unsigned Tnorm2)norm2[scfg2idx(scfg)]; */
    // if (sgn) 
    hbuf->acj[ii]=scfg; // else hbuf->acj[ii]=TBASE0;
    hbuf->asj[ii]=sgn;
    hbuf->an2[ii]=norm2;
  } /* for (ii=0;ii<hbuf->n;ii++) */
  return;
}

//#pragma GCC push_options
//#pragma GCC optimize ("unroll-loops")
/* VS + HBLrVS is defined in spins.h (ToDo: move to hilbert.h?)
 comp2+GNUC 4.1.2: SSE2 3dNOW 6+34 2015-02-23
 minsymcfg_lNbrk t[s]= 1.54 loops/s= 1.4e+06 t[ns]/cfs=      4.6 0
 minsymcfg_loopN t[s]= 1.02 loops/s= 2.6e+05 t[ns]/cfs=     24.4 0 4Ni
// bad old 88i, new=99i! ToDo: replace 88i
# new comp2-amd6134-3dNOW-2.3GHz gcc492.amdfam10
 minsymcfg_dflt  t=1.6s loops/s= 8.0e+01 t[ns]/cfs=    6.37 0 bn8    SH=12.4ns*160
 minsymcfg_lNbrk t=1.5s loops/s= 1.4e+06 t[ns]/cfs=    4.60 0 ud=6+34 SH=9.5ns*160
 minsymcfg_lNbrk t=1.1s loops/s= 9.8e+05 t[ns]/cfs=    6.36 0 ud=19+21 14.6clk/2.3GHz
 minsymcfg_bn    t=1.7s loops/s= 3.9e+05 t[ns]/cfs=   15.97 0 99i
 symcfg_bnv4     t=1.3s loops/s= 1.1e+08 t[ns]/cfg=    9.42 0 99i
 symcfg_bnv8     t=1.5s loops/s= 1.8e+08 t[ns]/cfg=    5.62 0 99i
 minsymcfg_bnVS  t=1.2s loops/s= 1.1e+06 t[ns]/cfs=    5.64 0 99i 8 13.0clk/2.3GHz
# vm8-201 E5-2660 AVX 2.2GHz clang -S -dA
 minsymcfg_dflt  t=1.3s loops/s= 4.8e+01 t[ns]/cfs=   10.53 0 bn4 clang
 minsymcfg_lNbrk t=1.2s loops/s= 1.8e+06 t[ns]/cfs=    3.48 0 ud=8+32 clang
 symcfg_bnv4     t=1.9s loops/s= 2.8e+08 t[ns]/cfg=    3.52 0 99i inlined clang
 symcfg_bnv4     t=1.9s loops/s= 2.8e+08 t[ns]/cfg=    3.58 0 99i clang !!!
 # 64bit shl,shr,andq,orq + 128bit vpunpcklqdq xmm1.. vpor vpand vpsllq
 #    + vextractf128 1,ymm0
 minsymcfg_dflt  t=1.4s loops/s= 9.0e+01 t[ns]/cfs=    5.68 0 bn4 gcc44
 symcfg_bnv4     t=1.4s loops/s= 1.9e+08 t[ns]/cfg=    5.24 0 99i gcc44
 symcfg_bnv8     t=1.3s loops/s= 2.0e+08 t[ns]/cfg=    5.00 0 99i gcc44
 minsymcfg_bnV4  t=1.4s loops/s= 1.2e+06 t[ns]/cfs=    5.40 0 99i gcc44
 # 11.9clks/2.2GHz IPC=99/11.9=8.3(i64) instructions per clock = 8.3 !!!
# RTL-haswell-PC AVX2 2.4GHz
 minsymcfg_dflt  t=1.3s loops/s= 1.9e+02 t[ns]/cfs=    2.65 0 bn32  SH=5.7ns*160
 minsymcfg_dflt  t=1.5s loops/s= 1.7e+02 t[ns]/cfs=    2.91 0 bn16  SH=6.0ns*160
 minsymcfg_lNbrk t=1.9s loops/s= 2.2e+06 t[ns]/cfs=    2.81 0 ud=10+30
 symcfg_bnv4     t=1.4s loops/s= 1.9e+08 t[ns]/cfg=    5.28 0 99i
 symcfg_bnv8     t=1.3s loops/s= 4.3e+08 t[ns]/cfg=    2.33 0 99i
 minsymcfg_bnVS  t=1.2s loops/s= 2.7e+06 t[ns]/cfs=    2.29 0 99i 16
 minsymcfg_bnVS  t=1.1s loops/s= 2.9e+06 t[ns]/cfs=    2.14 0 99i 32
 -O3 is slower than O2, so we disable
   inline-functions
   unswitch-loops
   gcse-after-reload
*/
//#pragma GCC reset_options
//#pragma GCC optimize ("O2","unroll-loops","tree-vectorize","unsafe-loop-optimizations")
// unroll+vect    VS=8 core2: 8unroll=4unroll*xmm
// no-unroll+vect VS=8 core2: 4loop*2unroll=4loop*xmm
// no-unroll+vect VS=2 core2: noloop*xmm
//   see hilbert.h for CPU-dependent best-choice table
// static inline  /* below 2% speedup, but -S -dA more difficult to find */
__attribute__ ((hot))
void symcfg_bnVS(
    const tbase ib1[VS] /* __attribute__ ((aligned (32)))*/,
                        /* 4*i64 = 32B input configs (vectorized) */
    const int ip, /* input sym_idx */
    tbase ib2[VS], /* output config_bits (vectorized) */
    tphase ss[VS], /* output sign (vectorized) ToDo: bit_is_changed */ 
    const int ms   /* input sign mode if(ms)sign*=w[k] (ToDo: needed?) */){
  /* register */ 
  tbase i2[VS] __attribute__ ((aligned (64)));
  int i;
#if CM(0,1,1)
  register int m1; /* unroll!? */ 
  tbase i3[VS] __attribute__ ((aligned (64)));
  int    s[VS] __attribute__ ((aligned (64))); register int m2;
#endif
  for (i=0;i<VS;i++) ib2[i]=ib1[i]; // auto-vectorization
  for (i=0;i<VS;i++) ss[i]=+1; // auto-vectorization
  if(ip<0) ERR(9999); // return; /* 1-Operator (0,identity) (ToDo: remove!) */
#if CM(0,1,1)
  // printf("\n symcfg_i= %s %d",ccfg(ib1),ip);
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
  // from:  uX.   dX.  3X.  dXu        3Xu   3Xd  see utils/fermion_sign.c
  // to:  +p.Xu +p.Xd +.X3 -uXd      +puX3 +pdX3   p=Parity(X)
  for (i=0;i<VS;i++) i3[i]=i2[i]=0l;
  for (i=0;i<VS;i++) s[i]=0; // auto unroll
  for (i=0;i<VS;i++)
  for(m1=2*nn-1;m1>=0;m1--)if(1&(int)(ib1[i]>>m1)){ /* m1 = old index */
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	    /* m2 = new index */
    i2[i] |= ((tbase)1) << m2;		    /* move spin to site m2 */
    s[i]^=(1&(int)(i3[i]>>m2));  /* get xor of lower i2-bits */
    i3[i] ^= ((~(tbase)0)<<m2);   /* change lower-bit-parity for higher bits */
    /* Todo: first test on bnv4, loop nn + nu+nd-unroll, remove if(m2>0)  */
  }
  // ----- tJ,tU not implemented !!! --- missing ib1[1..3]
  for (i=0;i<VS;i++) ss[i]=1-2*(s[i]&1);/* 2016-01 ToDo: return s in units of I*2Pi/2 */
  // ib2=i2;??? // for (i=0..3) ib2[i]=i2[i];
  for (i=0;i<VS;i++) ib2[i]=i2[i];
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  const tbase *m= sym_bn_m[ip]; // bn = bfly + ibfly mask
  for (i=0;i<VS;i++) i2[i]=ib1[i];
//    for (i=0;i<4;i++) {  // a&~b = a NAND b = 1 instruction
#if TEST_BFLY_128  || Ld_bits>7
  // symcfg_bnVS called by minsymcfg_blk_bnVS
  { int mi; // unroll i+mi, gcc -funroll-loops is slower than hand-unroll, why?
// lc28 N=64  for(i)for(mi)=96ns for(mi)for(i)=54ns
    for(mi=0;mi<Ld_bits  ;mi++)
      for (i=0;i<VS;i++) IBFLY(i2[i],m[        mi],1<<(Ld_bits-1-mi));
    for(mi=0;mi<Ld_bits-1;mi++)
      for (i=0;i<VS;i++) IBFLY(i2[i],m[Ld_bits+mi],2<<mi);
  }
#elif Ld_bits <= 6
#if NN>CM(32,16,16)
  for (i=0;i<VS;i++) IBFLY(i2[i], m[0],32); // 11* 8 instr.
#endif
  for (i=0;i<VS;i++) IBFLY(i2[i], m[1],16); //  PC  34/44clks
  for (i=0;i<VS;i++) IBFLY(i2[i], m[2], 8); //  avx 29/57clks
  for (i=0;i<VS;i++) IBFLY(i2[i], m[3], 4); // avx2=188i=4*11*4.2ymm
  for (i=0;i<VS;i++) IBFLY(i2[i], m[4], 2);
  for (i=0;i<VS;i++) IBFLY(i2[i], m[5], 1);
  for (i=0;i<VS;i++) IBFLY(i2[i], m[6], 2);
  for (i=0;i<VS;i++) IBFLY(i2[i], m[7], 4);
  for (i=0;i<VS;i++) IBFLY(i2[i], m[8], 8);
  for (i=0;i<VS;i++) IBFLY(i2[i], m[9],16);
#if NN>CM(32,16,16)
  for (i=0;i<VS;i++) IBFLY(i2[i],m[10],32);
#endif
#elif Ld_bits <= 7
        for (i=0;i<VS;i++) IBFLY(i2[i], m[0],64); // 11* 9 instr.
        for (i=0;i<VS;i++) IBFLY(i2[i], m[1],32); //  PC  
        for (i=0;i<VS;i++) IBFLY(i2[i], m[2],16); //  avx 
        for (i=0;i<VS;i++) IBFLY(i2[i], m[3], 8); // avx2=
        for (i=0;i<VS;i++) IBFLY(i2[i], m[4], 4);
        for (i=0;i<VS;i++) IBFLY(i2[i], m[5], 2);
        for (i=0;i<VS;i++) IBFLY(i2[i], m[6], 1);
        for (i=0;i<VS;i++) IBFLY(i2[i], m[7], 2);
        for (i=0;i<VS;i++) IBFLY(i2[i], m[8], 4);
        for (i=0;i<VS;i++) IBFLY(i2[i], m[9], 8);
        for (i=0;i<VS;i++) IBFLY(i2[i],m[10],16);
        for (i=0;i<VS;i++) IBFLY(i2[i],m[11],32);
        for (i=0;i<VS;i++) IBFLY(i2[i],m[12],64);
#else
#error panic
#endif
//     i2[i] |= ((tbase)(1& (ib1[i] >> m1))) << (int)a[m1];  /* vectorize? */
//  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
  for (i=0;i<VS;i++) ib2[i]=i2[i];
#endif /* JJ */
  if(ms){
#if CM(1,0,0)
    for (i=0;i<VS;i++) ss[i]=sym.w[ip]; // 11.7ns - 11.6ns
#else
    for (i=0;i<VS;i++) ss[i]*=sym.w[ip];
#endif
    /* better add fractional angles (integer) outside */
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    for (i=0;i<VS;i++) if(fabs(Im(ss[i]))<1e-9) ss[i]=C(Re(ss[i]),0);
#endif
  }  
  return;
}
// ToDo: adapt minsymcfg_blk_bn but replace loop call bn by call loop blk
// 2015-04-28 use blocked data
//   ToDo: replace thphase by int (use tphase outside!)
// ToDo: add pragma loop nu-unroll to lm ud loop
// not ollowed inside functions! gcc492
__attribute__ ((hot))
void minsymcfg_blk_bnVS(thbuf *hbuf/* in(acj)+out(acj,asj,an2) */){
  tbase ib1[VS] __attribute__ ((aligned (32))),
        ib2[VS] __attribute__ ((aligned (32))), *ib3,
        ib4[VS] __attribute__ ((aligned (32))),
       scfg[VS] __attribute__ ((aligned (32))); //struct sbase *b, tb;
  int ii,i,j,i2;   Tnorm2 n2[VS], norm2[VS];
  tphase s1[VS],s2[VS],s3[VS],sgn[VS];
//  b=&tb; /* for minimum source changes against old code, test code */
  /* ToDo: i<HBLrVS may be faster if NZXMAX fits better */
  if        (HBLrVS < ((hbuf->n+VS-1)/VS)*VS) ERR(9999);
  for (ii=hbuf->n; ii<((hbuf->n+VS-1)/VS)*VS;ii++){ // fill last VS
    hbuf->acj[ii]=hbuf->acj[ii-1]; // dummy data
    /* acj length must be round up to multiple of VS to avoid segfault */
    /* ToDo: fill once outside? or i<(hbuf->n | (VS-1))+1 faster? */
  }
  /* ToDo: put sym_bn_m[][] z.B. 16kB to L1 cache here! */
  for (ii=0;ii<hbuf->n;ii+=VS) {
    /* fill cpu SIMD vectors */
    for (i=0;i<VS;i++) ib1[i]=hbuf->acj[ii+i];
    for (i=0;i<VS;i++) sgn[i]=1;    /* sign of min cfg */ 
    for (i=0;i<VS;i++) norm2[i]=1;
    for (i=0;i<VS;i++) n2[i]=1;        /* multiple */
    for (i=0;i<VS;i++) scfg[i]=ib1[i]; /* min cfg */
    for (i=0;i<VS;i++) ib2[i]=ib1[i]; /* ud-sym */
    for (i=0;i<VS;i++) s2[i]=sgn[i]; /* ud-sgn */
    if (nu==nd && sym.wud) for(i=0;i<VS;i++) 
                   get_sudconfig(ib1[i],+1,ib2+i,s2+i); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
    if (sym.num_lm!=nn){ 
       for(i=0;i<VS;i++)    scfg[i]= ib1[i]=minsymcfg_lm(ib1[i],sgn+i);
       if (nu==nd && sym.wud) 
       for(i=0;i<VS;i++) ib2[i]=minsymcfg_lm(ib2[i],s2+i);
       for(i=0;i<VS;i++) if (s2[i]==0 || sgn[i]==0){ s2[i]=sgn[i]=0;/*norm2[i]=0;*/ }
       for(i=0;i<VS;i++) norm2[i] *= b_factor_lm(ib1[i]); /* fermionic sign? */
    }
#endif
    if (nu==nd && sym.wud){
      for(i=0;i<VS;i++) if (ib2[i]==ib1[i]){ /* better later, because its rarely */
          n2[i]=2; if(s2[i]!=1) sgn[i]=0; /* s1!=1 */ }
      for(i=0;i<VS;i++) if(sgn[i]!=0 && !(ib1[i]<=ib2[i])){ 
       ib4[i]=ib1[i];scfg[i]=ib1[i]=ib2[i];ib2[i]=ib4[i];/*ToDo: exchng,wozu?*/
       sgn[i]=s2[i];s2[i]=1; }/* new minimum */
    }
    /* s3 ib3 are pointers to input and udSym*input, sgn added 2015-12 */
    for(i2=0,ib3=ib1; ; i2++,ib3=ib2){  /* ud=+1,-1 */
      if (i2==0) for(i=0;i<VS;i++) s3[i]=sgn[i];
      else       for(i=0;i<VS;i++) s3[i]=s2[i];     
      /* search for minimum symmetric config as representator */
      for(j=0;j<sym.numall;j++){  /* take a SymOp=PermOp(j)=ps */
        // for (i=0;i<VS;i++) symcfg_loopN(ib3[i],j,ib4+i,s1+i,1);
        // for (i=0;i<VS;i++) symcfg_bn(ib3[i],j,ib4+i,s1+i,1);
        // symcfg_bnv8(ib3/*in_cfg*/,j/*in_sym*/,ib4/*out_cfg*/,s1/*out_sgn*/,1);
        symcfg_bnVS(ib3/*in_cfg*/,j/*in_sym*/,ib4/*out_cfg*/,s1/*out_sgn*/,1);
#if CM(1,0,0)
        for(i=0;i<VS;i++) if (ib4[i]<= scfg[i]) s1[i]=sym.w[j]; // ??? [sym.inverse[j]];
#else
	/* ToDo: multiply s1, s1[i]*=sym.w[j]; here, not in symcfg? */
#endif
        /* ToDo: howto vectorize that minimum select and sign zero? */
        for(i=0;i<VS;i++) if (ib4[i]==scfg[i]) {  /* same cfg (count) */
            n2[i]++; if(!phase_eq(s1[i]*s3[i],sgn[i])){s2[i]=sgn[i]=0;} }
        for(i=0;i<VS;i++) if (ib4[i]< scfg[i]) {  /* found smaller cfg */
            scfg[i]=ib4[i]; n2[i]=1; sgn[i]= s1[i]*s3[i];/* ToDo: int sgn */ }
      }   /* test next sym for(j<sym.numall) */
      if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 2xLoop ud+du */
      IFww( break );
    } /* test ud-sym for(i2=(0,1),ib3,s3) */
    for(i=0;i<VS;i++)
      norm2[i]*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2[i]; /* Ok? */
    /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scfg2idx(b->scfg)]; */
    for(i=0;i<VS;i++) hbuf->acj[ii+i]=scfg[i]; 
    for(i=0;i<VS;i++) hbuf->asj[ii+i]=sgn[i];
    for(i=0;i<VS;i++) hbuf->an2[ii+i]=norm2[i];
  } /* for (ii=0;ii<hbuf->n;ii+=VS) */
  /* ToDo: sym_bn_m[] may be released from L1 cache here */
  return;
}  /* minsymcfg_blk_bnVS() + _ud */


// speed test no ud-sym
// 2015-06 gcc492 --param max-unrolled-insns=2 is ignored! 
//                --param max-unroll-times=2   is ignored, unroll=4
__attribute__ ((hot))
void minsymcfg_blk_bnVS_no_ud(thbuf *hbuf/* in(acj)+out(acj,asj,an2) */){
  tbase ib1[VS] __attribute__ ((aligned (32))),
        ib4[VS] __attribute__ ((aligned (32))),
       scfg[VS] __attribute__ ((aligned (32)));
  int  ii, i, j;   Tnorm2 n2[VS], norm2[VS];
  tphase sgn[VS], s1[VS];
  if         (HBLrVS<((hbuf->n+VS-1)/VS)*VS) ERR(9999);
  for (ii=hbuf->n;ii<((hbuf->n+VS-1)/VS)*VS;ii++) // fill last VS
    hbuf->acj[ii]=hbuf->acj[ii-1]; // dummy data
  /* ToDo: hold sym_bn_m[] in the L1 cache from here, 
   *    core2 has 64B-cacheline base=64bit=8B  */
//__builtin_prefetch( sym_bn_m[0],   /* rw= */0,/*locality=*/2/*0=no,3=L1*/);
//__builtin_prefetch(&sym_bn_m[0][8],/* rw= */0,/*locality=*/2/*0=no,3=L1*/);
  for (ii=0;ii<hbuf->n;ii+=VS) {
    /* fill cpu SIMD vectors */
//__builtin_prefetch(&hbuf->acj[ii+VS],/* rw= */1,/*locality=*/0/*0=no,3=L1*/);
    for (i=0;i<VS;i++) ib1[i]=hbuf->acj[ii+i]; 
    for (i=0;i<VS;i++) sgn[i]=+1;   /* +sign of origin cfg, later of mincfg */
    for (i=0;i<VS;i++) norm2[i]=1;
    for (i=0;i<VS;i++) n2[i]=1;        /* multiple */
#if defined(CONFIG_S1SYM)
    if (sym.num_lm!=nn){ /* Howto? dont unroll or call vectorized lm */
       for(i=0;i<VS;i++) ib1[i]= minsymcfg_lm(ib1[i],sgn+i);
       for(i=0;i<VS;i++) norm2[i]= b_factor_lm(ib1[i]); /* fermionic sign? */
       /* ToDo: ib1 in extra pre-loop, norm2 after this loop? */
    }
#endif
    for (i=0;i<VS;i++) scfg[i]=ib1[i]; /* min cfg */
    /* ud-loop removed for speedup */
    { /* find min symmetric config as representator = minsymcfg_bnVS_no_ud() */
      for(j=0;j<sym.numall;j++){ /* take a SymOp=PermOp(j)=ps = data-stream */
        // gcc autounrolls it 4 times using 4*2*xmm-registers on core2
        // for (i=0;i<VS;i++) symcfg_loopN(ib3[i],j,ib4+i,s1+i,1);
        // for (i=0;i<VS;i++) symcfg_bn(ib3[i],j,ib4+i,s1+i,1);
#if CM(1,0,0)
        /* include instead of call symcfg (same speed on haswell) */
        const tbase *m= sym_bn_m[j]; // bn = bfly + ibfly mask (best if L1-cached!)
 // after unroll next loop j+2 or j+4 data
//  __builtin_prefetch(&sym_bn_m[j+3][0],/* rw= */0,/*locality=*/2/*0=no,3=L1*/);
  // loc=3 = prefetcht0   all caches
  // loc=2 = prefetcht1   n*64 L1+L2
  // loc=1 = prefetcht2   n*64 L2 only
  // loc=0 = prefetchnta  = non-temporary (?) (mips shows false sharing on threads)
  // L1-SRAM ~4cycles    4-256KiB/core
  // L2-SRAM ~10cycles  64-512KiB/core
  //  L-DRAM ~120cycles ~30ns
        for (i=0;i<VS;i++) ib4[i]=ib1[i];
#if TEST_BFLY_128  || Ld_bits>7  // minsymcfg_blk_bnVS_no_ud
  { int mi; // unroll i+mi
    for(mi=0;mi<Ld_bits  ;mi++)
      for (i=0;i<VS;i++) IBFLY(ib4[i],m[        mi],1<<(Ld_bits-1-mi));
    for(mi=0;mi<Ld_bits-1;mi++)
      for (i=0;i<VS;i++) IBFLY(ib4[i],m[Ld_bits+mi],2<<mi);
  }
#elif Ld_bits <= 6
#if NN>CM(32,16,16)
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[0],32); // 11* 9 instr.
#endif
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[1],16); //  PC  
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[2], 8); //  avx 
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[3], 4); // avx2=
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[4], 2);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[5], 1);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[6], 2);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[7], 4);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[8], 8);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[9],16);
#if NN>CM(32,16,16)
        for (i=0;i<VS;i++) IBFLY(ib4[i],m[10],32);
#endif
#elif Ld_bits <= 7
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[0],64); // 11* 9 instr.
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[1],32); //  PC  
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[2],16); //  avx 
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[3], 8); // avx2=
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[4], 4);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[5], 2);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[6], 1);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[7], 2);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[8], 4);
        for (i=0;i<VS;i++) IBFLY(ib4[i], m[9], 8);
        for (i=0;i<VS;i++) IBFLY(ib4[i],m[10],16);
        for (i=0;i<VS;i++) IBFLY(ib4[i],m[11],32);
        for (i=0;i<VS;i++) IBFLY(ib4[i],m[12],64);
#else
#error panic
#endif
        // symcfg_bnVS(ib1/*in_cfg*/,j/*in_sym*/,ib4/*out_cfg*/,s1/*out_sgn*/,0);
        for(i=0;i<VS;i++) s1[i]=sym.w[j]; /* .w should be cached 2015-09-22 */
#else /* additional fermionic sign must be computed, howto do it for bn? */
        symcfg_bnVS(ib1/*in_cfg*/,j/*in_sym*/,ib4/*out_cfg*/,s1/*out_sgn*/,1);
        /* s1 = signum * sym.w ToDo: sym.w outside? */
#endif
#if 0
        { // DEBUG-CHECK 2015-09  compare symcfg_bnVS to symcfg_loopN output
          tbase ib5; tphase s5=+1; static int dbg1=1;
          if (dbg1) { dbg1=0; printf("\nWARNING: slow bnVS==loopN-CHECK on!\n"); }
          for (i=0;i<VS;i++) {
            symcfg_loopN(ib1[i]/*in_cfg*/,j/*sym*/,&ib5/*out_cfg*/,&s5,1);
            if (ib5 != ib4[i] || s5!=s1[i]) {
              printf("\nERROR-DBG minsymcfg_blk_bnVS cfg=%s P%d %s %s\n",
                 ccfg(ib1[i]),j,ccfg(ib5),ccfg(ib4[i]));
              ERR(9999);
            }
          }
        } // DEBUG-CHECK 2015-09
#endif
        for(i=0;i<VS;i++) if (ib4[i]==scfg[i]) {  /* same cfg (count) */
            n2[i]++; if(!phase_eq(s1[i],sgn[i])){sgn[i]=0;} }
        for(i=0;i<VS;i++) if (ib4[i]< scfg[i]) {  /* found smaller cfg */
            scfg[i]=ib4[i]; n2[i]=1; sgn[i]= s1[i];/* ToDo: int sgn */ }
      }   /* test next sym for(j<sym.numall) */
    } /* test ud-sym for(i2=(0,1),ib3,s3) -- removed for speedup */
//  __builtin_prefetch(&hbuf->asj[ii],/* rw= */1,/*locality=*/0/*0=no,3=L1*/);
//  __builtin_prefetch(&hbuf->an2[ii],/* rw= */1,/*locality=*/0/*0=no,3=L1*/);
// ToDo: why lm_factor is left and n2 right of div? why div needed?
//      sqrt(sci->norm2/(double)scj->norm2);
//  OR  sqrt(scj->norm2/(double)sci->norm2); ??? can be inverted
#if defined(CONFIG_S1SYM)
    for(i=0;i<VS;i++) norm2[i]*=(sym.numall+1)/n2[i]; /* ToDo: avoid div? */
#else
    for(i=0;i<VS;i++) norm2[i]= (sym.numall+1)/n2[i]; /* Ok? speedup */
#endif
    /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scfg2idx(b->scfg)]; */
    for(i=0;i<VS;i++) hbuf->acj[ii+i]=scfg[i]; /* data stream, dont cache L1 */
    for(i=0;i<VS;i++) hbuf->asj[ii+i]=sgn[i];  /* data stream, dont cache L1 */
    for(i=0;i<VS;i++) hbuf->an2[ii+i]=norm2[i];/* data stream, dont cache L1 */
  } /* for (ii=0;ii<hbuf->n;ii+=VS) */
  /* ToDo: sym_bn_m[] can be released from L1 cache here, reload later */
  return;
} /* minsymcfg_blk_bnVS_no_ud() */


/* ToDo: 2014-06 try speedup doc/butterfly.txt ? bench it!
 * ToDo: 2014-06 try to use Generators + product-generation per graycode
 *   to reduce L1-cache needs to ngen*N + prod genorders genindizes
 *   generator - chain for all syms (1st step is finding generators)
 *   replace low order by high order generators (symmetry.c)
 *   bench it!
 */
/* --- last change 1998-169 JS --- */
/* static inline */ /* needed in spins.h + xval.c */
/*
   spins.h struct sbase, tbase scfg, Tnorm2 norm2, tphase sign, int num
   work variables:
    k1 - config ib1
    k2 - config ib1 or spin-inversed ib1 (not used for nu!=nd)
    ks - smallest config found actually 
   used by spins.c nhm() 107/135s, use b->sign, b->norm2 and b->scfg
   ToDo: remove b->bi[] array (for getbase call remalloc)
   see minsymcfg_old (old version 10times faster)
   called by: nothing (simpler code for FPGA replacement?)
   2015-02-24 renamed b_smallest_new to minsymcfg_loopN
 */
__attribute__ ((hot))
tbase minsymcfg_loopN(const tbase ib0, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  int j,i2,n2; tphase s1,s2,s3;
  tbase ib1=ib0, ib2, ib3, ib4;
  b->sign=1;  b->scfg=ib1; b->norm2=1;	/* default minimum */ n2=1;
  ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ b->scfg=    ib1=minsymcfg_lm(ib1,&(b->sign));
           if (nu==nd && sym.wud) ib2=minsymcfg_lm(ib2,&s2);
           if (s2==0 || b->sign==0){ s2=b->sign=0; return TBASE0; }
           else b->norm2 *= b_factor_lm(ib1); /* fermionic sign? */ }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
      n2=2; if(s2!=1){ s2=b->sign=0; return TBASE0;}	/* s1!=1 */ }
    if(!(ib1<=ib2)){ 
      ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; } /* new minimum */
  }
  for(i2=0,s3=b->sign,ib3=ib1; ; i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for minimum symmetric config as representator */
    /* ToDo: vectorize! j+=4 + rest (+bn replacing loopN) */
    for(j=0;j<sym.numall;j++){		/* take a SymOp=PermOp(j)=ps */
      /* better call/inline symcfg_loopN ? */
      symcfg_loopN(ib3,j,&ib4,&s1,1/* 1 until 2016-01-05 */);
      if (ib4> b->scfg) continue; /* its greater, forget it (how often?) */
//      s1 *= sym.w[j];  /* new 2016-01-05 */
      if (ib4==b->scfg) {    /* same (how often?) phase -> int */
        if(!phase_eq(s1*s3,b->sign)){s2=b->sign=0;return TBASE0;break;/*2015-09: return TBASE0;2016-01+brk*/}
        n2++; continue;
      }
      b->scfg=ib4; /* found smaller cfg */
      n2=1;
      b->sign=s1*s3; /* better use k(2PIi/l) */
    }   /* test next sym */
    if(nu!=nd || i2>=1 || (!sym.wud)) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  //b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scfg2idx(b->scfg)]; */
  return b->scfg;
}
// ToDo: adapt minsymcfg_blk_bnVS but replace loop call bn by call lNv blk
// 2015-06-04 use blocked data
//   ToDo: replace thphase by int (use tphase outside!)
__attribute__ ((hot))
void minsymcfg_blk_lNv(thbuf *hbuf/* in(acj)+out(acj,asj,an2) */){
  tbase ib1[VS] __attribute__ ((aligned (32))),
        ib2[VS] __attribute__ ((aligned (32))), *ib3,
        ib4[VS] __attribute__ ((aligned (32))),
       scfg[VS] __attribute__ ((aligned (32))); //struct sbase *b, tb;
  int ii,i,j,i2; Tnorm2 n2[VS],norm2[VS]; tphase s1[VS],s2[VS],s3[VS],sgn[VS];
//  b=&tb; /* for minimum source changes against old code, test code */
  /* ToDo: i<HBLrVS may be faster if NZXMAX fits better */
  if        (HBLrVS < ((hbuf->n+VS-1)/VS)*VS) ERR(9999);
  for (ii=hbuf->n; ii<((hbuf->n+VS-1)/VS)*VS;ii++){ // fill last VS
    hbuf->acj[ii]=hbuf->acj[ii-1]; // dummy data
    /* acj length must be round up to multiple of VS to avoid segfault */
    /* ToDo: fill once outside? or i<(hbuf->n | (VS-1))+1 faster? */
  }
  /* ToDo: put sym_bn_m[][] z.B. 16kB to L1 cache here! */
  for (ii=0;ii<hbuf->n;ii+=VS) {
    /* fill cpu SIMD vectors */
    for (i=0;i<VS;i++)   ib1[i]=hbuf->acj[ii+i];
    for (i=0;i<VS;i++)   sgn[i]=1;    /* sign of min cfg */ 
    for (i=0;i<VS;i++) norm2[i]=1;
    for (i=0;i<VS;i++)    n2[i]=1;        /* multiple */
    for (i=0;i<VS;i++)   ib2[i]=ib1[i]; /* ud-sym */
    for (i=0;i<VS;i++)    s2[i]=sgn[i]; /* ud-sgn */
    if (nu==nd && sym.wud) for(i=0;i<VS;i++) 
                   get_sudconfig(ib1[i],+1,ib2+i,s2+i); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
    if (sym.num_lm!=nn){ 
       for(i=0;i<VS;i++) ib1[i]=minsymcfg_lm(ib1[i],sgn+i);
       if (nu==nd && sym.wud) 
       for(i=0;i<VS;i++) ib2[i]=minsymcfg_lm(ib2[i],s2+i);
       for(i=0;i<VS;i++) if (sgn[i]==0 || s2[i]==0){ s2[i]=sgn[i]=0;/*norm2[i]=0;*/ }
       for(i=0;i<VS;i++) norm2[i] *= b_factor_lm(ib1[i]); /* fermionic sign? */
    }
#endif
    for (i=0;i<VS;i++)  scfg[i]=ib1[i]; /* min cfg */
    if (nu==nd && sym.wud){
      for(i=0;i<VS;i++) if (ib2[i]==ib1[i]){ /* better later, because its rarely */
          n2[i]=2; if(s2[i]!=1) sgn[i]=0; /* s1!=1 */ }
      for(i=0;i<VS;i++) if(sgn[i]!=0 && !(ib1[i]<=ib2[i])){ 
       ib4[i]=ib1[i];scfg[i]=ib1[i]=ib2[i];ib2[i]=ib4[i];/*ToDo: exchng,wozu?*/
       sgn[i]=s2[i];s2[i]=1; }/* new minimum */
    }
    /* s3 ib3 are pointers to input and udSym*input */
    for(i2=0,ib3=ib1; ; i2++,ib3=ib2){  /* ud=+1,-1 */
      if (i2==0) for(i=0;i<VS;i++) s3[i]=sgn[i];
      else       for(i=0;i<VS;i++) s3[i]=s2[i];     
      /* search for minimum symmetric config as representator */
      for(j=0;j<sym.numall;j++){  /* take a SymOp=PermOp(j)=ps */
        // for (i=0;i<VS;i++) symcfg_loopN(ib3[i],j,ib4+i,s1+i,1);
#if CM(1,0,0)
        { int m1; const t_sites *a=sym.i[j];
          tbase bt1[VS] __attribute__ ((aligned (32)));
          for (i=0;i<VS;i++) ib4[i]=TBASE0;
          for (i=0;i<VS;i++) bt1[i]=ib3[i];
          for (m1=0;m1<NN;m1++) {  // unroll + vectorize 4x-8x!
             for (i=0;i<VS;i++) ib4[i] |= B_AND1(bt1[i]) << a[m1];
             for (i=0;i<VS;i++) bt1[i]>>=1;
          }
        }
#else
        symcfg_loopNv(ib3/*in_cfg*/,j/*i_sym*/,ib4/*out_cfg*/,s1/*o_sgn*/,1);
#endif
#if CM(1,0,0)
        for(i=0;i<VS;i++) s1[i]=sym.w[j]; // ??? [sym.inverse[j]];
#else
	/* ToDo: multiply s1, s1[i]*=sym.w[j]; here, not in symcfg? */
#endif
        /* ToDo: howto vectorize that minimum select and sign zero? */
        for(i=0;i<VS;i++) if (ib4[i]==scfg[i]) {  /* same cfg (count) */
            n2[i]++; if(!phase_eq(s1[i]*s3[i],sgn[i])){s2[i]=sgn[i]=0;} }
        for(i=0;i<VS;i++) if (ib4[i]< scfg[i]) {  /* found smaller cfg */
            scfg[i]=ib4[i]; n2[i]=1; sgn[i]= s1[i]*s3[i];/* ToDo: int sgn */ }
      }   /* test next sym for(j<sym.numall) */
      if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 2xLoop ud+du */
      IFww( break );
    } /* test ud-sym for(i2=(0,1),ib3,s3) */
    for(i=0;i<VS;i++)
      norm2[i]*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2[i]; /* Ok? */
    /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scfg2idx(b->scfg)]; */
    for(i=0;i<VS;i++) hbuf->acj[ii+i]=scfg[i]; 
    for(i=0;i<VS;i++) hbuf->asj[ii+i]=sgn[i];
    for(i=0;i<VS;i++) hbuf->an2[ii+i]=norm2[i];
  } /* for (ii=0;ii<hbuf->n;ii+=VS) */
  /* ToDo: sym_bn_m[] may be released from L1 cache here */
  return;
}
//#pragma GCC pop_options

/* called by scfg2idx_err nhm_err1 nhm hamilton_nhv? wopij_block
 *  ToDo: array of ib1's, output array, return number if calculation stopped?
 *  ToDo: better return 0 or 1=smallest to replace if_smallest
 *        input ib1 via *b (less arguments, type tbase not need to declared)
 *   10times faster than old code calling symcfg_loopN numsym times
 * called by: xval.c.mrule ...
 *   tuning-test-version=minsymcfg_test3
 * 2015-02: rename b_smallest to minsymcfg_lNbrk
 */
__attribute__ ((hot)) /* 2015-06-10 icore2 old 19+21 = 21 clks*/
const tbase minsymcfg_lNbrk(const tbase ib0, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  /* L1-caching? NN*MaxSym=40*410 */
  int j,i2,n2; register int i/*,d*/; tphase s1,s2,s3;
  tbase ib1=ib0, ib2; tbase ib3, ib4=TBASE0;
  /* set default minimum, better set by caller? also reduces arguments */
  b->sign=1;  b->scfg=ib1; b->norm2=1; n2=1;
  ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ 
    b->scfg=               ib1= minsymcfg_lm(ib1,&(b->sign));
    if (nu==nd && sym.wud) ib2= minsymcfg_lm(ib2,&s2);
    if (s2==0 || b->sign==0){ s2=b->sign=0; /* 2015-09: return TBASE0;*/ }
    else b->norm2 *= b_factor_lm(ib1); /* ToDo: fermionic sign? */ }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
      n2=2; if(s2!=1){ s2=b->sign=0; /* 2015-09: return TBASE0;*/}	/* s1!=1 */ }
    if(!(ib1<=ib2)){ 
      ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1;} /* new minimum */
  }
  // NN faster code as nn ?
//  for(i=0;i<NN;i++){ ks[i]=k1[i]=geti(ib1,i); k2[i]=geti(ib2,i); }
  for(i2=0,s3=b->sign,ib3=ib1;;i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp(j)=Permutation */
      /* better call/inline symcfg_loopN ? */
      ib4=TBASE0;
      /* ToDo: use NN and unroll tests for even, %4 %8 etc. */
      /* can be unrolled (with lot of jumps to loop-end), speculative CPU,
         but bad to vectorize (possible? lossy?
          jump out if all elements meet the jump-out-criteria) */
      i=nn; do { i--;                 /* Tj, high -> low */
        /* cache miss? -> no, its accessed ca. 700 times per call, 14s */
        /* 32+8: J2=.5 0.10205747 0.25766905 53s,5.5m -> 49s,4.9m */
        /* 32+8: 700*hnz=4.4*sym*hnz max=N*sym*hnz */
        /* unroll possible, generates loop-core and conditional-jumps */
        /* vectorization may be implementable, loop will break a bit later */
#if 0 && NN % 4 == 0 /* 2014-05-14, failed NN=32 tU nn=3 and NN>128 ? */
//#if 0
        i-=1; /* compiler opt: (a|b) | (c|d) = 2 steps, 1 parallel */
        ib4 =                             (ib4 << (2*CM(1,2,2)))
            |  (CM(1,3,3)&(int)(ib3>>(sym.i[j][i  ]<<CM(0,1,1))))
            | ((CM(1,3,3)&(int)(ib3>>(sym.i[j][i+1]<<CM(0,1,1))))<<CM(1,2,2))
          //  | (CM(1,3,3)&(int)((ib3>>(sym.i[j][i+2]<<CM(0,1,1))))<<CM(2,4,4))
          //  | (CM(1,3,3)&(int)((ib3>>(sym.i[j][i+3]<<CM(0,1,1))))<<CM(3,6,6))
        ;
#else   /* 2(ud)*4(NN) unroll+jmpin+out, no vectors */
        ib4 =                             (ib4 << CM(1,2,2))
            | (CM(1,3,3)&(int)(ib3>>(sym.i[j][i]<<CM(0,1,1))));
#endif
        /* d= (ib1>>i)-ib4; */  /* compare cfg with SymOp*cfg */
        /* [j][i]-access vs. pointer good for x86, bad for sparc? */
        /* check if cfgs are still equal after permutation of site i */
        /* ToDo which is faster? test ib4 vs. ib1 or not? */
      } while (/*d==0*/(ib1>>(i<<CM(0,1,1)))==ib4 && i!=0); /* true: 34+6=83% 32+8=78% */
      /* speedup: greater cfg, 157*hnz calls, true(32+8)=99.2% */
      if ((ib1>>(i<<CM(0,1,1)))<(ib4)/*d< 0*/) continue;
#if     CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[ j ]; 
#elif   CM(0,1,1)  /* signum is cfg-dependend */
      /* permute sites and get sign, 2014-05-23 j replaced by inverse! */
      symcfg_loopN(ib3,sym.inverse[j],&ib4,&s1,1);
#endif
      if (ib1!=ib4/*d!=0*/) {  /* 32+8 J2=.5: 1.2*hnz calls, true: 99.7% smaller cfg */
#if CM(1,0,0) /* store rest config */
        if (i) do { i--; ib4 = (ib4 << CM(1,2,2))
           | (CM(1,3,3)&(int)(ib3>>(sym.i[j][i]<<CM(0,1,1)))); } while (i!=0);
#endif
#if CM(0,0,0)
        if (s1*s3==1) /* 2014-05-22 not if bad sign! ToDo does not work */ 
#endif
        { ib1=ib4; n2=1; b->sign=s1*s3; continue; } /* found smaller cfg */
      }
      /* get same config: S|k> to s|k> */
      /* sign=0? 32+8 true=0..3% sym_k-depending, better use k(2PIi/l)? */
      if(!phase_eq(s1*s3,b->sign)){s2=b->sign=0;break;/* 2015-09: remove return TBASE0; */}
      n2++;
    }   /* for(j=0;j<sym.numall;j++) test next sym */
    /* 2015-12-31 add sign-break, fixes tJ_lc problem */
    if(nu!=nd || i2>=1 || (!sym.wud)) break; /* Sep00, Dec15 add sign */
    IFww( break );
  }     /* test ud-sym */
  b->scfg=ib1;
  //b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scgfg2idx(b->scfg)]; */
  return b->scfg;
}
/* 2015-04-10 like minsymcfg_lNbrk, but operating on block of data for performance
 *  ToDo: not complete, which data format? one pointer, 
 * inp: cfg_ib1[n] out: {scfg=acj,sign=asj,norm2=an2}[n]
 */
__attribute__ ((hot))
void minsymcfg_blk_lNbrk(thbuf *hbuf){ /* about 63% CPU -O1 10+10 */
  /* L1-caching? NN*MaxSym=40*410 */
 tbase ib1, ib2;  tbase ib3, ib4=TBASE0; struct sbase tb, *b;
 int ii, j,i2,n2; int i/*,d*/; tphase s1,s2,s3;
 b=&tb; /* for minimum source changes against old code, test code */
 for (ii=0;ii<hbuf->n;ii++) {  
  ib1=hbuf->acj[ii];
  /* start old code minsymcfg_blk_lNbrk(ib1, b); */
  b->sign=1; b->scfg=ib1; b->norm2=1; n2=1; ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ b->scfg=    ib1= minsymcfg_lm(ib1,&(b->sign));
           if (nu==nd && sym.wud) ib2= minsymcfg_lm(ib2,&s2);
           if (s2==0 || b->sign==0){ b->sign=s2=0; }
           else b->norm2 *= b_factor_lm(ib1); /* fermionic sign? */ }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
      n2=2; if(s2!=1) b->sign=s2=0; /* s1!=1 */ }
    if(!(ib1<=ib2)){ /* 2015-01-04 old: s2=1 failes for tJ, tU */
      ib3=ib1;b->scfg=ib1=ib2;ib2=ib3;s1=b->sign;b->sign=s2;s2=s1; } /* new minimum */
  }
  // NN faster code as nn ?  Dec15+b->sign(tJ_lc_s1!)-loop-break
//  for(i=0;i<NN;i++){ ks[i]=k1[i]=geti(ib1,i); k2[i]=geti(ib2,i); }
  for(i2=0,s3=b->sign,ib3=ib1; ;i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
//if (ib3==0x29) printf(" ib3=%s\n",ccfg(ib3));
    for(j=0;j<sym.numall;j++){		/* take a SymOp(j)=Permutation */
      /* better call/inline symcfg_loopN ? */
      ib4=TBASE0;
      /* ToDo: use NN and unroll tests for even, %4 %8 etc. */
      i=nn; do { i--;                 /* Tj, high -> low */
        /* cache miss? -> no, its accessed ca. 700 times per call, 14s */
        /* sym.i should be and stay in L1 cache here, hbuf->acj is stream */
        /* 32+8: J2=.5 0.10205747 0.25766905 53s,5.5m -> 49s,4.9m */
        /* 32+8: 700*hnz=4.4*sym*hnz max=N*sym*hnz */
#if   0 && NN % 4 == 0 /* 2014-05-14, 2016-01 failed for tU NN=32 and nn=3 */
        i-=1; /* compiler opt: (a|b) | (c|d) = 2 steps, 1 parallel */
        ib4 = (ib4 << (2*CM(1,2,2)))
            |  (tbase)(CM(1,3,3)&(int)(ib3>>(sym.i[j][i  ]<<CM(0,1,1))))
            | (tbase)((CM(1,3,3)&(int)(ib3>>(sym.i[j][i+1]<<CM(0,1,1))))<<CM(1,2,2))
        ;
#else
        ib4 = (ib4 << CM(1,2,2))
//          | (tbase)(CM(1,3,3)&(int)(ib3>>(sym.i[j][i]<<CM(0,1,1))));
            | ((ib3>>(sym.i[j][i]<<CM(0,1,1)))&(tbase)CM(1,3,3));
#endif
        /* [j][i]-access vs. pointer good for x86, bad for sparc? */
        /* check if cfgs are still equal after permutation of site i */
      } while ((ib1>>(i<<CM(0,1,1)))==ib4 && i!=0); /* true: 34+6=83% 32+8=78% */
      /* speedup: greater cfg, 157*hnz calls, true(32+8)=99.2% */
      if ((ib1>>(i<<CM(0,1,1)))<(ib4)) continue;
#if     CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[ j ];
#else              /* signum is cfg-dependend, 2014-05-23 inverse[j]! */
      s1=1;
      symcfg_loopN(ib3,sym.inverse[j],&ib4,&s1,1); /* s1=ferm_sign*sym.w[j] */
      // s1=s1*sym.w[sym.inverse[j]]; /* added 2015-12-31 tJ_lc_s1 bug, loop(...,1 to 0) */
#if 0 /* 2015-12-31 fix tJ_lc8_s1 */
      if (ib3==0x29){
        printf("\nDBG loopN(%d=%s,sym%d) min=%s ib4=%s s1  % .1f %+.1fi s0? %d",
           ii,ccfg(ib3),j,ccfg(ib1),ccfg(ib4),Re(s1),Im(s1), (!phase_eq(s1*s3,b->sign)));
        printf("\nDBG loopN(%d=%s,sym%d) min=%s ib4=%s s13 % .1f %+.1fi s0? %d",
           ii,ccfg(ib3),j,ccfg(ib1),ccfg(ib4),Re(s1*s3),Im(s1*s3), (!phase_eq(s1*s3,b->sign)));
        printf("\nDBG loopN(%d=%s,sym%d) min=%s ib4=%s sgn % .1f %+.1fi s0? %d\n",
           ii,ccfg(ib3),j,ccfg(ib1),ccfg(ib4),Re(b->sign),Im(b->sign), 
           Norm(s1*s3-b->sign)>1e-4);
      }
#endif
#endif
      if (ib1!=ib4) {  /* 32+8 J2=.5: 1.2*hnz calls, true: 99.7% smaller cfg */
#if CM(1,0,0) /* store rest config */
        if (i) do { i--; ib4 = (ib4 << CM(1,2,2))
           | (((tbase)CM(1,3,3))&(ib3>>(sym.i[j][i]<<CM(0,1,1)))); } while (i!=0);
#endif  /* s1= added 2015-12-31 tJ_lc_s1 bug */
        ib1=ib4; n2=1; b->sign=s1*s3; continue; /* found smaller cfg */
      } /* else: get same config: S|k> to s|k> */
      /* sign=0? 32+8 true=0..3% sym_k-depending, better use k(2PIi/l)? */
      if(!phase_eq(s1*s3,b->sign)){s2=b->sign=0;break;}/*2015-04-29-brk,2015-12+brk */
      n2++;
    } /* for(j=0;j<sym.numall;j++) test next sym */
    if(nu!=nd || i2>=1 || (!sym.wud)) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  b->scfg=ib1;
  //b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scgfg2idx(b->scfg)]; */
  /* end old code minsymcfg_blk_lNbrk(ib1, b); */
  hbuf->acj[ii]=ib1;
  hbuf->asj[ii]=b->sign;
  hbuf->an2[ii]=b->norm2;
 } /* for (ii=0;ii<hbuf->n;ii++) */
 return;
}
// double computation to see if we have L1I problems (half or same speed?)
tbase minsymcfg_test3(const tbase ib0, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  /* L1-caching? NN*Sym=40*160=6400 lt. optofile 5% misses */
  int j,i2,n2; register int i/*,d*/; tphase s1,s2,s3;
  tbase ib1=ib0, ib2; tbase ib3, ib4=TBASE0;
  /* set default minimum, better set by caller? also reduces arguments */
  b->sign=1;  b->scfg=ib1; b->norm2=1; n2=1; ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=wud*T|ib1> */
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ b->scfg=    ib1= minsymcfg_lm(ib1,&(b->sign));
           if (nu==nd && sym.wud) ib2= minsymcfg_lm(ib2,&s2);
           if (s2==0 || b->sign==0){ s2=b->sign=0; return TBASE0; }
           else b->norm2 *= b_factor_lm(ib1); /* fermionic sign? */ }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
      n2=2; if(s2!=1){ s2=b->sign=0; return TBASE0;} /* s1!=1 */ }
    if(!(ib1<=ib2)){ 
      ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; } /* new minimum */
  }
  // NN faster code as nn ?
//  for(i=0;i<NN;i++){ ks[i]=k1[i]=geti(ib1,i); k2[i]=geti(ib2,i); }
  for(i2=0,s3=b->sign,ib3=ib1;;i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp(j)=Permutation */
      /* better call/inline symcfg_loopN ? */
      ib4=TBASE0;
      i=NN /*nn*/; do { i--;                 /* Tj, high -> low */
        /* cache miss? -> no, its accessed ca. 700 times per call, 14s */
        /* 32+8: J2=.5 0.10205747 0.25766905 53s,5.5m -> 49s,4.9m */
        /* 32+8: 700*hnz=4.4*sym*hnz max=N*sym*hnz */
        ib4 = (ib4 << CM(1,2,2))
            | ((ib3>>(sym.i[j][i]<<CM(0,1,1)))&(tbase)CM(1,3,3));
        /* d= (ib1>>i)-ib4; */  /* compare cfg with SymOp*cfg */
        /* [j][i]-access vs. pointer good for x86, bad for sparc? */
        /* check if cfgs are still equal after permutation of site i */
      } while (/*d==0*/(ib1>>(i<<CM(0,1,1)))==ib4 && i!=0); /* true: 34+6=83% 32+8=78% */
      if ((ib1>>(i<<CM(0,1,1)))<(ib4)/*d< 0*/) continue; /* greater cfg, 157*hnz calls, true(32+8)=99.2% */
#if     CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[ j ]; 
#elif   CM(0,1,1)  /* signum is cfg-dependend */
      /* permute sites and get sign, 2014-05-22 j replaced by inverse */
      symcfg_loopN(ib3,sym.inverse[j],&ib4,&s1,1);
#endif
      if (ib1!=ib4/*d!=0*/) {  /* 32+8 J2=.5: 1.2*hnz calls, true: 99.7% smaller cfg */
#if CM(1,0,0) /* store rest config */
        if (i) do { i--; ib4 = (ib4 << CM(1,2,2))
          | (CM(1,3,3)&(int)(ib3>>(sym.i[j][i]<<CM(0,1,1)))); } while (i!=0);
#endif
        ib1=ib4; n2=1; b->sign=s1*s3; continue; /* found smaller cfg */
      }
      /* get same config: S|k> to s|k> */
      /* sign=0? 32+8 true=0..3% sym_k-depending, better use k(2PIi/l)? */
      if(!phase_eq(s1*s3,b->sign)){s2=b->sign=0;return TBASE0;}
      n2++;
    }   /* test next sym */
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  b->scfg=ib1;
  //b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scgfg2idx(b->scfg)]; */
  return b->scfg;
}

// next_minsymcfg: start highest (all bits 1) set 0 starting from MSB(=bitN-1)
// next_maxsymcfg: start lowest  (all bits 0) set 1 starting from LSB(=bit0) ?
//                                                       ... from MSB(bitN-1)!
// ToDo: check on 6+4 s1-chain
// next_maxsymcfg: start highest (all bits 1) set 0 starting from LSB(=bit0) 
//    stop or go deeper if found
//   
//   not easy to find, so we try to find a good upper bound
//   scan full subspace nu=4 and than walk only nu+1-max-paths
//  ToDo: get_maxscfg parallel ab nu-4 ?
//  ToDo: stepwise nu+1 statt rekursiv (parallel)
//        umbenennen in lastsymcfg ?
tbase get_maxscfg(){  /* get max(last) sym config to speedup scfg search */
  // ULong i2=0;
  int ii;
  tbase maxscfg= 0;
  ULong nt=0l, nt0=0l, nt1=0l;   /* number tested cfgs for speed analyse */
  if (vvv && !mpi_i) { printf(" get_maxscfg vvv=%d\n", vvv); }

  extern ULong x_max;
  struct sbase b;   /* working buffer (big) */
  int tnu=nu, tnd=nd; /* store the original nu,nd and restore later */
  tbase state[2*NN+1]; /* recursion buffer (max nu+nd) */
  int id[2*NN+1], iu[2*NN+1]; /* number of bits set */
  int rd=0; /* recursion depth */
  int rs;   /* return isminsymcfg */
//  int hus[2*NN+1]; /* highest unset site(bit), goes from high to low */
  int lus[2*NN+1]; /* lowest  unset site(bit), goes from low to high ?? */
  int nbp[2*NN+1]; /* next bitpattern, 3=u+d 2=d 1=u 0=. */
  // long t2, t1=itime();
  //b.o=2; /* ??? */

  /* the initial state is all sites with maximum bits set (000...000) */
  state[0]=TBASE0;  // maxscfg
  for(ii=0;ii<nn;ii++) state[0]=puti(state[0],ii,CM(2,2,3)); // 2=d minscfg
  // now we have maxcfg CM(0xff=dddddddd,0xaaaa=dddddddd,0xffff=33333333)

  // hus[0]=nn;                  // start from MSB+1=nn go down to 0
  nbp[0]=CM(0,-1,-1);         // nbp+1=CM(1=u,0=.,0=.)  goes up ++ ?
  nbp[0]=CM(2,2,3);         // nbp+1=CM(21=du,210=du.,3210=3du.)  goes down
  id[0]=nn; iu[0]=CM(0,0,nn); // ini_maxcfg
  lus[1]=lus[0]=0;            // start from LSB=0 go up to MSB=nn-1
#if CM(1,0,0)
  lus[1]=lus[0]=nu-1;  // speedup 2012-06-24, enough space for rest 0s on LSBs
#elif CM(0,0,1)
  lus[1]=lus[0]=min(nn-nu-1,nn-nd-1); // enough space for rest zeros
#endif
  
  if ((vvv&4) && !mpi_i)
    printf(" start     %s rd %2d lus %2d\n",
           ccfg(state[rd]), rd, lus[rd]);
  /* now reduce  bits from lowest bits, test if smallest and recursive */
  /* ToDo: hus and nn-rd is the same! remove hus[] */
  if (CM(nu==0, nu==0 && nd==nn, nu==nn && nd==nn)) {
    if (!mpi_i)printf("DBG: this trivial case of nu, nd is set extra\n");
    /* ToDo: use fastest _dflt() */
    rs=isminsymcfg_lNbrk(state[rd],&b); nt++; /* rs: 1=true 0=false -1=destr_sign */
#ifdef CONFIG_ANSATZ
    if (!ansatz_test(state[rd])) { rs=0; }; // ToDo: untested!
#endif
    if (rs==1) maxscfg= state[rd];
  } else // trivial case
  for(;;) { /* can we do that for a thread specifique range? */
    // break if nu,nd reached!?
    // next bit pattern on current site
    //  old: JJ: [0] 1 1 ...  tJ: [-1] 0 1 0 1 ... tU: [-1] 0 1 2 0 1 2 0 1 2
    //  min: JJ: [2] 1 1 ...  tJ:  [2] 1 0 1 0 ... tU:  [3] 2 1 0 2 1 0 2 1 0
    nbp[rd]--;    /* next bitpattern, 3=u+d 2=d 1=u 0=. at site lus[rd+1] */
    if (nbp[rd]<CM(1,0,0)) {
//    lus[rd]--; /* highest unset site goes down to next lower site */
      lus[rd+1]++; /* highest unset site goes up to next higher site */
      nbp[rd]=CM(2,2,3)-1; /* next bit pattern at site lus[rd+1] */
    }
    // if (geti(state[rd],lus[rd+1])==nbp[rd]) continue; // no double set
    if (lus[rd+1]>=nn || (rd>0 && lus[rd+1]>=lus[rd]) )
    { /* now we have tried all sites, so we can return */
      if (rd==0) break; /* all work done */
      /* not a max, resume search at rd-1 */
      // if (rd>FullDepth) rd=FullDepth; else 
      rd--;     /* go up stack */
      if (nt<x_max && (vvv&5)==5 && !mpi_i)
        printf(" come back %s rd %2d lus %2d %2d+%2d %5.0e\n",
               ccfg(state[rd+1]), rd+1, lus[rd+1], iu[rd+1], id[rd+1],1.*nt);
      continue; /* return and try next deeper state */
    }
#ifdef CONFIG_S1SYM
#endif    
    /* probewise reset bit at hus/lus */
//    nbp[rd+1]=CM(0,-1,-1);
    nbp[rd+1]=CM(2,2,3)-1; // start with overflow nn to nn-1
    /* set nu, nd to have correct working behaviour of minsymcfg */
    id[rd+1]=id[rd]-1+(nbp[rd]>>1); /* 1: nd--  0: nd-- */
    iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(1,1,0):CM(0,0,-1));  /* 1: nu++  0: nu-- */
//    id[rd+1]=id[rd]+(nbp[rd]>>1); /* JJ: 2: nd++  */
//    iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(0,1,1):CM(-1,0,0));  /* */
    if (lus[rd+1]<0) continue; // tJ, tU only
    state[rd+1]=puti(state[rd],lus[rd+1],nbp[rd]); /* lower bit pattern, down -> up */
#if CM(0,1,1)
    // dont go a wrong path 2012-07-10
    if (iu[rd+1]<tnu || id[rd+1]<tnd) continue;
#endif

    // prepare for going deeper (before rd++! 2013-02-07)
    lus[rd+2]=0; /* start from lowest site when going deeper */
#if CM(1,0,0)  // 2012-06-24 speedup for nu>>nd 39+15 13s, 19+35 1s
    // 14+2: 0x7fff 14-1-3 ??? 
    lus[rd+2]=tnu-iu[rd+1]-2; // (much) faster if starting on higher bits
#elif CM(0,0,1)
    lus[rd+2]=min(iu[rd+1]-tnu,id[rd+1]-tnd)-1; // -3 ??? 0..nn
    // printf(" rd+2= %d  lus= %d\n",rd+2,lus[rd+2]);
#endif

#if 1
    // 39+1 speedup, 37+17 speedup 21s 3880724 of 10988161 (old 37s)
    // if there is only one choice, skip all pre-cfgs and go to final cfg
    CM(\
    if (lus[rd+1]>0 && tnu-iu[rd+1]==lus[rd+1]){rd++;continue;},{},{})
/*
    CM(\
    if (lus[rd+1]>0 && id[rd+1]-tnd==lus[rd+1] \
                   ) { rd++; continue; },{},{})
*/
#endif

    /* ToDo: filter thread specifique range here? */
    nu=0; nd=nn; /* trick to use sym_ud to early */
    if (iu[rd+1]==tnu && id[rd+1]==tnd) {nu=tnu; nd=tnd;} /* sym.ud=on */
    rs=isminsymcfg_lNbrk(state[rd+1],&b); /* 1=true 0=false -1=destr_sign */
    nt++; /* count isminsymcfg calls for speed output */
    if (((rs && iu[rd+1]==tnu && id[rd+1]==tnd && (nt0<x_max || !(nt0&(nt0-1))))
       || nt<x_max || !(nt&(nt-1))) && (vvv&5)==5 && !mpi_i) // || hus[rd+1]!=nn-rd-1)
      printf(" mtry % d   %s rd %2d lus %2d %2d+%2d %5.0e\n",
        rs, ccfg(state[rd+1]),rd+1, lus[rd+1],iu[rd+1],id[rd+1], 1.*nt);
    if (rs) { /* go deeper or count for 1=true and -1=destructiv_sign */
      if (iu[rd+1]==tnu && id[rd+1]==tnd) { /* found symmmetric state */
#ifdef CONFIG_ANSATZ
        if (!ansatz_test(state[rd+1])) { rs=0; } // untested
#endif
        nt0++; // count tests for reduced debug output
        if (rs==1) {  /* without destructive phase */
          // tbase old= maxscfg;
          // 2015-10 ri hits 1e9 on 18+18 chain, set to 2e9
          if (maxscfg < state[rd+1]) maxscfg= state[rd+1];
          if (nt>2e9) E1("maxscfg stopped after 2e9 tests (mpi=%d)",mpi_i);
          if (rd==0 || nt>2e9) break;
          break; // 2012-06-24 fixed version
          // break; /* ToDo: return to next subtree tnu=4 */
        } /* rs==1 found smallest, nondestructive sign */
      } else { /* found symmetric mother cfg, go deeper */
        nt1++;
        rd++; /* go deeper, ToDo: start by putting rest up's to next sites? */
        if (rd>2*NN) { ERR(9999); } continue;
      }
    } /* smallest cfg */
  } /* for(;;) */
  nu=tnu; nd=tnd; /* restore gobal nu, nd */

  if (!mpi_i) {
    printf("  end.maxscfg nt0= %lu nt1= %lu nt= %lu t+= %s\n",
      nt0, nt1, nt, sclock(~0l));
    printf("  maxscfg= %s\n", ccfg(maxscfg));
  }
  if (maxscfg==TBASE0 && !mpi_i) E0("bad maxscfg code?, ignore this?"); 
  return(maxscfg);
}


#define BLOCKS	8	/* can be only 2^n=1,2,4,8,etc */
/* version 0.1 Feb 2000 first try version, not finaly, for checks 
 * ideal way for part l1 into blocks => multi processing
 * in ideal case fast and invariant against symmetry operations 
 * min=0 max=n*(n-1)/2 ??? besser mit Untergittern
 */
int block_no(tbase ib1){ /* use pair correlations, count d-d pairs */
  int no,i1,i2; no=0;
  for(i1=0;i1<nn;i1++) if (zi(ib1,i1)==-1) /* nn+nd*nn Ops */
  for(i2=0;i2<nn;i2++) if (zi(ib1,i2)==+1)
   no+=sym.corr[i2][i1];
  return (no>>1)&(BLOCKS-1);
} 

/* --- all what be needed for pthreads of numsonf --- */
typedef struct sp_ns {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  int block;       /* block nr (0..threads*nodes-1) for debugging */
  ULong i1,n;      /* idx range of configuration to check (rw,r) [i1...n-1] */
                   /* ToDo: on return i1 is i1(nothing done)...n(all done) */
                   /* ToDo: return time too, for dynamic balance */
                   /* ToDo: timeout in seconds dflt=60 */ 
  ULong numans;    /* num_ansatz (w) symcfgs filtered out by ansatz_test() */
  Tnorm2 norm2;     /* check sum (w) */
  ULong num;       /* number of stored sym configs (w) */
  ULong num0;      /* number of sym configs with sign=0 (w) v2.27 */
  tbase ib1;        /* spin-configuration (t) */
  struct sbase b;   /* working buffer (t,big) */
  struct sp_ns *next; /* pointer to next data set */
  // char fname[32];     /* filename tmp_shared/l1_%d%4d.tmp (r) Oct07,Jan12 */
  unsigned char *buf; /* space for results (sym-configs) (r[w]) 2012-04 */
  ULong buflen;       /* size of buffer in units of Bsize (r) 2012-04 */
  int time0;          /* computing time in seconds to calculate balance */
} tp_ns;

/* void-function for pthreads called by numsymconf(), see below */
#define L1_BUFLEN (1*256*1024)  /* in units of Bsize(5-8B) for 1-10write/s */
/* reduce write calls for performance reasons (?), but 8M is to big */
/* change 2014-07-16 8M*Bsize zu gross fuer low-mem-tests, SHsize=4M */
void ns_thread( void *ptr ){
    tp_ns *p=(tp_ns *)ptr;
    tbase ib0;      /* min config (overflowtest) */
    tbase ib_stop;    /* in case of s=1 next skips LM-cfgs 2012-01 */
    ULong i1 /*, i2, i3*/;

    long t2,t1=itime();  // ToDo: syscalls only if ptr->tflag=1 
    ULong i_out=1l<<14; /* counter for output, speed estimation */
    int rt;              /* return code from isminsymcfg_lNbrk */
    /* 2012-01 p->n may be to big, in case LM-cfgs are skipped by next() */
    /* ll1(n) may cause trouble on s=1 and N=64 bit systems */
    ib_stop = ll1(p->n /* - 1*/); /* overrun nosym_n1? 2012-03: ll1(n1)==0 bad!? */
//    ib_stop = next(ib_stop); /* trick to handle overflow correct */
    p->ib1 = ll1(i1= p->i1); /* start cfg */
//  if (i1) { p->ib1 = next( ll1(i1-1) ); } /* trick to get correct s1sym */
// ToDo: 9s1=3s3 next(aaaaaaaaabbbbbbbbb
//                   =aaabbbcccdddeeefff
//                   =100000000011111111) may fail (noS1sym)! N=2*9 2012-05-16
    ib0 = ll1(0);
    if (!ib_stop) ib_stop=ib0;
    // 2013-03-28 add threadsafeness and fix bug S=2++
//    if (ib_stop<p->ib1) F0("stop<start");
    /* but its better to do this outside run, because s=1 overlap problem */
#if 0
    // if (mpi_i<=8) 
    {
      usleep(mpi_i*10000);
      printf("DBG mpi %d %s - %s i %9lu - %9lu\n",
        mpi_i, ccfg(p->ib1), ccfg(ib_stop),
               (unsigned long)(p->i1), (unsigned long)(p->n));
    }
#endif
    if    (i1 < p->n) // JS20120306 minimum condition
    for( ; i1 < p->n ; ){
      rt=isminsymcfg_lNbrk(p->ib1,&(p->b)); /* 1=true, 0=false -1=sign0 */
#ifdef CONFIG_ANSATZ
      if (!ansatz_test(p->ib1)) { p->numans++; rt=0; }
#endif
//printf("DBG i1= %6.0f %s = %d\n", 1.*i1, ccfg(p->ib1), rt);
      if (1==rt) { /* 1=true, 0,-1=false */
        if (p->num >= p->buflen) break;
        putl1(/*i2*/p->num,p->ib1,p->buf);   /* store mincfg to l1p[l1s] */
        p->num++; /* i2++; */            /* count stored cfgs */
        p->norm2+=p->b.norm2;
      }
      if (-1==rt) { p->num0++; p->norm2+=p->b.norm2; }
      i1++;                     /* loop-counter */
      // if (p->num>=0) printf("DBG1: %3d %s\n",(int)p->num,ccfg(p->ib1));
      p->ib1=next(p->ib1);      /* next higher config (skip LM-cfgs!?) */
      // if (p->num>=0) printf("DBG2: %3d %s\n",(int)p->num,ccfg(p->ib1));
//printf("DBG i1= %6.0f %s stop=%s ib0=%s\n",
//   1.*i1, ccfg(p->ib1), ccfg(ib_stop), ccfg(ib0));
      /* for 2s=2,3,4... p->n is bigger than the number of next calls */
      /* next() skips LM configs, so the of= value and 100% estim. fails */
      /*  may be next works to &(p->ib1) and returns skipped numbers
       *    also getting index difference between ib1 and ib2 would be nice 
       * next may jump over ib_stop for s=1 (usually s=1/2 cant jump)
       */
      if (p->ib1 <= ib0) break;     /* avoid overrun? */
      if (p->ib1 >= ib_stop
       && ib_stop != ib0) break; /* part of next thread (s=1,3/2...) */ 
      if (i1==i_out && mpi_i==0 && p->block==0) { long t2=itime();
        i_out<<=2; /* x4 */
        if (t2-t1>8)
        /* i1 is incorrect for LM, ll2 can be used for correct estimations */ 
        printf(" ns %6.2f%% found= %9.0f of %9.0f t= %5.2fm t100%%= %6.2fm\n",
           (i1 - p->i1)*100./(p->n - p->i1),  // % of work?
             1.*p->num,                     // found sl1
             1.*i1, /* i1 < p->n - p->i1 for s=1++ */
              (t2-t1)/60.,
           // (t2-t1)/60.*p->n/i1);
              (t2-t1)/60.*(p->n - p->i1)/(i1 - p->i1)); // JS20120306
      }
    }
    /* this was buggy before 2012-01 and therefore disabled for s=1 pt>1 */
    /* ToDo: output only min/mean/max + time in the main thread later */
    /* MPI: "\n ... " may buffer ... until end (SC), better " ... \n" */
    t2=itime();
    p->time0 = t2-t1;
#if 0 // Mar2013 for debugging only, set back to 0 after
    if ((mpi_i*pt_n+p->block%pt_n<2 /* 2012-05 output first and last two */
      || mpi_i*pt_n+p->block%pt_n>mpi_n*pt_n-3) &&
        (p->num>0 || t2-t1>10) && (vvv&4)) 
                /* but not the less interesting things */
      printf(" ns_t %2d %2d found= %9.4g of %9.4g of %9.4g t=%s\n",
        mpi_i, p->block%pt_n, 1.*p->num, 1.*i1, 1.*(p->n - p->i1),
        // p->i1*1., 1.*p->n, /* for debugging s=1 from-to */
        sclock(t2-t1)); /* block=0..mpi_n*B_NUM-1 */
#endif
    /* i3=fwrite(buf,Bsize,i2,fl1); if(i3!=i2)F1("fwrite %s",p->fname);
     fclose(fl1); */ /* flush */
    /* free(buf); 2012-04 */
}  // ns_thread

/* number of (symmetric) base-states(NN,nu,nd) -> n1_using_Sym 
 *  global vars: sym,nu,nd!
 * until v2.19:
 *   dividing in equal blocks for threads is unbalanced (2/3 only one cpu)
 *   because 1st thread has more work as the others (3 times more)
 * ToDo: better thread-balance
 *   1) create nt=4 threads, set results=0
 *      pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
 *      better pthread_mutex_init(&mutex1,NULL)
 *   2) pthread_mutex_lock (only one thread can pass)  (OMP_critical)
 *   3) store results, reset results, get new work=block(ns/(16*nt)) (serialized)
 *   4) pthread_mutex_unlock
 *   5) does the job and return to (2) -or- leave if work=0
 *   ... (MPI_Barrier, OMP_barrier)
 *   6) destroy threads
 * exsample: jj square 160sym nn=36+4 sym_k=0 ns(91390)=632=523+85+22+1+1
 * ToDo: try IsingGS + non-sym-excitations until IsingMax (how to divide?)
 * ToDo: dont use disk (MPI+NFS is slow and instable)
 *    better use MPI-disk emulation OR parallel CPU and save local part
 *    store buffered blocks to node (i++%mpi_n)?
 * ToDo: transfer to local disks of cluster nodes
 *       get final_n1=sum(local_n1)
 *       + send local_n1 (0 to n0, ... local_n1 to n3, 0 to rest
 *       receive from n0,n1,n2 ... until final_n1/tasks reached
 * ToDo: estimation of n1? upper boundary?
 *        MC? 100*blocks random symmetrized startvec
 *            divided into blocks + storeh2?
 *            hash = ising+meanerg_of_excitations
 *  (minimize off_diag_block_elements for maximum performance)
 *   
 */
// ToDo: new pipelined recursion
// l1_0(u=0) -> task0(u=1) -> task1(u=2) -> task2+3(u=3) -> task4-7(u=4) -> l1_4(u=4)
//   repeat recursion using l1_4 (einfacherer [paralleler] code)
//  ToDo: chpt, read last l1_0000.dat + test if nu+nd+symcfg = OK
//     + continue from last symcfg+1
// ToDo2016: add args nn,nu,nd,sym,l1?
ULong numsymconf(){  /* short: ns() */
  ULong i2=0, i3, numans=0, num0=0;
  Tnorm2 sum=0;
  tp_ns pns, *ppns, *p2;
  tbase maxscfg0,   // ToDo: upper boundary (is: maybe max)
        startcfg0,  // startcfg, resumed after checkpointing, default=ll1(0);
        blast;      // checkpointing, last written valid scfg
  int ii, i4, nsalgo=0;
  FILE *fl1=NULL;   /* tmp_shared/l1_0000-file */
  long l1=itime();
extern int algo;
  p2=NULL ; /* dummy access, avoiding cc warnings */
  ppns=&pns; ppns->buflen=0; ppns->buf=NULL;
  i3=numconf(nn,nu,nd); 
  if (!i3) E1("numsymconf numconf-OVL = %.1e", slow_dbl_np(nn,nu,nd));
  // ToDo2016 handle slow_np on OVL
  i4=pt_n*mpi_n; /* i3=(N!)(nu!nd!) i4=num_threads */

#if CM(1,0,0)
  maxscfg0= get_maxscfg(); /* 2012-05-24 for scan cfg speedup */
#else
  maxscfg0= 0; // get_maxscfg(); /* fails OR runs very long if sym= 0 */
  if (!mpi_i) printf("\nWARNING: maxscfg is badly slow for 0+2 tU + sym (ToDo) !!!\n\n");
#endif

  if (checkpoint>=2) { //  try reading n1 from chkpt2
    FILE *f1; n1=0;
    if (!mpi_i) { 
      if (vvv) printf(" chkpt2.r"); // 2016-10 add macro spacing for gcc-6.2
      f1=fopen(FCHKPT "2","r"); if(!f1){E0("chkpt2.o");}
      if (f1 && fscanf(f1,"%lu",&n1)!=1) {E0("chkpt2.r");n1=0;} // 2012-05
      if (f1) fclose(f1);
      /* new warning, for break + edit sym_k + restart without rm chkpt* */
      printf("\nWARNING: checkpoint loaded, this may be unintended\n"
             " tmp_shared/* should be deleted after modification of data set\n");
    } // if (!mpi_i) ... ToDo: test last symconfig!?
    MPI_Bcast(&n1,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);
    if (n1==0) checkpoint=1; // failed, restart calculation
    // ToDo check correct l1.dat (below?)
  } // if (checkpoint>=2)
  // 2012-06-06 read last state
  blast=0;
  if(!mpi_i){ // check point resume? read l1.end and progress
    tbase b0=TBASE0;  // must be initialized with 0, see fread(..,Bsize,...) ??? 2012-06
    off_t fl2o; int fl2, rt=0,i,tu,td; struct sbase b; // long t1;
    //b.o=2; /* ? */
    fl2=open("tmp_shared/l1_0000.dat",O_RDWR); // seq. written 
    if (fl2!=-1) {
      // EOF-Bsize may fail if parts of Bsize=f(NN) has changed! 2012-06-19 len%8=6 %6=4
// * fseek(long)  32bit or 64bit (v2.32 and less)
// * lseek(off_t) 32bit or 64bit (also 64bit for 32bit + LFS) v2.33
      fl2o=lseek(fl2,(off_t)0,SEEK_END);
      if (fl2o==-1){ E0("lseek");
      } else { //ok 
        if (fl2o%Bsize) {
          printf(" WARNING: l1-file size %% Bsize%d = %d"
                 " trying to cut ...\n", Bsize, (int)(fl2o%Bsize));
          if (ftruncate(fl2,fl2o-(fl2o%Bsize))!=0) E0("truncate");
          fl2o-=fl2o%Bsize;
        }
        fl2o=lseek(fl2, (off_t)(-Bsize), SEEK_END)+Bsize;
        //  ToDo: cut to ftell%BSize==0 2012-06  (writing multiple of Bsize=7 + chkpt1)
        if (read(fl2,&b0,Bsize)==Bsize) {  // may read only Bsize=7 of 8 bytes ???
          for (td=i=0;i<nn;i++) td+= 1&(int)(b0>>(CM(i,2*i+1,2*i+1)));
          for (tu=i=0;i<nn;i++) tu+= 1&(int)(b0>>(CM(0,2*i  ,2*i  )));
          CM(tu=nn-td,{},{});
          rt=isminsymcfg_lNbrk(b0,&b); /* 1=true, 0=false -1=sign0 */
          printf(" chkpt12.r_last %s tu+td=%d+%d scfg?=%d n1?=%ld %%%d=%d\n",
           ccfg(b0), tu, td, rt, (long int)(fl2o/Bsize), Bsize, (int)(fl2o%Bsize));
          // ToDo: check if valid scfg rt==1 td==nd,
          //       then start search from it
          //  ToDo: vorsicht nicht zuverlaessig genug!
          if (tu!=nu || td!=nd || rt!=1) {
            if (checkpoint>=1) {
              E0("bad chkpt1 (rename), NN+Bsize changed?"); 
              rename(FCHKPT "1",FCHKPT "1.old"); // next start may succeed
              // MPI_Abort(MPI_COMM_WORLD,1);
            }
            n1=0; checkpoint=0; // ToDo: restart?
          }
          if (checkpoint>=2 && fl2o*1. != n1*1.*Bsize) {
            n1=0 /* fl2o/Bsize */; checkpoint=1;
            printf(" WARNING: bad chkpt2 (rename), NN+Bsize changed? fsize?\n");
            rename(FCHKPT "2",FCHKPT "2.old"); // next start may succeed
            // MPI_Abort(MPI_COMM_WORLD,1); // ToDo: restart?
          } 
          // ToDo: try to load chkpt1a (letzter gepruefte config) hex
          if (checkpoint>=1) blast= b0; // valid
        } // fread ok
      } // fseek ok
      close(fl2); fl2=0;
    }
  }
  MPI_Bcast(&checkpoint,1,MPI_INT,0,MPI_COMM_WORLD);
  if (checkpoint>=2) return (n1); // skip computation
  
  startcfg0= ll1(0);   // 2012-06-11 default start configuration (no syms)
  if (checkpoint>=1) {
    FILE *f1; tbase t0=0; int i1,rt=0,tu,td; char s1[NN+2]; // + LF + \0
    if (vvv && !mpi_i) fputs(" r." FCHKPT "1\n",stdout); // meaning changed (old n1)
    n1= 0;
    if (!mpi_i) {
      f1=fopen(FCHKPT "1","r");
      if (!f1) { printf("\nERROR o." FCHKPT "1\n"); rt=1; }
      // read ooooxoxoxoxxx or ...ooxxQQoxQ + \0
      // ToDo: additional read numstoredcfgs
      if (f1 && !fgets(s1,nn+1,f1)) {E0("r." FCHKPT "1");rt=2;} // 2012-06
      if (rt==0) {
        if (s1[0]=='0' && s1[1]=='x') { /* new hex format 2012-06-27 */
          for (i1=0;i1<(nn+3)/4;i1++)
            { t0<<=4; t0|=s1[2+i1]-((s1[2+i1]<'a')?'0':'a'+10); }
          for (tu=td=0,i1=0;i1<nn;i1++) {
            if (geti(t0,i1)&1) tu++;
            if (geti(t0,i1)&2) td++;
          }
        } else /* old site wise format */
        for (tu=td=0,i1=0;i1<nn;i1++) {
          t0<<=CM(1,2,2);
          if (s1[i1]=='.') {t0|=CM(~0,0,0);          continue; }
          if (s1[i1]=='o') {t0|=CM( 0,1,1);tu++;     continue; }
          if (s1[i1]=='x') {t0|=CM( 1,2,2);     td++;continue; }
          if (s1[i1]=='Q') {t0|=CM(~0,3,3);tu++;td++;continue; }
          E0("chkpt1: bad format (ignore chkpt)"); rt=3; break;
        }
        if ((tu!=nu || td!=nd)) { E0("chkpt1: bad u+d (ignore chkpt)"); rt=4; }
        if ((tu==nu && td==nd) && i1==nn) { 
          startcfg0= t0; // valid startcfg
          printf(" ckpt1cfg= %s dflt_start\n", ccfg(startcfg0));
          printf(" l1endcfg= %s higher?\n", ccfg(blast));
        }
      } // rt==0 no errors
      if (f1) fclose(f1);
      if (rt==0 && blast>startcfg0) {
        startcfg0=next(blast);
        printf(" startcfg= %s (l1end+1)\n", ccfg(startcfg0));
      } // else  checkpoint=0; // needed?, not if chkpt2 + l1 is ok
    } // !mpi_i
  } // chkpt1
  
  MPI_Bcast(&startcfg0,Bsize,MPI_BYTE,0,MPI_COMM_WORLD);
  MPI_Bcast(&checkpoint,1,MPI_INT,0,MPI_COMM_WORLD);
  if (!mpi_i)
    printf(" startcfg= %s chkpt=%d\n", ccfg(startcfg0), checkpoint);
  /* the estimation of n1 failes if there are local symmetries. n1 bigger? */
  /* we now try to choose the faster algorithm automatically JS 2011-10-28 */
  /* nsalgo: 0 = conventional check (parallel), 1 = tree (serial only) */
  /* for single threads switch to ca. 10*faster serial code JS 2011-10-28 */
  /* worst case nosym:  N=40  8+32  120s/threads vs. 127s (Faktor 0.94 !!) */
  /*     for S=1 Factor 4 faster, for S=2 Factor 1000 2013-04-02 */
  /*  ToDo: parallel recursive would the best way! forget old next() */
  if (mpi_n==1 && pt_n==1) nsalgo= 1;  /* serial-tree-like-scfg-search */
  if (sym.num_lm*2 < nn) nsalgo= 1; /* for s=3/2++ speedup 48++ 2013-04-02 */
  /* but user can switch to the other mode by algo+=16 */
  /* old switching to oposite is missleading for debugging, now ... */
  /* 2013-04-24 a0=auto_switch  a16=recursive_ns a32=old_seq_linpack_scan */
  /* if(algo&16)nsalgo^=1; */ /* explicite switch to other algo for tests */
  if (algo&16) nsalgo= 1; /* explicite switch to recursive_ns */
  if (algo&32) nsalgo= 0; /* explicite switch to old_seq_linpack_scan */
  /* 2014-09 tU chain 30 0+2 n1=0 for sym!, n1=190 for sym_k= -1000 */
#ifdef __MIC__
  if(nsalgo==0 && !mpi_i)
    printf("WARNING: a32 failed on XeonPhi-tests 2015-02, a16 is ok\n");
#endif
  if (vvv && !mpi_i) {
    printf(" --- starting hilbert space generation ---\n"
           " filter out symmetric ising base states (n1 .ge. N!/(nu!nd!))/sym\n");
    printf(" algo= %d (0=auto, 16=recursive, 32=old_parallel)\n", algo&48);
    printf(" using algorithm %d (0=scan_bitpermutations, 1=recursive)\n", nsalgo);
    /* JS-Apr2011: mpi*pt=i4 removed (diff) */
    printf("\n numsymconf(%.15g)/(mpi*pt) slow dbl\n", slow_dbl_np(nn,nu,nd));
    printf(" numsymconf(%.15g)/(mpi*pt) out=tmp_shared/l1_0000.dat\n", i3*1.);
    printf(" ncfgs=%.2e nsym=%d estimated nscfgs n1: %.2e\n",  
             i3*1., sym.numall+1, i3*1./(sym.numall+1));
    /* example: chain6=3+3 n1: 6!/(3!3!)/sym=20/(6..12)=0..3 */
  }

  /* MPI_Barrier(MPI_COMM_WORLD); -- may hang on oversubscription 2016-03 */

  /* nsalgo==0: parallel scan next()-loop for scfgs (to slow for S=1++) */
  /* nsalgo==1: (serial) recursive scfg generation */
  if(nsalgo==1 && !mpi_i)  /* new serial recursive method v2.27 */
  { /* ToDo: tJ + tU, nd>nu is much faster, test=algo&16
     * number of isminsymcfg calls dominates compute time (speedup for sym>=2)
     * sym_k != 0 works, ToDo: tU,tJ
     * ToDo: if nu>nd do inversion! tU -> .3 inversion + U -> -U
     *                                for U*(niu-1/2)*(nid-1/2)==Ham2
     *        for nu>nd do inversion before nsymconf and again before storing
     *  see ../doc/numsymcfg.html (explanation+speed+test+examples)
     *   tJ code checked, tU checked (tune+further checks later!)
     * parallel idea 2015-12:
     *   start all tasks + count depth u=3(p.e.), u+1 at task count%num_tasks
     * ...   
     * ToDo2012-05: parallel-recursive: (worst case = 0sym, best case = s=4)?
     *   - step1: 1cfg 1+39 to file0 (mem0!)
     *   - step2: read file0, compute 2+38 parallel to file1 (mem1)
     *   - step3: read file1(2+38) file1, split +compute 3+37 parallel to file0
     *   - etc. (per ping-pong-piping+sorting 2 task-groups of n/2?)
     * ToDo: checkpointing!? + recompute on bad checkpoints
     */
    extern ULong x_max;
    struct sbase b;     /* working buffer  b.{scfg+norm2+sign} */
    int tnu=nu, tnd=nd; /* store the original nu,nd and restore later */
    tbase state[2*NN+1]; /* recursion buffer (tU.max=nu+nd) */
    int id[2*NN+1], iu[2*NN+1]; /* number of bits set */
    int rd=0; /* recursion depth */
    int rs;   /* return isminsymcfg */
    int hus[2*NN+1]; /* highest unset site(bit), goes from high to low */
                     /* better rename it to (work)position? */
//  int lus[2*NN+1]; /* lowest  unset site(bit), goes from low to high */
    int nbp[2*NN+1]; /* next bitpattern, 3=u+d 2=d 1=u 0=. */
    unsigned char *buf=NULL; ULong bi2; /* write buffer and its counter */
    tp_ns *p=&pns;
    long  t1=itime();
    // ULong i_out=1L<<8; /* output of estimated time */
    ULong nt=0l;        /* number tested cfgs for speed analyse */
    if (!mpi_i) printf(" running fast recursive ns(), but single_thread\n");
    p->norm2=0;   p->numans=0; p->num=0; // p->b.o=2;
    // sprintf(p->fname,"tmp_shared/l1_%04d.dat",0);
    // 2012-06-11 append if checkpoint>=1  ToDo: problems if chkpt_load=1
    //      and stored_chkpt=2++ (ToDo)
    checkpoint=0; // ToDo: fix ckpt-behaviour
    if (!mpi_i) {
      fl1=fopen("tmp_shared/l1_0000.dat",((checkpoint)?"ab":"wb"));
      if (!fl1) F0("hilbert.fopen tmp_shared/l1_0000.dat");
      // can happen if Bsize=7 write_buffer=512 + ^C + chkpt.resume 2012-06-20
      if (ftell(fl1)%Bsize) fseek(fl1, 0L-Bsize-ftell(fl1)%Bsize, SEEK_END);
    }
    bi2=0;   /* ini counter for write buffer */
    buf=(unsigned char *)malloc(L1_BUFLEN*Bsize); /* 512k*5B */
    if (!buf) F2("malloc %e for l1_buf node %d", L1_BUFLEN*1.*Bsize, mpi_i);
    // b.o=2; /* ??? */

    /* the initial state is all sites with maximum bits set (111...111) */
    state[0]=TBASE0;
    for(ii=0;ii<nn;ii++) state[0]=puti(state[0],ii,CM(2,2,3)); // JJ: 2=d
    // start with maxcfg CM(0xff=dddddddd,0xaaaa=dddddddd,0xffff=33333333)
    
    hus[0]=nn;                // start from MSB+1=nn go down to 0
    nbp[0]=CM(0,-1,-1);       // nbp+1=CM(1=u,0=.,0=.)  goes up ++ (min-1)
    hus[0]=nn+1;              // start from MSB+1=nn go down to 0
    nbp[0]=CM(1,1,2);         // nbp+1=CM(1=u,0=.,0=.)  goes up ++ (max-1)
    id[0]=nn; iu[0]=CM(0,0,nn); // ini_maxcfg
    
    // ToDo: prepare hus,nbp,id,iu,state[] dependend from startcfg0
    
    i2=0; num0=0; sum=0; /* counter scfg states + scfgs_with_sign0 */
    if ((vvv&5) && !mpi_i)  /* hus = highest unset bit, rd = recursion depth */
      printf("DBG: start  %s rd %2d hus %2d (maxcfg), vvv&4 for more info\n",
             ccfg(state[rd]), rd, hus[rd]);
    /* now reduce bits from highest bits, test if smallest and recursive */
    /* ToDo: hus and nn-rd is the same! remove hus[] */
#if 1
    /* nu,nd=0,40 will not work else Apr09 */
    if (CM(nu==0, nu==0 && nd==nn, nu==nn && nd==nn)) { // n1=1 only maxcfg
      if (!mpi_i) printf("DBG: do trivial case endcfg=startcfg extra\n");
      rs=isminsymcfg_lNbrk(state[rd],&b); nt++; /* 1=true 0=false -1=destr_sign */
#ifdef CONFIG_ANSATZ
      if (ansatz_test(state[rd])) rs=0; // ToDo: untested!
#endif
      /* nondestructive phase, store mincfg=state[rd] to l1[bi2*BSize] */
      if (rs== 1) { putl1(bi2,state[rd],buf); /* count symcfg, sign!=0 */
                    i2++;p->num++;bi2++;p->norm2+=b.norm2; }
      if (rs==-1) { num0++; p->norm2+=b.norm2; } /* count sign0 */
    } else
#endif
    // ToDo: check first for valid/go_deeper and do next/break after that
    //  from dddd to dddd=break
    //  from uddd via dudd ddud dddu to uddd=break
    //  from uudd via uuud uudu udud uduu uddu uddd=break
    for(;;) { /* can we do that for a thread specifique range? */
      // next bit pattern on current site
      //   JJ:  [0] 1 1 ...   tJ: [-1] 0 1 0 1 ... tU: [-1] 0 1 2 0 1 2 0 1 2
      // ToDo: check tU, should it not go down ([3] 2 1 0 2 1 0 ...)
      nbp[rd]++; // JJ: 1=u 2=d 3to1=u 2=d ... 
      if (CM(tnu!=0,tnd!=nn,(tnu!=nn || tnd!=nn)) // exception n1==1
       && nbp[rd]>CM(1,1,2)) {
        hus[rd]--; /* highest unset site goes down to next lower site */
        nbp[rd]=CM(1,0,0); /* next bit pattern JJ: u */
      }
      if (hus[rd]<=0 CM(|| tnu-iu[rd]>hus[rd],\
                        || id[rd]-hus[rd]>tnd || id[rd]<tnd || iu[rd]>tnu || iu[rd]+hus[rd]<tnu || id[rd]+iu[rd]<tnu+tnd,\
                        || hus[rd]<iu[rd]-tnu || hus[rd]<id[rd]-tnd)) {
        /* now we have tried all sites, so we can return */
        if (rd==0) break; /* all work done */
        rd--;     /* go up */
//        if (i2<x_max && vvv&4)
//          printf("\n come back %s rd %2d hus %2d %2d+%2d %.0f",
//                 ccfg(state[rd]), rd, hus[rd], iu[rd], id[rd],1.*nt);
        continue; /* return and try next deeper state */
      }
      /* probewise reset bit at hus=highest_unset_site (hus runs from nn to 0) */
      hus[rd+1]=hus[rd]-1;/* go to next lower site index when going deeper */
      nbp[rd+1]=CM(0,-1,-1); // min-1=max, next_bit_pattern
//      hus[rd+1]=hus[rd];/* go to next lower site index when going deeper */
//      nbp[rd+1]=CM(1,1,2); // max-1
      /* set nu, nd to have correct working behaviour of minsymcfg */
      id[rd+1]=id[rd]-1 +(nbp[rd]>>1); /* npb=0,1: nd-- */
      iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(1,1,0):CM(0,0,-1));  /* 1: nu++  0: nu-- */

      // 2012-05 add {} for empty macro
      CM({},if (iu[rd+1]!=tnu && id[rd+1]==tnd) { continue; },{})
      CM({},{},if (iu[rd+1]<tnu || id[rd+1]<tnd) { continue; })
      /* much faster for N=40 1u1d */
      CM({},{},if (iu[rd+1]-hus[rd+1]>tnu || id[rd+1]-hus[rd+1]>tnd) { continue; })
      CM({},if(id[rd+1]<tnd || iu[rd+1]>tnu || iu[rd+1]+hus[rd+1]<tnu || id[rd+1]+iu[rd+1]<tnu+tnd) continue;,{})

#ifdef CONFIG_S1SYM  /* skip invalid LM-patterns 2016-01-11 only 8-10% */
//     recursion start ffff=~vacuum set 0 highest bit first going down
//     if LSize[i]>1 && LNext[i]<nn && bit[LNext[i]>bit[i] invalid
//        bit = 1&(state[rd+1]>>bit)
//    test next higher bit within LM is same or less as this bit
      if (/*sym.ls[hus[rd+1]]>1 &&*/ sym.ln[hus[rd+1]]<nn /* valid LM sites */
       && geti(state[rd],sym.ln[hus[rd+1]]) > nbp[rd]) continue; /* noLM? */
#endif    

      state[rd+1]=puti(state[rd],hus[rd+1],nbp[rd]); /* lower bit pattern, down -> up */
      /* go deeper if we have only one possibility to set the rest cfg */
      CM({},{},\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==hus[rd+1]) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==0) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]-tnd==0) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]    ==hus[rd+1]) { rd++; continue; })
#if 0
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]    ==hus[rd+1] \
                      && id[rd+1]-tnd==hus[rd+1]) { rd++; continue; })
#endif
      // 38+1
      CM({},\
      if (hus[rd+1]>0 && id[rd+1]-hus[rd+1]==tnd \
                      ) { rd++; continue; },{})
#if 0
      // 38+1
      CM({},\
      if (hus[rd+1]>tnd && id[rd+1]==hus[rd+1] \
                      ) { rd++; continue; },{})
#endif

#if 1  // 2012-05-25 no speedup for nu<nd, speedup 2..3 for nu>nd (10x slower)
      // 39+1 speedup
      CM(\
      if (hus[rd+1]>0 && tnu-iu[rd+1]==hus[rd+1] \
                      ) { rd++; continue; },{},{})
#endif
#if 1 // useless? s1-triangle 46+8 2.3m vs. 0.28m 2012-07-03
      CM(\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      ) { rd++; continue; },{},{})
#endif

      /* ToDo: filter thread specifique range here? */
      nu=0; nd=nn; /* trick to use sym_ud to early */
      if (iu[rd+1]==tnu && id[rd+1]==tnd) {nu=tnu; nd=tnd;} /* sym.ud=on */
      rs= isminsymcfg_lNbrk(state[rd+1],&b); /* 1=true 0=false -1=destr_sign */
      // rs|=isminsymcfg_lNbrk(state[rd+1],&b); /* 120629 sym=324 4.7/2.38m=1.97 */
      nt++; /* count isminsymcfg calls for speed output */
      if (!mpi_i)
      if (   (  i2<x_max     && (vvv&4))  // first iterations 
          || ((!(nt&(nt-1))) && (vvv&5))) // show step 2^n only (ln2-output)
          // || hus[rd+1]!=nn-rd-1)
      if ((vvv&4) || itime()-t1>10) /* 2015-12 more output, ToDo PRF */
       printf("PRF: try % d %s rd %2d hus %2d %2d+%2d %5.0e %7.1e t/m=%4.0f\n",
          rs, ccfg(state[rd+1]),rd+1, hus[rd+1],iu[rd+1],id[rd+1], 1.*nt,
          1.*i2+((rs==1 && rd+1>=tnu)?1:0), (itime()-t1)/60.); // found
      if (rs) { /* go deeper or count for 1=true and -1=destructiv_sign */
        if (iu[rd+1]==tnu && id[rd+1]==tnd) { /* found symmmetric state */
#ifdef CONFIG_ANSATZ
          if (!ansatz_test(p->ib1)) { p->numans++; rs=0; }
#endif
          if (rs==1) {  /* without destructive phase */
            if (bi2 >= L1_BUFLEN) {  /* write out buffer, disk full? */
              if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F0("fwrite l1");
              bi2=0;
            }
            putl1(bi2,state[rd+1],buf);   /* store mincfg to l1p[l1s] */
            i2++; p->num++; bi2++; p->norm2+=b.norm2; /* count stored cfgs */
            /* found: hus[rd+1]=nn-(rd+2) */
          } /* rs==1 found smallest, nondestructive sign */
          if (rs==-1) { num0++; p->norm2+=b.norm2; } /* count sign0 */
        } else { // dive into the deeper subtree
          /* ToDo: + receive + write results from other tasks too
             max cfgs which can be expected? tasks split at max_rd-fixnr
             end rd is tnu?  JJ: rd=0..nu=0..nn-nd
                             tU: rd=0..(nn-nu + nn-nd)
                             tJ: rd=0..nn-nd ???  ToDo: badLM 2015-12
                sample tU: nn=40 nu=35 nd=38 chain cxx0- n1=255018
              ToDo: if rd==nn-nd-6 use mpi_i==tn%mpi_n or similar 2015-12
             */
          /* found symmetric mother cfg, go deeper */
          rd++; /* go deeper, ToDo: start by putting rest up's to next sites? */
// v2.38 gave n1=0 where it should not, nut sure if I fixed this correct
//          if (rd>nu && CM(1,0,0)) break; // fix nud=0,40 endless 2^40 loop 
//          if (rd>tnu && CM(1,0,0)) break; // fix nud=0,40 endless 2^40 loop 
            /* ToDo: store found+nt, on return print found as success rate */
          if (rd>2*NN) { ERR(9999); } continue;
        }
      } /* smallest cfg */
//      else { nbp[rd]+=2; } /* 38+1 but 37+1? if . not possible, u also dont work??? tJ++ */
    } /* for(;;) */
    sum=p->norm2;
    nu=tnu; nd=tnd; /* restore gobal nu, nd */
    if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F0("fwrite l1");
    fclose(fl1); fl1=NULL; /* flush */
    free(buf);
    printf(" ns2.27 end found= %8lu of %8.0f srate= %5.2f sgn0= %lu ns.t= %.2f min\n",
           (ULong)i2, 1.*nt, nt*1./i2, (ULong)num0, (itime()-t1)/60.);
  } /* mpi_i=0 only, ToDo change to all mpi_i's for ofs,len */
#ifdef CONFIG_MPI
  if (nsalgo==1) {
    /* send (i2) from root and receive (i2) by all other ranks */
    MPI_Bcast(&i2,sizeof(ULong),MPI_BYTE,0,MPI_COMM_WORLD);
  }
#endif

  if (nsalgo==0) /* old pre2.27 method, parallel-scanning of scfgs */
  {
    double i3b= i3;
    if (mpi_n<20 && !mpi_i) printf(" HINT: (algo^=16) may be faster\n");
#if CM(1,0,0)  /* reduce search space by factor 1/2 for ud-sym */
    if (nu==nd && sym.num_lm==nn && sym.wud) i3/=2;
      /* ??? does not work for LM<NN */
//    if (sym.num_lm!=nn) { /* test JS 2012-05-20 */
//      tbase ib0=ll1(0);
//      if (i3>3 && next(ll1(i3/2-1))==ib0) i3=i3/2; // does not work!
//      if (!mpi_i) printf("DBG next(i3/2)=%s\n",ccfg(next(ll1(i3/2))));
//      // N=12*s1 0x00800bff 1000 0000 0000  1011 1111 1111
//    }
#endif                  
    // new 2012-06-24 we trust maxscfg with latest bug fixes
    // ToDo: problem on 2012-11-05 for CONFIG_ABEL + read 3syms + -t2
    // ToDo: better name lastsymcfg instead of maxsymcfg?
#if 0 // buggy 2012-11-05, replace get_maxscfg !!!
    i3= ll2(maxscfg0)+1;
#else
    if (!mpi_i) printf(" ToDo: fix get_maxsymcfg code to reenable speedup (Joerg2012)\n");
#endif
    if (!mpi_i) printf(" i3 reduced by factor %.2f\n",i3b*1./i3);
#ifdef CONFIG_S1SYM
    if (sym.num_lm!=nn) { 
      /* avoid a bug for (s=1,JJ,mpt), cfgs counted twice */
      /* 2012-01 test N=7 s=1 LC n1=19 -9.57219835 mpi2=OK pt2=OK noProbs */
      // N=7 mpi=8 n1=19(+1 bad) see lc_s1.gpl
#if 0
      i4=1; /* ToDo: try to remove this again, bad sum_n2! i4=pt*mpi */
#endif
    }
#endif
/*
    Better scaling diskless versions:
    - searchpackets= 1/n...1e6 store result-packets on free nodes or local
    - write to local buffer size = 1e6
      copy from buffer to new chunk size=(0...1e6)
        + repeat with cfg += threads*n
      on 1000 chunks per node? interleaved for nodes (simple resort)?
      table/list of tbase-chunks and length_of_chunk
       mit pointer vorgaenger? + nachfolger thread/position
     1. create l1buf  of length n1e=min(maxveclen/2,max_estimated_node_n1)
        and    l1buf2 of length 1M
     2. check 0..1M(@node1)..2M(@node2)... and write to local-l1buf2
     3. distribute n1found[] over nodes
     4. copy l1buf2 to l1buf
        goto 2. +nodes*1M
     5. redistribute l1buf, shorten using remalloc 

 */
    numans=i2=0; num0=0; sum=0;
    /* write to files within threads to avoid unbalanced memory */
    // rewrite so that can return earlier (after fiex time) and rest work
    //   will be distributed under the tasks? 
    /* i3=(N!)(nu!nd!) i4=num_threads=mpi*pt or 1 */
    /* new: 2...m steps, first distribute 1/16 i3, than 1/4 i3, than 1/2?
     *   distribution depends from prefered run, go from smaller to bigger
     *   steps (which a parallel computed)
     * ToDo: divide in compute chunks (extra loop of about 2..8)
     */
{   // split for better balance (ToDo)
    // for heterogen clusters? (every thread tries to stop after
    //   1 minute, 2 minute, 4minute etc. ???) more status messages
    // send different lengths to the workers? depending from 
    ULong idx_start=0, // last start index 0...i3 within chunk
          idx_stop=0,  // last end index i4 .... i3 (end_idx+1)
          l1range=i3;    // range of first run 1/256 1/16 1, renamed 2012-11-14
    int  t1=itime();   // chkpt_time
    int io=0; // outer loop index
    int mintime=999999, maxtime=0, sumtime=0; // computing time per task, summed threads

/* 2012-04 new version use mpi to transfer data to n0, n0 writes to file */
    /* truncate/reset file l1_0000.dat */
    if (!mpi_i) {
      fl1=fopen("tmp_shared/l1_0000.dat",((checkpoint)?"ab":"wb"));
      if (!fl1) F0("hilbert.fopen tmp_shared/l1_0000.dat");
      // setvbuf( fl1, NULL, _IONBF , 0);
      i2= ftell(fl1)/Bsize; // long
      printf(" start_i2= %.0f Bsize=%d mpi_n=%d\n", 1.*i2, Bsize, mpi_n);
      // ToDo: full disk on big jobs!?
      // fclose(fl1);
    }
    /* big values alloc more memory but do less communication (can cause slowdown) 
     * search range per thread 1**20 causes slow down on satuated 100tasks*1GbE 
     *  sicortex.mips.5GHz 1<<20 cpu=80%(32MB,3-30s/io), 1<<25 cpu=100%(254MB,2-39m/io) tri48
     *   but cpuX=50min cpu0=33min diff=17min(MPI+NFS)
     * range=((ULong)1)<<22; // 100MB 1024 256 64 16 (2^(8*2)) start with small enough?
     * set to highest value is good idea for biggest systems but is limited by hardware 
     *  ToDo: 2*buf needed for async transfer_mpi/disk 17min vs. 33min
     */
    l1range=((ULong)1)<<25; // 254MB * tasks/cpu, start with small enough?
    l1range=((ULong)1)<<24; // 145MB * tasks/cpu, start with small enough?
    // 2014-08 l1to big for small mem footprint analyzes
    l1range=((ULong)1)<<18; //   2MB 2014-08-07
    // 1<<28 = 1.8GB for 32GB/8cores but cpu=20% on 14*1GbE (dont fit to 4GB/6cores)
    //  ToDo: better fill 16MB buffer (partly) by computing 8*16M-chunks,
    //        on overflow restart with chunk/2, ggf. underflow chunk*2?
    // ToDo: testmalloc ggf l1range/2 + MPI_BCast (lowmem-nodes)
    //       may fail, malloc race condition may alloc more than allowed 2012
    if (l1range>i3/2/mpi_n) l1range=i3/2/mpi_n; // 2012-06-22
    if (l1range==0) l1range=1;   // avoid stupid error
    if (l1range>i3) l1range=i3;
    // if (l1range<i3>>16) l1range=i3>>16;  // 27er s=1 nc=2e+15 l1range=3e10 OOM
    idx_start= 0; // default start, we may use higher value for chkpt1.resume
    if (checkpoint>=1) idx_stop= idx_start= ll2(startcfg0); // 2012-06-11
    if (!mpi_i) 
      printf(" start_i1= %s %.0f\n", ccfg(ll1(idx_start)), 1.*idx_start);
    // if (!mpi_i) printf(" malloc %.0f * %d\n",l1range*1.,(int)Bsize);
    // ToDo: HowTo do this nonhomogen data production locally with
    //    parallel pragmas?
    for ( ppns=&pns, ii=0; ii<pt_n; ii++ ) { /* all threads */
      if (!ppns) ERR(9011);                     /* check buffer for thread */
      ppns->buf=(unsigned char *)malloc(/*L1_BUFLEN 2012-04*/l1range*Bsize);
      if (!ppns->buf)
         F2("malloc %.2e L1_buf on node %d, plz reduce range", l1range*1.*Bsize, mpi_i);
      ppns->buflen= l1range; // ULong in units Bsize, available flag 2012-11-14
      if( ii<pt_n-1) {
         ppns->next = (tp_ns *) malloc(sizeof(tp_ns)); // ToDo: alloc array outside loop!?
         if (!ppns->next)
           F3("malloc %.2e tp_ns on node %d thread %d",
               sizeof(tp_ns)*1., mpi_i, ii);
      }
      ppns=ppns->next;
    } // start threads ii
    i4=pt_n*mpi_n; // 2012-06-26 
    /* divide into data chunks (balance compute time)*/
    for (io=0; idx_start < i3; idx_start=idx_stop, io++) {
      // stop=from would(should) be dummy call! ToDo
      if (l1range!=(l1range<<2)/4) // may be zero, avoid false alarm
         F0("l1range overflow, please report to the author");
      // ToDo: if loop is to small and enough mem enlarge l1range
      // l1range<<=2; // 1/1024 1/256 1/64 1/16 1/4 rest (2^(8*2)) growing size
      // ... but remalloc needed!
      idx_stop+=i4*l1range; // end of chunk = start of next chunk
      if (idx_stop>i3) { l1range=(i3-idx_start+i4-1)/i4; idx_stop=i3; }
      if ((!mpi_i)&&(io<1/*||(io&(io-1))==0*/)) /* io == power of 2 */
        /* 9 chars - .e+xx = 4 digits */
        printf("DBG: io= %6d start%%= %6.3f range%%= %9.5f t+= %s\n",
          io, idx_start*100./i3, l1range*100.*mpi_n*pt_n/i3, sclock(~0l));
      // ToDo min+maxtime
      for ( ppns=&pns, ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
        if (!ppns) ERR(9011);                     /* check buffer for thread */
        ppns->block=ii;                           /* block number */
        /* v2.26 (propto 2^ii)=unbalanced, try ii^2?, v2.27 (propto ii) */
//        ppns->i1=idx_start+(l1range+i4-1)/i4*(ii);   /* start of block */
//        ppns->n =idx_start+(l1range+i4-1)/i4*(ii+1); /* stop = start block+1 */
 // no overlap, but overflow? 2012-01-20
        ppns->i1=idx_start+l1range*(ii)  /*/i4*/; /* start of block */
        ppns->n =idx_start+l1range*(ii+1)/*/i4*/; /* stop = start block+1 */
        if (ppns->i1>i3) ppns->i1=i3;             /* aware of rest block */
        if (ppns->n >i3) ppns->n =i3;             /* aware of rest length */
        if (ppns->i1>i3) ppns->n = ppns->i1 = 0;  /* empty range */
        // malloc outside loop, l1range/2 in case of malloc fail (malloc-slowdown?)
        //  ppns->buf=(unsigned char *)malloc(/*L1_BUFLEN 2012-04*/l1range*Bsize); 
        //  ppns->buflen=l1range; // old: L1_BUFLEN 2012-04
        ppns->norm2=0;   ppns->numans=0; // ppns->b.o=2;
        ppns->num=0;     ppns->num0=0;
        // sprintf(ppns->fname,"tmp_shared/l1_%d%04d.tmp",io,ii); /* unique name */
        /* each thread writes its file, aware of max file handles! */
        parallel_start(ii%pt_n, ns_thread, ppns);  /* ! ! ! */
        // if( ii<i4-1 && ii%pt_n<pt_n-1)
        //   ppns->next = (tp_ns *) malloc(sizeof(tp_ns)); // ToDo: alloc array outside loop!?
        ppns=ppns->next;
      } // start threads ii
{ // write computed data-packets via task0 to l1-file
      // 2012-04: write one file, old 33000++ files caused problems
      // 2016-01: writing one file from multiple nodes and read from one
      //       causes problems on nfs (rm + t0.write + t1.write + t0.read)
      //  ToDo: send all data to task0
      static ULong count_per_io=0; // debugging number found configs per printf
#ifdef CONFIG_MPI
      MPI_Status   status;
#endif
      // fl1=fopen("tmp_shared/l1_0000.dat","ab"); if(!fl1)F0("reopen_l1");
      // ToDo: send to mpi=0 thread and store full l1 there
      // if (vvv && !mpi_i) printf(" join threads\n");   /* easier to debug */
      for ( ppns=&pns,ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
        parallel_wait(ii%pt_n);
        // save_l1(ppns->l1p,ppns->l1s,i2); /* write to tmp_shared/l1_%04d.dat @pos i2 */
           /* uchar *l1p, ulong stored, ulong ofs */
        count_per_io+= ppns->num; // for debugging
        i2+=ppns->num;        sum+=ppns->norm2; 
        numans+=ppns->numans; num0+=ppns->num0;
        sumtime += ppns->time0;
        if (mintime>ppns->time0) mintime=ppns->time0;
        if (maxtime<ppns->time0) maxtime=ppns->time0;
        /* may be we should store ppns->num too for later MPI distribution */
        /* distribute ppns->num per mpi-all-to-all for mpi-code? */
#if 0 // to much output!?
        if (vvv&2 && (ii<2 || ii>=mpi_n*pt_n-2))
          printf(" node %2d.%2d n1= %9.0f n_ans= %9.4g sum_n2= %9.4g t[m]= %5.2f\n",
               mpi_i, ii%pt_n, (double)i2, (double)numans, (double)sum, ppns->time0/60.);
#endif
        // ToDo: send to task 0 which allone writes to the l1-file = ok
#ifdef CONFIG_MPI
// if (mpi_i) printf("DBG.n%d: wait for token from 0\n",mpi_i);
// without blocking, task0 will get OOM on some MPIs 2012-06 (mpich)
// ToDo: do passiv wait MPI_Isend MPI_Irecv MPI_Wait MPI_Waitall MPI_Test
//        MPI_Test + man 3 usleep or man 2 nanosleep
        if (mpi_i) { int b2; // wait for a token, see OOM-Problem on task0
          MPI_Recv(&b2,1,MPI_INT,0,1,MPI_COMM_WORLD,&status); // blocking
          // ToDo: receive free packets and/or next destination node
        }
// if (mpi_i) printf("DBG.n%d: send to 0 count=%d\n",mpi_i,(int)ppns->num);
        if (mpi_i)
          MPI_Send(ppns->buf,ppns->num*Bsize,MPI_CHAR,0,0,MPI_COMM_WORLD);
// if (mpi_i) printf("DBG.n%d: send to 0  OK\n",mpi_i);
        if (!mpi_i)
#endif
        if (ppns->num) /* ToDo: pt_n>1 untested 2012-05 */
        if (fwrite(ppns->buf,Bsize,ppns->num,fl1)
                                != ppns->num) F0("fwrite l1");
        // now free outside io compute disptribution loop 2012-06-22
        // free(ppns->buf); // 2012-4
        // p2=ppns;
        ppns=ppns->next;
        // if (ii%pt_n>0) free(p2); // free(ppns) after access!
      } // join threads ii 
      // if (fl1) fclose(fl1);
#ifdef CONFIG_MPI
// receive and write to file
if (!mpi_i)
{     int ii;
      unsigned char *buf; /* space for results (sym-configs) (r[w]) 2012-04 */
      int count;       /* size of buffer in units of Bsize (r) 2012-04 */
      // buf=(unsigned char *)malloc( l1range*Bsize);
      buf=pns.buf; // reuse compute buffer 2012-06-22
        /* should never happen but it did 2012-05-11 after 512+ loops mpi=0 */
        /*   pot. reasons: slurmd was 400MB? at n0, L1_BLen was 8*l1range */
        if (!buf) { E0("malloc");system("free;ps aux");F0("malloc"); }
                      
      for (ii=pt_n;ii<mpi_n*pt_n;ii++) { int b2=0;
// printf("DBG.n0: send token to ii=%d\n",ii);
        MPI_Send(&b2,1,MPI_INT,ii/pt_n,1,MPI_COMM_WORLD); // token to ii
        // token is needed to avoid OOM by MPI caching (n-to-1-Data-transfer) 2012-06
// printf("DBG.n0: recv from ii=%d max=%.0f\n",ii,1.*pns.buflen);
        MPI_Recv(buf,pns.buflen*Bsize,MPI_CHAR,ii/pt_n,0,MPI_COMM_WORLD,&status);
        MPI_Get_count(&status, MPI_CHAR, &count);
// printf("DBG.n0: recv from ii=%d OK count=%d\n",ii,count);
        if (fwrite(buf,Bsize,count/Bsize,fl1)
                  != (size_t)count/Bsize) F0("mpi0.fwrite l1");
        count_per_io += count/Bsize;
      }
      /* lot of output (sumcount!=0) if nosym! */
      if (/*sumcount ||*/ (io<4||(io&(io-1))==0)) { /* io == power of 2 */
        /* %9.4g = 9 chars - .e+xx = 4 digits */
        printf("DBG: io= %6d stop%%= %7.3f count= %10.5g t+= %s\n",
          io, idx_stop*100./i3, 1.*count_per_io, sclock(~0l));
        count_per_io= 0;
        fflush(fl1); // better checkpointing
      }
} // (!mpi_i)
#endif /* CONFIG_MPI */
} // join results mpi+pthreads
      // ToDo: writing chkpt1 after time>=chktp_time/2
      if (checkpoint<=1 && !mpi_i)
      { int t2=itime();FILE *f1;
        extern int chkpt_time;
        if (t2-t1 >= chkpt_time/2) {
          // we need consistency, else s1cfgs may be lost 2012-06-25
          fflush(fl1); // consistent checkpointing
          /* fsync: _BSD_SOURCE || _XOPEN_SOURCE || _POSIX_C_SOURCE >= 200112L */
          /* fdatasync(): _POSIX_C_SOURCE >= 199309L || _XOPEN_SOURCE >= 500 */
          fdatasync(fileno(fl1)); /* sync to disk for consistency, unistd.h */
          printf(" w.chkpt1 %s\n",ccfg(ll1(idx_stop)));
          f1=fopen(FCHKPT "1","wt");fputs(ccfg(ll1(idx_stop)),f1);fclose(f1);
          // ToDo: store i2 + idx_start, umlauf? ???
          if (checkpoint<1) { f1=fopen(FCHKPT,"wt");fputs("1\n",f1);fclose(f1); }
          checkpoint=1;
          t1=t2; // chkpt1 timer reset
        }
      }
    } // io++ idx_start+=... outer loop (l1range chunks)
    for ( ppns=&pns,ii=0; ii<pt_n; ii++ ) { // syscall outside main loop
      free(ppns->buf); /* 2012-4 */
      ppns->buflen=0;  /* 2012-11-14 not available flag */
      p2=ppns; ppns=ppns->next;
      if (ii>0) free(p2); /* free(ppns) after access! */
    } // join threads ii 
    /* 2016-05-18 fix bad fclose_l1 msg */
    if (!mpi_i){ if (fl1){ if (fclose(fl1)) E0("fclose_l1"); fl1=NULL; }}
#if 0 // to much output!?
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (vvv&2 && (mpi_i<2 || mpi_i>=mpi_n-2))
      printf(" node %2d n1= %9.0f n_ans= %.4g sum_n2= %9.4g\n",
           mpi_i, 1.*i2, 1.*numans, 1.*sum);
#endif
} // split for better balance
#ifdef CONFIG_MPI
    { /* on Tru64+SLES9 MPI_UNSIGNED_LONG_LONG is not available */
      /* ToDo: compare to float sum for overflows ? */
      ULong lsend[4], lrecv[4];/* int err;*/ 
      lsend[0]=i2;     /* number of states counted in this MPI task */
      lsend[1]=numans; lsend[2]=sum; lsend[3]=num0;
      /*err=*/MPI_Allreduce( lsend, lrecv, 4,
                         MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
      /* ToDo: min/max for debugging/balance */
      i2=    lrecv[0]; /* overall number of counted states */
      numans=lrecv[1]; sum=   lrecv[2];
      num0=  lrecv[3]; /* number of forbidden symmetry states */ 
    }
#endif
    /* old: concat cat l1_*.tmp > l1_*.dat # replaced by seq. function */
    MPI_Barrier(MPI_COMM_WORLD); /* wait for node0 */
  } // nsalgo==0 for parallel-scanning scfgs

  if (!mpi_i) {
    printf(" ns: n1= %.15g sgn0= %.8g n_ans= %.15g sum_n2= %.15g\n",
           1.*i2, 1.*num0, 1.*numans, 1.*sum);
    if (sum!=i3 && !mpi_i) printf(" sum!=i3");
    // get last element Little-Endian if (L1_PACKED)! DBG-2012-05-23
    tbase maxscfg=0;
    if (0!=load_l1((unsigned char*)&maxscfg, 1, i2-1)) {
      E0("numsymconf mpi0.load_l1 maxscfg (nfs-problems?)"); /* relax? */
      sleep(8);
      if (0!=load_l1((unsigned char*)&maxscfg, 1, i2-1)) /* again */
        printf(" load_l1=success");
    }
    // ToDo: error from load_l1?
    printf(" gt.maxscfg= %s from getmaxcfg()\n", ccfg(maxscfg0) );
    printf(" ns.maxscfg= %s from l1[%.0f-1]\n",  ccfg(maxscfg), i2*1.);
    if (maxscfg != maxscfg0 && !mpi_i) E0("bad maxscfg ???, ignore");
    /* write n1 to disk, does not cost much CPU */
    /* if (chkpt_mode & 1) */
    {FILE *f;f=fopen(FCHKPT "2","w");if(f)fprintf(f,"%lu",i2); // n1=i2
             else E0("w." FCHKPT "2");if(f)fclose(f);}
    {FILE *f;f=fopen(FCHKPT,"wt");fputs("2\n",f);fclose(f);}
    printf("PRF ns.t=%s checkpoint2",sclock(itime()-l1));
  }
  if (checkpoint<2) checkpoint=2;
  return(i2);
} /* ULong numsymconf(),  short: ns() L2782..3534 */

/* return a unique(?) list of counted different pairs = ising energy index
   used to block hilbertspace using model-symmetries in future versions
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode|=2: output list
   mode|=1: return list (else free list, return 0)
   buf = result-array or NULL (if NULL, alloc memory)
   called by: ini_norm2 storeh2 x_out => replace by
      int get_ising_erg(b1,mode,weights)
      return product j1ud*a1+j2ud*a2+... removes need of storing array
 */
int *get_dcorr(tbase b1, int mode, int *buf) {
  int *result, i1, i2, i3, ib1, ib2;
  /* map an site pair nisnjs to a Ising index to count up */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
   if (buf) result = buf;
       else result = (int *) malloc(sym.num_dcorr*CM(2,4,5)*sizeof(int));
   if (result) {
     int max_result=0;
     for (i1=0;i1<sym.num_dcorr*CM(2,4,5);i1++) { result[i1]=0; } /* reset */
     for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
     for (i3=i2+1;i3<nn  ;i3++) {
           i1=sym.corr[i2][i3];    /* access N*N pair array */
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           // printf("DBG: %d %d  %d %d  %d\n",i2,i3,ib1,ib2,i1);
           /* PM600MHz 32+8 isinghash=14s for 4*if(&&||)else */
           /* PM600MHz 32+8 isinghash=17s(int),12s(char) for array-access */
           result[CM(2,4,5)*i1+map[4*ib1+ib2]]++;
           // => ising+=weight[CM(2,4,5)*i1+map[4*ib1+ib2]]; /* +Ji1 for ud */
     }
     for (i1=0;i1<sym.num_dcorr*CM(2,4,5);i1++) // get max for most compact
       if (max_result<result[i1]) max_result=result[i1]; // ... output
     if (mode&2) {
       /* printf("\n dcorr: cfg=%s", ccfg(b1)); */
       printf(" dcorr:");  /* different pair correlations */
       for (i1=1;i1<sym.num_dcorr;i1++)
       for (i2=0; i2<CM(2,4,5)-1; i2++) /* dont print CM(uu,.u,.3) */
       if (max_result>9)
         printf(" %2d",result[CM(2,4,5)*i1+i2]); /* [j1,ud-.u] [j2,ud-.u] .. */
       else
         printf( "%d",result[CM(2,4,5)*i1+i2]); /* [j1,ud-.u] [j2,ud-.u] .. */
     }
   }
   /* free if temporary allocated, for output only */
   if ((!(mode&1)) && (!buf) && result) { free(result); result=NULL; }
   return result;
}

/* idea: sort l1 for norm2 (better ising-erg), to avoid storing norm2 */
#define N2Blen 1021  /* histogram size, prime number */
int     n2b[N2Blen]; /* histogram, hashed */
Tnorm2  n2h[N2Blen]; /* hash table, norm2 */

#if 0
tbase *b_l1=NULL; /* table: first_sym-config of each block (global) */
#else
tbase *mpi_l1=NULL; /* table: first_sym-config of each node (global) */
#endif

extern ULong x_max; /* ToDo: maximum number output xval.h => debug.h */

/* 2015-07 declare alloc_hbuf from spins.c */
thbuf *alloc_hbuf();
void *free_hbuf( thbuf *hbuf );
#define MyMIN(x,y) (((x)<(y))?(x):(y))  /* 2015-01-05 better than if redef */
/* --- ToDo: define fastest variant of minsymcfg_* as default here: --- */
/*  this is very CPU and compiler dependent (SIMD, pipelines etc.) */
/* #define minsymcfg_dflt b_smallest  */ /* hilbert.h */
/* -----
 *  early 1s speedtests moved from ini_norm2
 * ToDo: 2014 add rndmem, generators-getsym, butterfly benchmarks, maxflop/iop
 * 2015-03-26 add multithread to test speed vs. cache coherence effects
 */
void speed_tests(void *ptr){ /* called as thread, one pointer as arg */
  thbuf *hbuf = (thbuf *)ptr; /* 2015-05 change from local hbuf3 to argument */
  int pt = hbuf->block; /* 2015-04-16 use thbuf */
  ULong loops=0, l0=0; double tm2=0.0, tm1=0.0;
  static int flag_once[B_NUM]= {0};  /* warning: if pthread */
/* 2015-05: bigger than CPU cache (3MB) for realistic runtime estimation 
 * 2015-05-06 see no cache effects, speed 16K*8B == 1M*8B on 3M-cache
 *           ulimit -s unlimited # to allow big arrays on stack (segfault)
 *           found no difference between 16K*8B and 1M*8B
 * 2015-12-31 reduce to 4K*8B, because its to slow on ATOM -O0
 *
 * dont put big arrays to stack, ToDo: use malloc + free
 * HBLrVS is NUM_AHEAD_LINES*NZXMAX roundup to nVS
 */
#define NUM_CFG_SAMPLES MyMIN(HBLrVS,4*1024) /* 4K*8B, lg2 and VS multiple */
  /* static is bad for PTHREAD, also a reason to make it not to big 2016 */
  tbase ib4[NUM_CFG_SAMPLES+32]; /* __attribute__ ((aligned (32))); */ 
                                     /* 8*4k per thread temp stack-space */
  /* 2015-04 use hbuf->acj[hbuf->n=NZXMAX*NUM_AHEAD_LINES] for speed-test */
  if (flag_once[pt]) return; /* only once! */
  if (pt_n==1) MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (vvv && (!mpi_i) && !pt) {
    printf("\nspeed_tests (variants of %d-bit permutation, Ld_bits=%d",
           NN, Ld_bits);
#if defined(_Vlint_h)
    printf(", vlint[%d*uint%d]",B2size,(int)sizeof(int)*8);  /* 2016-01 */
#endif
    printf(")\n  blocksize=%dk*%dB vvv&3=%d \n",
           NUM_CFG_SAMPLES/1024, (int)sizeof(tbase), vvv&3);
  }
 // b1.o=0;
  /* get random configs for benchmarks, speed may be config dependend */
  { int i1, i2, i3; unsigned int rnd1=0;
    srand((int)(1000*fstime())); /* 2015-10-08 get more statistics, ms-rand */
    tbase ib0=TBASE0;
#if CM(1,0,0)
    /* 2016-01-08 tbase - int(1) gave bad results for vlint, fixed, avoid */ 
    // ib0=(((tbase)1)<<nd)-1; /* initial cfg before random permutaion */
    ib0=~((~TBASE0)<<nd); /* initial cfg before random permutation */
#else
    for (i2=0;i2<nu;i2++) ib0 |= ((tbase)1)<<(2*i2); /* set up-spins */
    for (i2=0;i2<nd;i2++) ib0 |= ((tbase)2)<<(2*i2); /* set down-spins */
#endif
#ifdef _Vlint_h   // was failing for NN=200 2016-01-07
    if ((ib0>>(CM(1,2,2)*nn-1))>>1 && pt<1) { /* found set bits in zero area */
      if (mpi_i==0 && pt==0) {printf("DBG: rcfg1 ");pout(ib0);}
    } else 
#endif
    { if (mpi_i==0 && pt<1)  printf("DBG: rcfg1  %s\n",ccfg(ib0)); }
    ib4[0]=ib0; /* have minimum state in it which is also smallest 2016-01 */
    for (i1=1; i1<NUM_CFG_SAMPLES /* || i1<NZXMAX*NUM_AHEAD_LINES*/; i1++) {
      /* randomly exchange nd bitpairs, ToDo: use kind of bfly? */
      /* ToDo: better scan rnd1-th 1 bit and rnd2-th 0 bit, 0+1 */ 
      for (i2=0;i2<CM(nd,(nu+nd+3)/2,(nu+nd+3)/2);i2++) {
        i3=(rnd1%nn)*CM(1,2,2)+CM(0,1,1)*(i2&1); rnd1>>=Ld_bits;
        if (!rnd1) rnd1 = rand(); /* use bitwise, minimize rand-calls */
        if     ((1&(int)((ib0>>i3) ^ (ib0>>i2))) == 1) 
             ib0^=(((tbase)1)<<i3) ^ (((tbase)1)<<i2);
      }
      ib4[i1]=ib0;            /* initial cfg before random permutaion */
#ifdef _Vlint_h
      if ((ib4[i1]>>(CM(1,2,2)*nn-1))>>1) { /* found set bits in zero area */
        if (mpi_i==0 && i1<3+(int)x_max/8 && pt<2)
          {printf("DBG: rcfg %d",i1);pout(ib4[i1]);}
      } else 
#endif
      { if (mpi_i==0 && i1<3+(int)x_max/8 && pt<2)
           printf("DBG: rcfg %d %s\n",i1,ccfg(ib4[i1])); }
    } /* for i1=0 i1<16K */
  } // ini ib4[16k]=rnd 

  /* ToDo: temporarely fill 1e6 l1 for realistic tests
      using nextNAHcfgs or better nextNAHscfgs 
    2014-05-23 PC-E5400-2700MHz-2MB-1200MHz g++4.1.2 -O3 -funroll-loops 20+20
    2015-01-28 quantum-E5450-3000MHz-6MB g++4.8.2 -O2 20+20
    2015-01.28 32core64threads quanta1-E5-4620-2200MHz-16MB g++4.4.7 -O2 20+20
      # SSE2,XOP 2*i64 may give scfg4 an advantage of 2 (not much)
      # XOP.VPCMOV=VectorCondMoves(d,a,b,s))
      #          for(i=0..255)d[i]=s[i]?a[i]:b[i]; = BFLY0
      # XOP.VPPERM=32BytePerm (64BPerm=4*32BPerm+2OR)  256bit
      # BFLYn: b=a&~mask;a&=mask;a=(a&maskn)>>n | (a&(maskn>>n));a|=b; 9clks*6
      # loop: b|=((a>>i)&1)<<p[i]; 4clks * N   ca: 1:3 (const.clks+pipe)
    ToDo: hamilton_nhv( i, hbuf->num_ahead_lines, hbuf, 2); incl. nhm_line
    ToDo: replace getl1() by pointer.part.l1 + len=ahead
          remove mode m1 and put extra functions outside (better readable)
          nhm_line sends H-elements to other tasks and receives others
           for computing and/or storing (store received not send)??? 2014
          for perftest
// ToDo: BMI2: intel-haswell-2013 amd-excavator-2015(geplant)
//    Parallel bit deposit and extract  (for const. runtime bitperm)
//    6 times  pext(v,m)<<32 | pext(v,~m)   6*14cycles ? vs 64*4cycles = 1:3
//   better t/cfg/sym?
//        6*14=90 (6*64bit=48B) vs. 4*40=160 (40*6bit ca. 40B)/sym
// ToDo: XOP.VPPERM 6*? (4*32B=128B)/sym vs. ... 40*1B=40B 3*Cache
//       + 8*8bit to 64*8bit expansion table (256*8B)
//  PTable: 8*256*8B=16kB/sym,8OPs  or 16*16*8B=2KB,16*10cyc
//                                  or  32*4*8B=1KB,32*10cyc 64*1*8B=512B
//          L1-SRAM ~4cycles    4-256KiB/core
//          L2-SRAM ~10cycles  64-512KiB/core
//          L3-RAM  ~40cycles    2-32MiB/chip
//          L-DRAM ~120cycles ~30ns            # 1% miss rate = half speed
//          R-DRAM           ~100ns 
//             see valgrind --tool=cachegrind ...
//    Best: BFLY/PEXT-version
// 2006 Y.Hillewitz+R.B.Lee Fast Bit Compression and Expansion with 
//      Parallel Extract and Parallel Deposit Instructions
//      papers/rechner/bit_math/hilewitz06FastBitCompression.pdf
//      papers/rechner/bit_math/Hilewitz_JSPS_08.pdf
// ToDo: benchmark symcfg_loopN (vs. get_smallest)
 * see speed_mpi.gpl comp2 2015-02 minsymcfg_lNbrk=minsymcfg=80%*SH
 * 2015-02-24: output t/cfs=t/(cfg*numsym) easy compare minsymcfg to symcfg
 */    
  hbuf->n= NUM_CFG_SAMPLES;
  sym_bn_m_ini(pt); /* call bn_mask initialization before any *symcfg_bn*-call */
  if (mpi_i==0 && pt==0) printf(
   "DBG: check_minsymcfg dflt=%5s nsym=%d, N=%d ncfgs=%d vvv&64 ...\n",
    ((CFG_USE_BENES)?"bnV  ":"lNbrk"),sym.numall+1, NN, (int)hbuf->n);
  if (mpi_i==0 && pt==0){ /* check benes/bfly correctness, recode 2016-01 */
    tbase ib3=TBASE0, ib5=TBASE0; /* struct sbase b; */
    tphase s1,s2; int i, nj=sym.numall;/* biggest 1111 below ns */
    thbuf *hbf2;  /* 2015-07-08 replace big local var by malloc */
    /* if enlarge stacksize (1G), orterun will fail on lowmem 2015-07 meggie */
    hbf2= alloc_hbuf();  
    if (!hbf2) F2("alloc_hbuf failed, mpi=%d thread=%d", mpi_i, pt);
    hbf2->block= hbuf->block; hbf2->n= hbuf->n;
    for (i=0;i<hbuf->n;i++) hbuf->acj[i]=ib4[i]; /* bad additional copy */
    for (i=0;i<hbf2->n;i++) hbf2->acj[i]=ib4[i]; /* bad additional copy */
    for (i=0;i<hbf2->n;i++) hbf2->asj[i]=1;
    for (i=0;i<hbuf->n;i++) hbuf->asj[i]=1;
    /* before first test T=1 symmetrie */
    if (sym.numall>=MaxSym) printf("WARNING: no space for Testsym=1\n");
    if (sym.numall<MaxSym) { /* 2016-01 noSym avoid div0 */
      /* 1-sym should be the default initialization */
      /* happens after "symtable full" */
      /* 2016-03-22 fix buggy 1st index nn to sym.numall */
      if (sym.i[sym.numall][nn-1]!=nn-1){
        printf("\nWARNING: no T=1 found at sym.i[numall], auto correction\n");
        for (i=0;i<NN;i++) sym.i[sym.numall][i]=i; // correct it
      }
    }
    if (sym.numall<MaxSym) /* 2016-01 testing T=1 */
    for (l0=0;l0<(unsigned long)hbuf->n;l0++) {
      static int nerr=0; int fail=0; nj=sym.numall;
      symcfg_loopN(ib4[l0],nj,&ib5,&s1,1);  /* reference */
      symcfg_bn(   ib4[l0],nj,&ib3,&s2,1);
      /* 2015-12-19 conj() as work around (bad), but same results ToDo */
      if (s1!=0 && ib3!=ib5) fail|=1;
      if           (s1!=s2)  fail|=2;
      if ((ib5    >>(CM(1,2,2)*nn-1))>>1) fail |= 16; /* bit problems */
      if ((ib3    >>(CM(1,2,2)*nn-1))>>1) fail |= 32; /* bit problems */
      if ((ib4[l0]>>(CM(1,2,2)*nn-1))>>1) fail |= 64; /* bit problems */
      if (fail || (l0<=x_max && vvv&64)) {
        /* better Norm(s1-hbuf->asj[l0])>1e-9 ??? */
        printf("DBG: cmp  lN cfg=%s Tsym=1 scfg=%s sgn=%s l0=%d fail=%d\n",
        ccfg(ib4[l0]),ccfg(ib5),outphase(s1),(int)l0,fail);
        printf("DBG:      bn cfg=%s Tsym=1 scfg=%s sgn=%s\n",
        ccfg(ib4[l0]),ccfg(ib3),outphase(s2));/*exit(1);*/
#ifdef _Vlint_h
    pout(ib4[0]);
    pout(ib5);
    pout(ib3);
#endif
        if (fail) nerr++; 
        if (fail && nerr>7) {printf("DBG: ... stop error report ...\n");break;}
      } // if bad
    } // for l0=0..(hbuf->n-1)
    if (sym.num_lm!=nn) /* 2016-01-12 testing LM */
    for (l0=0;l0<(unsigned long)hbuf->n;l0++) {
      static int nerr=0; int fail=0;
      /* reference */
      struct fullsbase b; b.is[0]=1; b.ib[0]=ib4[l0]; b.num=1;
            ib3=allsymcfg_lm(&b, 1);  s1=b.sign;
      s2=1; ib5=minsymcfg_lm(ib4[l0],&s2);  
      /* 2015-12-19 conj() as work around (bad), but same results ToDo */
      if (s1!=0 && ib3!=ib5) fail|=1;
      if           (s1!=s2)  fail|=2;
      if ((ib5    >>(CM(1,2,2)*nn-1))>>1) fail |= 16; /* bit problems */
      if ((ib3    >>(CM(1,2,2)*nn-1))>>1) fail |= 32; /* bit problems */
      if ((ib4[l0]>>(CM(1,2,2)*nn-1))>>1) fail |= 64; /* bit problems */
      if (fail || (l0<=x_max && vvv&64)) {
        /* better Norm(s1-hbuf->asj[l0])>1e-9 ??? */
        printf("DBG: cmp aLM cfg=%s LMsym  scfg=%s sgn=%s n2=%.0f l0=%d fail=%d\n",
        ccfg(ib4[l0]),ccfg(ib3),outphase(s1),1.0*b.norm2,(int)l0,fail);
        printf("DBG:      LM cfg=%s LMsym  scfg=%s sgn=%s n2=%.0f\n",
        ccfg(ib4[l0]),ccfg(ib5),outphase(s2),1.0*b_factor_lm(ib4[l0]));
        /*exit(1);*/
#ifdef _Vlint_h
    pout(ib4[0]);
    pout(ib5);
    pout(ib3);
#endif
        if (fail) nerr++; 
        if (fail && nerr>7) {printf("DBG: ... stop error report ...\n");break;}
      } // if bad
    } // for l0=0..(hbuf->n-1)
    /* first test single symmetries */
    if (sym.numall) /* 2016-01 noSym avoid div0 */
    for (l0=0;l0<(unsigned long)hbuf->n;l0++) {
      static int nerr=0; int fail=0; nj=sym.numall;
      symcfg_loopN(ib4[l0],l0%nj,&ib3,&s1,1);  /* reference */
      symcfg_bn(   ib4[l0],l0%nj,&ib5,&s2,1);
      /* 2015-12-19 conj() as work around (bad), but same results ToDo */
      if (s1!=0 && ib3!=ib5) fail|=1;
      if           (s1!=s2)  fail|=2;
      if ((ib5    >>(CM(1,2,2)*nn-1))>>1) fail |= 16; /* bit problems */
      if ((ib3    >>(CM(1,2,2)*nn-1))>>1) fail |= 32; /* bit problems */
      if ((ib4[l0]>>(CM(1,2,2)*nn-1))>>1) fail |= 64; /* bit problems */
      if (fail || (l0<=x_max && vvv&64)) {
        /* better Norm(s1-hbuf->asj[l0])>1e-9 ??? */
        printf("DBG: cmp  lN cfg=%s sym%03d scfg=%s sgn=%s l0=%d fail=%d\n",
        ccfg(ib4[l0]),(int)(l0%nj),ccfg(ib3),outphase(s1),(int)l0,fail);
        printf("DBG:      bn cfg=%s sym%03d scfg=%s sgn=%s\n",
        ccfg(ib4[l0]),(int)(l0%nj),ccfg(ib5),outphase(s2));/*exit(1);*/
#ifdef _Vlint_h
    pout(ib4[0]);
    pout(ib5);
    pout(ib3);
#endif
        if (fail) nerr++; 
        if (fail && nerr>7) {printf("DBG: ... stop error report ...\n");break;}
      } // if bad
    } // for l0=0..(hbuf->n-1)
    /* prepare for 2nd blk-test */
    minsymcfg_blk_lNbrk(hbuf); /* ToDo: ueberschreibt input = bad! pt!=0 */
    //minsymcfg_blk_bn(&hbf2); /* OK */
    minsymcfg_blk_bnVS(hbf2); /* BAD! ToDo: ueberschreibt input = ungut! */
    /* second test minsymcfg */
    for (l0=0;l0<(unsigned long)hbuf->n;l0++) {
      static int nerr=0; int fail=0; struct sbase b0; 
      minsymcfg_loopN(ib4[l0],&b0);  /* reference */
      /* 2015-12-19 conj() as work around (bad), but same results ToDo */
      if (b0.sign != 0) {
        if (b0.scfg  !=   hbuf->acj[l0])  fail |= 1;
        if (b0.scfg  !=   hbf2->acj[l0])  fail |= 2;
        if (b0.norm2 !=   hbuf->an2[l0])  fail |= 128;
        if (b0.norm2 !=   hbf2->an2[l0])  fail |= 128;
      }
      if (b0.sign !=      hbuf->asj[l0])  fail |= 4;
      if (b0.sign != Conj(hbf2->asj[l0])) fail |= 8;
      if ((b0.scfg      >>(CM(1,2,2)*nn-1))>>1) fail |= 16; /* bit problems */
      if ((hbuf->acj[l0]>>(CM(1,2,2)*nn-1))>>1) fail |= 32; /* bit problems */
      if ((hbf2->acj[l0]>>(CM(1,2,2)*nn-1))>>1) fail |= 64; /* bit problems */
      if (fail || (l0<=x_max && vvv&64)) {
        /* better Norm(s1-hbuf->asj[l0])>1e-9 ??? */
        printf("DBG: src      cfg=%s sym%03d l0=%6d fail=0x%02x\n",
          ccfg(ib4[l0]), (int)nj, (int)l0, fail);
        printf("DBG:      lN scfg=%s sgn=%s n2=%7.1e\n", /* compare base */
         ccfg(b0.scfg),      outphase(b0.sign),      (double)b0.norm2);
        printf("DBG:  blNbrk scfg=%s sgn=%s n2=%7.1e\n",
         ccfg(hbuf->acj[l0]),outphase(hbuf->asj[l0]),(double)hbuf->an2[l0]);
        printf("DBG:   bbnVS scfg=%s sgn=%s n2=%7.1e\n",
   ccfg(hbf2->acj[l0]),outphase(Conj(hbf2->asj[l0])),(double)hbf2->an2[l0]);
#ifdef _Vlint_h
    pout(b0.scfg);
    pout(hbuf->acj[l0]);
    pout(hbf2->acj[l0]);
#endif
        if (fail) nerr++; 
        if (fail && nerr>7) {printf("DBG: ... stop error report ...\n");break;}
      } // if fail
    } // for l0=0..(hbuf->n-1)
    free_hbuf(hbf2);
  } /* correctness-test */
#ifdef CONFIG_PTHREAD   /* sync, not overwrite cfgs! */
#endif  
  if (mpi_i==0 && pt==0) printf("DBG: check_minsymcfg.end\n");
  if (vvv && mpi_i==0 && pt==0) printf(
   "PRF: speedtest.start (1s/test, cfs=cfg*nsym, nsym=%d, N=%d, 80%%*SH)\n",
    sym.numall+1, NN);
  /* no need for warmup, switching clock in 10us = 1e-5 of 1s */
  /* 1st test on vvv=1|2 rest on vvv=2 only(safe time) */
  if (pt_n==1) MPI_Barrier(MPI_COMM_WORLD); /* debug */
  /* 2016-04-07 remove mpi_i */
  if ((vvv&3) /* && mpi_i==0*/) { /* ToDo: 2013 speed test minsymcfg() using blocks! */
    int i, j=0; tbase ib3=TBASE0; /* tU: tbase,tbase1,vlint */
    // printf(" hbuf->n=%d\n", hbuf->n);
    l0=0; tm2=0.0; tm1=fstime();
    /* hold as much as possible of sym.i in L1-cache! ToDo: howto */
    /* hbuf->acj buffers the datastream into VS-size blocks */
    for (loops=1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) {
        for (i=0;i<hbuf->n;i++,j=(j+1)&(NUM_CFG_SAMPLES-1))
           hbuf->acj[i]=ib4[j]; /* bad additional copy */
        if(nu==nd && sym.wud) minsymcfg_blk_dflt(hbuf); /* ToDo: overwrite input = bad */
          else                minsymcfg_blk_dflt_no_ud(hbuf); /* faster!? */
        for (i=0;i<hbuf->n;i++) ib3^=hbuf->acj[i];
      }  // why not removed by optimization?
      tm2= fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    } /* output speed */
    if (pt==0 && mpi_i==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d",
      "minsymcfg_dflt", tm2, /* hbuf->n* */ loops*1./tm2,
       tm2*1000000000./loops/hbuf->n/(1+sym.numall), 63&(int)ib3);
    if (pt==0 && !mpi_i) {
#if  CFG_USE_BENES
      printf(" bn%d",VS);
#else
      printf(" lNbrk");
#endif
      printf("\n");
    }
#ifdef CONFIG_MPI
    /* kautz2016-04-08.lc21+21 k00+ 950*6=5700cpus lNbrk
     *   0.25s++ best=175..183ns min=90.6..91.6% bad=2..166 m31n8(wechselnd)
     *   1.00s++ best=173..182ns min=89.4..92.9% bad=9..162 m31n14,m31n0(wechselnd)
     *   2.00s++ best=178..181ns min=90.3..93.8% bad=1-6 m32n10(wechselnd)
     *     950*1 best=180..184ns min=92.7..92.9% bad=27..736
     *       1*6 best=180..186ns min=95.3..96.5% bad=1..5 m0n0.2
     *  .25s 1*6 best=171..190ns min=92.2..99.7% bad=1..5 m0n0.0-5(wechselt) */
    if (mpi_n>1 && pt==0) { double inp[4], min[1], max[1], sum[4];  /* 2016-04-07 added */
      extern char nodename[21]; /* 2016-04-08 set global for err() see spins.c */   
      inp[0]=loops*1./tm2; // num min/max = weight
      inp[1]=0; inp[2]=0; inp[3]=0;
      MPI_Allreduce(inp, max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
      MPI_Allreduce(inp, min, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
      if (max[0]-min[0]>0 && max[0]!=0) { /* lowest 10% */
        if (inp[0]/max[0]<(0.900*min[0]+0.100*max[0])/max[0]) inp[1]=1.0; }
      if (max[0]-min[0]>0 && max[0]!=0) { /* lowest 1% */
        if (inp[0]/max[0]<(0.990*min[0]+0.010*max[0])/max[0]) inp[2]=1.0; }
      if (max[0]-min[0]>0 && max[0]!=0) { /* lowest */
        if (inp[0]==min[0]) inp[3]=1.0; }
      MPI_Allreduce(inp, sum, 4, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
      /* number of slow nodes (5% or 10%)? ToDo: better 32 slowest! */
      if (max[0]>0 && /*min[0]*100./max[0]<0.99 &&*/ !mpi_i)
        printf("PRF:    mpi speed range  %4.1f..%4.1f..100%%    t[ns]/cfs="
               " %7.2f bad=%.0f\n",
          min[0]*100./max[0], sum[0]/mpi_n*100./max[0], 
            1000000000./hbuf->n/(1+sym.numall)/max[0], sum[1]);
      /* ToDo: output best speed + t[ns]/cfg */
      if (inp[1]>0.0 && sum[1]<8.0) { usleep(10000);
        printf("DBG: slow task %5d %-20s %4.1f%%\n",
                mpi_i, nodename, inp[0]*100./max[0]); } else
      if (inp[2]>0.0 && sum[2]<8.0) { usleep(10000);
        printf("DBG: slow task %5d %-20s %4.1f%%\n",
                mpi_i, nodename, inp[0]*100./max[0]); } else
      if (inp[3]>0.0 && sum[3]<8.0) { usleep(10000);
        printf("DBG: slow task %5d %-20s %4.1f%%\n",
                mpi_i, nodename, inp[0]*100./max[0]); }
      MPI_Barrier(MPI_COMM_WORLD);
    }
#endif
  }
  if ((vvv&2) && (mpi_i+1)%mpi_n==0) { /* ToDo: 2013 speed test minsymcfg() using blocks! */
    tbase ib2=TBASE0, ib3=TBASE0; struct sbase b; // int i1,i2;
    b.scfg=(1UL<<nd)-1; /* JJ only! */
    l0=0; tm2=0; tm1=fstime();
    for (loops=64;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) {
        ib2=ib4[l0 & (NUM_CFG_SAMPLES-1)];
        minsymcfg_lNbrk(ib2,&b); ib3^=b.scfg;
        /* ToDo: force cache misses? by jumping sym-set or big ib2 steps?
             or PERF analysis for cache misses OR call nhm_lines (counts) */
       }
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d ud=%d+%d\n",
      "minsymcfg_lNbrk", tm2, loops*1./tm2,
       tm2*1000000000./loops/(1+sym.numall), 63&(int)ib3, nu, nd);
  }  /* min sym config */
  /* 2nd different test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+2)%mpi_n==0) {
#if 0 // old non-blocked-data code
    tbase ib2=TBASE0, ib3=TBASE0; struct sbase b;
    b.scfg=(1UL<<nd)-1; /* JJ only! */
    /* prepare cfg-array */
    l0=0; tm2=0; tm1=fstime();
    for (loops=128*CM(4,1,1);loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        ib2=ib4[l0 & (NUM_CFG_SAMPLES-1)];
        minsymcfg_loopN(ib2,&b);  ib3^=b.scfg; }
        /* new is using symcfg_loopN */
         /* ToDo: runtime doubled? less runtime on ATOM (2 cores, 2 MHz) */
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    /* 4Ni = 4N instructions per loop and symmetry (valid for JJ only) */
    /* ToDo: instructions per loop for tU? */
    if (pt==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d 4Ni\n",
     "minsymcfg_loopN", tm2, loops*1./tm2,
      tm2*1000000000./loops/(1+sym.numall), 63&(int)ib3);
#else
    tbase ib3=TBASE0; int i, j=0; l0=0; tm2=0; tm1=fstime();
    // printf("DBG: n=%d\n",hbuf->n);
    /* hold as much as possible of sym_bn_m in L1-cache! ToDo: howto */
    for (loops=1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        /* sym_bn_m should keep in the L1 cache, hbuf->acj is data-stream */
        /* here bad speed test, no datastream, just same data repeated */   
//        for (i=0;i<hbuf->n;i++) hbuf->acj[i]=ib4[i]; /* bad additional copy */
        for (i=0;i<hbuf->n;i++,j=(j+1)&(NUM_CFG_SAMPLES-1))
           hbuf->acj[i]=ib4[j]; /* bad additional copy, data-stream */
        /* ToDo: prefetch acj[i+VS] */
        if(nu==nd && sym.wud) minsymcfg_blk_lNv(hbuf); /* ToDo: ueberschreibt input = ungut! */
        else                  minsymcfg_blk_lNv(hbuf); /* faster!? */
        for (i=0;i<hbuf->n;i++) ib3^=hbuf->acj[i]; /* !0 if SAMPLES>2*hbuf->n? */
        /* ToDo: release acj[i] from cache */
      } tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) 
      printf("PRF: %13s%-2d t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d 4Ni\n",
     "minsymcfg_lNv", VS, tm2, hbuf->n*loops*1./tm2,
      tm2*1000000000./loops/hbuf->n/(1.+sym.numall), 63&(int)ib3);
#endif
  }
  /* 3th different test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+3)%mpi_n==0) {
    tbase ib2=TBASE0, ib3=TBASE0; struct sbase b;
    b.scfg=(1ul<<nd)-1; /* JJ only! */
    l0=0; tm2=0; tm1=fstime();
    for (loops=128*CM(4,1,1);loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        ib2=ib4[l0 & (NUM_CFG_SAMPLES-1)];
        minsymcfg_test3(ib2,&b);  ib3^=b.scfg;}
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d\n",
     "minsymcfg_test3", tm2, loops*1./tm2,
      tm2*1000000000./loops/(1+sym.numall), 63&(int)ib3);
  }
  /* 4th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+4)%mpi_n==0) {
    tbase ib2=TBASE0, ib3=TBASE0, ib5=TBASE0; /* struct sbase b; */
    tphase ss=1; int nj;
    for (nj=1;nj<=sym.numall;nj<<=1) ;
    nj>>=1; if(nj)nj-=1; /* biggest 1111 below ns */
    l0=0; tm2=0; tm1=fstime();
    for (loops=1024*4;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        ib2=ib4[l0 & (NUM_CFG_SAMPLES-1)];
        symcfg_loopN(ib2,(int)(l0&nj),&ib5,&ss,1);
        ib3^=ib5;}
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfg= %7.2f %d 4Ni\n",
     "symcfg_loopN", tm2, loops*1./tm2, tm2*1000000000./loops, 63&(int)ib3);
  }
  /* 5th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+5)%mpi_n==0) { /* vectorized */
    tbase *ib2, ib3[VS], ib5[VS]; /* struct sbase b; */
    tphase ss[VS]; int nj, i;
    for (i=0;i<VS;i++) { ib3[i]=TBASE0; ss[i]=+1; }
    for (nj=1;nj<=sym.numall;nj<<=1); 
    nj>>=1; if(nj)nj-=1; /* biggest 1111 below ns */
    l0=0; tm2=0; tm1=fstime();
    for (loops=16;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0+=VS) { 
        ib2=&(ib4[l0 & (NUM_CFG_SAMPLES-1)]);
        symcfg_loopNv(ib2,(l0/VS)&nj,ib5,ss,1); // ib2 must be aligned 8B
        for (i=0;i<VS;i++) ib3[i]^=ib5[i];
//      if (l0<8) printf("DBG 4v %d %s %s %s %s\n",
//        (int)l0,ccfg(ib5[0]),ccfg(ib5[1]),ccfg(ib5[2]),ccfg(ib5[3]));
        }
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    for (i=1;i<VS;i++) ib3[0]^=ib3[i];
    if (pt==0) 
    printf("PRF: %13s%-2d t=%3.1fs loops/s= %7.1e t[ns]/cfg= %7.2f %d 4Ni\n",
     "symcfg_loopNv", VS, tm2, loops*1./tm2, tm2*1.e9/loops, 63&(int)
     ib3[0]); /* %-2d means left justified, fill spaces after */
  }
  /* 6th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+6)%mpi_n==0 && sym.numall /* 2015-09-23 x%0 */) {}
  /* correctness-test moved to the start */
  if ((vvv&2) && (mpi_i+6)%mpi_n==0) {
#if 0
    tbase ib2=TBASE0, ib3=TBASE0, ib5=TBASE0; /* struct sbase b; */
    tphase ss=1; int nj;
    for (nj=1;nj<=sym.numall;nj<<=1); nj>>=1; if(nj)nj-=1; /* biggest 1111 below ns */
    l0=0; tm2=0; tm1=fstime();
    for (loops=128;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        ib2=ib4[l0 & (NUM_CFG_SAMPLES-1)];
        symcfg_bn(ib2,l0&nj,&ib5,&ss,1); /* 0=noSign, 1=getSign */
        ib3^=ib5;}
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) 
    printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfg= %7.2f %d 99i\n",
     "symcfg_bn", tm2, loops*1./tm2, tm2*1000000000./loops, 63&(int)ib3);
#else
    tbase ib3=TBASE0; int i; l0=0; tm2=0; tm1=fstime();
    // printf("DBG: n=%d\n",hbuf->n);
    for (loops=1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        /* sym_bn_m should keep in the L1 cache, hbuf->acj is data-stream */
        /* here bad speed test, no datastream, just same repeated */ 
        for (i=0;i<hbuf->n;i++) hbuf->acj[i]=ib4[i]; /* bad additional copy */
        minsymcfg_blk_bn(hbuf); /* ToDo: ueberschreibt input = ungut! */
        for (i=0;i<hbuf->n;i++) ib3^=hbuf->acj[i];
      } tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) printf("PRF: %-15s t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d 99i\n",
     "minsymcfg_bn", tm2, hbuf->n*loops*1./tm2,
      tm2*1000000000./loops/hbuf->n/(1.+sym.numall), 63&(int)ib3);
#endif
  }
  /* 7th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+7)%mpi_n==0) { /* vectorized */
    // bnv4 removed for code reduction
  }
  /* 8th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+8)%mpi_n==0) { /* vectorized */
    // bnv8 removed ToDo: space for VS/2 test
  }
  /* 9th different dummy test on other CPU (ToDo) */
  if ((vvv&2) && (mpi_i+9)%mpi_n==0) { /* vectorized */
#if 0
    tbase *ib2, ib3[VS] __attribute__ ((aligned (64))),
                ib5[VS] __attribute__ ((aligned (64))); /* struct sbase b; */
    tphase       ss[VS] __attribute__ ((aligned (64))); int nj, i;
    for (i=0;i<VS;i++) { ib3[i]=TBASE0; ss[i]=+1; }
    for (nj=1;nj<=sym.numall;nj<<=1); nj>>=1; if(nj)nj-=1; /* biggest 1111 below ns */
    l0=0; tm2=0; tm1=fstime();
    for (loops=128*1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0+=VS) { 
        ib2=&(ib4[l0 & (NUM_CFG_SAMPLES-1)]);
        symcfg_bnVS(ib2,(l0/VS)&nj,ib5,ss,1); /* 0=noSign, 1=getSign */
        for (i=0;i<VS;i++) ib3[i]^=ib5[i];
//      if (l0<VS) printf("DBG 8v %d %s %s %s %s\n",
//        (int)l0,ccfg(ib5[0]),ccfg(ib5[1]),ccfg(ib5[2]),ccfg(ib5[3]));
        }
      tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    ib5[0]=TBASE0; for (i=0;i<VS;i++) ib5[0]^=ib3[i];
    if (pt==0) 
    printf("PRF: %-10s%-5d t=%3.1fs loops/s= %7.1e t[ns]/cfg= %7.2f %d 99i\n",
     "symcfg_bnV", VS, tm2, loops*1./tm2, tm2*1.0e9/loops, 63&(int)ib5[0]);
#else
    tbase ib3=TBASE0; int i, j=0;  l0=0; tm2=0; tm1=fstime();
    // printf("DBG: n=%d\n",hbuf->n);
    /* hold as much as possible of sym_bn_m in L1-cache! ToDo: howto */
    for (loops=1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        /* sym_bn_m should keep in the L1 cache, hbuf->acj is data-stream */
        /* here bad speed test, no datastream, just same data repeated */ 
//        for (i=0;i<hbuf->n;i++) hbuf->acj[i]=ib4[i]; /* bad additional copy */
        for (i=0;i<hbuf->n;i++,j=(j+1)&(NUM_CFG_SAMPLES-1))
           hbuf->acj[i]=ib4[j]; /* bad additional copy, data-stream */
        /* ToDo: prefetch acj[i+VS] */
        if(nu==nd && sym.wud) minsymcfg_blk_bnVS(hbuf); /* ToDo: ueberschreibt input = ungut! */
          else                minsymcfg_blk_bnVS_no_ud(hbuf); /* faster!? */
        for (i=0;i<hbuf->n;i++) ib3^=hbuf->acj[i];
        /* ToDo: release acj[i] from cache */
      } tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0) printf("PRF: %13s%-2d t=%3.1fs loops/s= %7.1e t[ns]/cfs= %7.2f %d 99i\n",
     "minsymcfg_bnV", VS, tm2, hbuf->n*loops*1./tm2,
      tm2*1.0e9/loops/hbuf->n/(1.+sym.numall), 63&(int)ib3);
#endif
  }
#if 1
  /* 10th different dummy test on other CPU (ToDo: hamilton_nhv op_blk_h) */
  if ((vvv&2) && pt_n>1 && mpi_i==0 && pt==0)
                                          printf("DBG: ToDo fix pt_n>1!\n");
  if ((vvv&2) && pt_n<=1/* && (mpi_i+10)%mpi_n==0 parallel!*/) { /* vectorized */
    int i, j=0, maxnzx=0; ULong nzx=0; l0=0; tm2=0;
    if (!mpi_i) printf(  /* 2016-04-07 after bug fixed */
      "DBG: on error, please try verbose=1 too before reporting bug\n");
    MPI_Barrier(MPI_COMM_WORLD);  /* nhv is parallel, so we sync here 2015 */
    tm1=fstime();
    // printf("DBG: n=%d\n",hbuf->n);
    b_ofs[pt+1]=(pt+1)*NUM_AHEAD_LINES;   /* needet in _nhv, ToDo: argument */
    if (pt==0) b_ofs[pt]=pt*NUM_AHEAD_LINES;
    node_n1=NUM_AHEAD_LINES*pt_n; /* needed in _nhv extra check */
    /* ToDo2016-03: fill scfg2-tables */
    /* hold as much as possible of sym_bn_m in L1-cache! ToDo: howto */
    for (loops=1;loops<(1lu<<31);loops*=2) { /* run min. 1s */
      for (;l0<loops;l0++) { 
        for (i=0; i<NUM_AHEAD_LINES; i++,j=(j+1)&(NUM_CFG_SAMPLES-1)) {
           hbuf->cfg[i]=ib4[j]; /* usually inside _nhv from l1[] */
        }
        /* ToDo: set l1(0...ahead) + b_ofs + free after */
        /* nhv calls minsymcfg + scfg2idx using l1[] (uninitialized!) */
        // ToDo: remove ~ismincfg?
        hamilton_nhv(0,NUM_AHEAD_LINES,hbuf,0); // bad access to l1! out NZXMAX*AH
        for (i=0; i<NUM_AHEAD_LINES; i++) {
          if (maxnzx<hbuf->nzx[i]) maxnzx=hbuf->nzx[i];
          nzx += hbuf->nzx[i];
        }
        /* ToDo: release acj[i] from cache */
      } tm2=fstime()-tm1; if(tm2>0.25) break; /* 2015-10-08 skip loops*=2 */
    }    /* output speed */
    if (pt==0 && !mpi_i) /* 2015-11-17 add !mpi_i */
     printf("PRF: %13s   t=%3.1fs   nzx/s= %7.1e t[ns]/cfs= %7.2f"
     " nzx= %5.2f %d\n",
     "nhv_no_l1", tm2, nzx*1./tm2, tm2*1.0e9/nzx/(1.+sym.numall),
      nzx*1./(1.*loops*NUM_AHEAD_LINES), maxnzx );
  }
#endif
  /* ToDo: SH speed-test */
  if (pt==0 && !flag_once[pt]) flag_once[pt]=1; /* only once! */
  hbuf->n=0; /* reset hbuf */
  return;
}

/* ----- initialize symmetric base (cfg-tables l1) v1.53 ----- 
 * calculate orbit length or norm of each sym-config
 *  also do some speed tests and test new ideas
 * ToDo: should be done with creation of l1
 * v2.33 norm2[] table removed to simplify code, less MEM, little more CPU
 * 2014-05 speed test moved to earlier speed_test()
 */
void ini_norm2(void){
  ULong i1;
#if 0  // Aug2010 not used anymore?
  Tnorm2 sum2=0l,n2max=0; int n2hash, n2num=0;
#endif
  int i2;
  tbase ib1;
//  static struct sbase b1 /* b2=tmp */;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (vvv && !mpi_i) printf("\n ini_norm2");
 // b1.o=0;
// 2016-01-12 looks like unneeded code but err is 1 here
  ERR(-1); /* reset errno for ERR(600) */
  /* initialize hash table (for norm2 distribution) */
  for (i1=0l; i1<N2Blen; i1++) n2h[i1]=n2b[i1]=0;

  /* create scfg-range per block table */
  if (vvv&2 && !mpi_i) printf(" ini_mpi_l1[]");
  if (!mpi_l1) mpi_l1=(tbase *)malloc(sizeof(tbase)*mpi_n);
  if (!mpi_l1) ERR(9011);
  for (i2=0;i2<mpi_n;i2++) mpi_l1[i2] = 0; /* clear for debugging */
  /* mpi_l1 must be set also for len=0 */
  mpi_l1[mpi_i] =
       ((b_ofs[pt_n])? getl1(b_ofs[0]) /* first config of this node */
                     : ~(tbase)0 ); /* dummy cfg > max_cfg */
  // b_l1[mpi_i*pt_n+pt_n]= getl1(node_n1-1); /* ??? */
  /* MPI_Allgather(sbuf, 100, MPI_INT, rbuf, 100, MPI_INT, comm);
   *     AllGather: rbuf[  0.. 99]=proc0.sbuf[0..99]
   *                rbuf[100..199]=proc1.sbuf[0..99]
   *                ...           =procn.sbuf[0..99]
   */
  /* can we use a part of the receiverarray as sendarray? yes v2.27 */
#ifdef CONFIG_MPI
{
  tbase tmp_l1=mpi_l1[mpi_i]; /* sendbuf != recvbuf, v2.36 */
#if 0
  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (mpi_i<4 || mpi_i>=mpi_n-4)
      printf("\n DBGpreAG  node= %2d mpi_l1= %s",mpi_i,ccfg(mpi_l1[mpi_i]));
    else if (mpi_i==4) printf("\n ...");
  }
#endif
  MPI_Allgather(&tmp_l1, sizeof(tbase), MPI_BYTE,
                 mpi_l1, sizeof(tbase), MPI_BYTE, MPI_COMM_WORLD);
#if 0
  if (mpi_i==mpi_n-1 && vvv&4)
  for(i2=0;i2<mpi_n;i2++) 
    if (i2<4 || i2>=mpi_n-4)
      printf("\n DBGpostAG node= %2d mpi_l1= %s", i2, ccfg(mpi_l1[i2]));
    else if (i2==4) printf("\n ...");
#endif
}
#endif
  if ((vvv&2) && mpi_i==0) puts(""); /* NewLine */

  if (vvv&2) { /* speed test, see scfg2idx() for results */
    ULong j=~0, loops; double tm2=0, tm1=0;
    tbase ib2=0; /* JS2012-01 do a simple check that i1<next(i1) */
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    tm1=fstime();
    /* if (!mpi_i) printf("PRF: scfg2idx_speedtest.start\n"); */
    for (loops=0;loops<400000;loops++) { /* run until 1s gone */
      for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
       ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */
       if (ib2>=ib1 && i1!=0) break; /* also plausible check correctness */
       j=scfg2idx(ib1); /* ToDo 220ns vs. 5*(rndmem=40ns) dSH=5ns*64=320ns */
       /* 16+16 lc 220ns/64sym= 3.4ns vs.  diff=SH=39ns/cfg - nhv=34ns/cfg */
       if (j!=i1) break;
       ib2=ib1;
      }
      if (i1<node_n1) { // JS2012-01-20
        int j2;
        printf("ERR task%d (j=%lu)!=(i1=%lu) or ib2>=ib1 ll2=%lu\n",
          mpi_i, j, i1, ll2(ib1));
        printf("ERR ib2 %s %lu\n",ccfg(ib2),i1-1);
        printf("ERR ib1 %s %lu\n",ccfg(ib1),i1);
        for (j2=-3;j2<3;j2++) if(-j2<=(signed long)j && j+j2<node_n1)
        printf("ERR j%+d %s %lu\n",j2,ccfg(getl1(j+j2)),j+j2);
        sleep(10); // give other tasks a chance to show further errors
        F0("inkonsistent l1")
      }
      tm2=fstime()-tm1; if (tm2>0.25) break; /* 2015-10-08 correct loops! */
    }
    if (!mpi_i)
      printf("PRF: scfg2idx t=%3.1fs loops= %lu * %.2e t[ns]/n1= %.1f\n",
             tm2, (loops+1), 1.*node_n1, tm2*1.0e9/((loops+1)*1.*node_n1));
  }
  
  if (vvv&4) /* not needed at the moment */
  { /* create Isingtable for blocking criteria, Idea from Oct07 */
    /* ToDo: also output a num_states_per_IsingErg I1-I2-matrix */
    int j1,                    /* loop-vars */ 
        min0, max0,             /* min/max numPairs */
        ising,                  /* ising erg, may be exotic hash of all Js */ 
        *buf=NULL;              /* buffer for counted pairs (IsingErg) */
#define MaxIsing (B_NUM)    /* should be prime but 4*(pt_n)..4*NN */
    long numIsing[MaxIsing];   /* IsingErg[j1,j2,..] for load balance */
    long minI,maxI,t1=itime();

    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i)
      printf("\n ini_norm2.IsingBlocks:");
    min0=MaxIsing; max0=0;
    for(j1=0;j1<MaxIsing;j1++) numIsing[j1]=0; /* init hash array */
    /* ToDo: hash should leave near J1-Ising-ergs at near hashs */
    if (sym.num_dcorr*CM(2,4,5)>22)
    for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
      ib1=getl1((ULong)i1);
      buf=get_dcorr(ib1,1,buf); /* get dcorr-table, count ud-pairs */
      /* which bonds we take? worst case is spin glas -> add different bonds */
      /* Idea: make a monte carlo test to find a good function */
      /* IsingErg can be arteficcial build from dcorr, ex: j1+j4/2+... */
      /* ToDo: do hashing within get_dcorr? => hashed j1-jmax-Ising erg */
//    ising=((731*buf[ 5]+23*buf[15]+171*buf[25]+3*buf[35]+17*buf[45])/2)%pt_n;
      ising=((9551*buf[ 2]+23*buf[ 6]
             +1327*buf[10]+ 3*buf[14]
             +  17*buf[18]+ 7*buf[22])/2)%pt_n;
      if (min0>ising) min0=ising;
      if (max0<ising) max0=ising;
      if (ising>=MaxIsing) ising=MaxIsing-1;
      /* hash indicees to avoid overflow */
      numIsing[ising]++;
    }
    if (!mpi_i)
      printf("\n Ising range for node= %2d: %d-%d", mpi_i, min0, max0);
    /* output Ising distribution, ToDo: only min, max, variance? */
    minI=n1; maxI=0;
    for(j1=min0;j1<MaxIsing && j1<=max0;j1++){
      if (!mpi_i)
        printf("\n node= %2d Ising hash=%2d: %6ld", mpi_i, j1, numIsing[j1]);
      if (numIsing[j1]<minI) minI= numIsing[j1];
      if (numIsing[j1]>maxI) maxI= numIsing[j1];
    }
    if (!mpi_i)
      printf("\n node= %2d Ising block size: %lu-%lu-%lu (%.0f%%-%.0f%%)"
             "\nPRF ising.t=%s",
            mpi_i, minI, node_n1/pt_n, maxI,
            minI*100./(node_n1/pt_n),
            maxI*100./(node_n1/pt_n), sclock(itime()-t1));
    if (buf) free(buf); /* malloc by get_dcorr() */
  }
  /* ToDo: sum hash tables? + pthread */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

#if 1
  if (vvv && !mpi_i)
    printf(" create n2hashtables n2b,n2h - skipped (ToDo)\n");
#else /* slow! */
  if (vvv && !mpi_i)
    printf(" create n2hashtables n2b,n2h - (ToDo parallel)\n");
  /* ToDo: remove n2, no significant speedup and MPI-Probs */
  /* if (norm2) ToDo: ini sqrt[n2a][n2b]-table when n2a!=n2b often? */
  for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
   ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */

   // isminsymcfg_lNbrk(ib1,&b1); /* called only to get norm2 2016-01 = o n2 */
   minsymcfg_lNbrk(ib1,&b1); /* called only to get norm2 2016-01 = o n2 */
   sum2+=b1.norm2; /* sum2 == numconf ! */
   if(b1.norm2 > n2max) n2max=b1.norm2;	/* store maximum orbit length */

   n2hash=(int)(((ULong)b1.norm2)%N2Blen);
   if (!n2b[n2hash]) { n2b[n2hash]++; n2h[n2hash]=b1.norm2; n2num++; }
              else if (n2h[n2hash]==b1.norm2) n2b[n2hash]++;
              else { n2b[0]++; /* collision */}
  }
  if (err_fatal==600) ERR(-1); /* reset error 600 */
  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i) printf("\n orbit lengths (norm2): ");
    for(i1=1; i1<N2Blen; i1++) if (n2b[i1]) {
      if (mpi_i<4 || mpi_i>=mpi_n-4)
      printf("\n node %2d norm2= %15.10g %8d times",
              mpi_i,(double)n2h[i1],n2b[i1]);
      else if (mpi_i==4) printf("\n ...");
    }
    if (n2b[0]) printf("\n hash_collisions= %8d (should be zero)",n2b[0]);
  }

  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (mpi_i<4 || mpi_i>=mpi_n-4)
    printf("\n node %2d n2num= %d  n2max= %.10g sum_norm2= %.15g t+= %s",
            mpi_i, n2num, (double)n2max, (double)sum2, sclock(~0l));
    else if (mpi_i==4) printf("\n ...");
  }

  /* does not work for Tnorm2 == double */
  if ( ((ULong)n2max)>>(sizeof(ATnorm2)/2*8) == 0 && !mpi_i)
    printf("\n# spare memory: Change ATnorm2 to short or char.\n");

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
#endif

#ifndef CONFIG_MPI
  /* MaxSBase set to 1 to spare Memory but allsymcfg will fail v2.33 */
  if ((vvv&128) && !mpi_i) {
    static struct fullsbase b1; b1.o=0; /* added 2014-08-05 MaxSBase*size */
    struct sbase b2;  // b2.o=0; /* added 2015-03 small sbase */
    printf("\nDBG: ini_norm2 vvv&128 (node=0 only):");
    for(i1=0l;i1<node_n1 && i1<x_max;i1++){
      /* show base for debugging purpose, ToDo: extra norm2.lm-output? */
      // ismin gives no norm2 since 2016-01 (overflows etc.)
      // isminsymcfg_lNbrk(getl1((ULong)i1),&b2); /* different results for S>=1 ??? */
      minsymcfg_lNbrk(getl1((ULong)i1),&b2); /* different results for S>=1 ??? */
      if (b2.norm2 < 9999999) printf("\n %7d",(int)b2.norm2); /* NAN? */
      else               printf("\n %7.1e",(double)b2.norm2); /* s=5/2 */
   /* minsymcfg_dflt(getl1((ULong)i1),&b1); */
      allsymcfg(getl1((ULong)i1),&b1);
      printf(  " %6d |%.15g>=",(int)b1.norm2,(double)i1);
      for(i2=0l;i2<b1.num;i2++) 
        printf("\n %32s*%s",outphase(b1.is[i2]),ccfg(b1.ib[i2]));
    }
  }
#endif
  if (vvv && !mpi_i) printf(" ini_norm2.end\n");
  return;
}


/* last modifications: 
   1998-148 type tbase j,m substituted by ulong etc.
   Dez98: general tbase-type independent scfg2idx-fkt. + better errorhandling
   Jun14 triggered by ri@20*16comp2 star54 NFS? "belongs to wrong node"
 */
 /* nu=dbg_numbits(cfg,1);
  * nd=dbg_numbits(cfg,2);  */
int dbg_numbits(tbase cfg, int bit){ /* 2015-03 found defect RAM bits, check l1[] */
  int i1, i2;
  /* if getl1 was called with index_out_of_range, we have wrong bits */
  for (i2=i1=0;i1<nn;i1++) { if (geti(cfg,i1)&bit) i2++; }
  return i2;
}

/* convert base to index, this is good break point for debugging */
ULong scfg2idx_err(tbase cfg){ /* called on [blk_]scfg2idx() error */
  ULong j, m; tbase cfg2;
  static int err=0;
  static ULong m2=1l;    /* not thread safe */
  static struct sbase b; /* static makes it 100x faster, only used if error */
  static struct fullsbase fb; /* static makes it 100x faster, only used if error */
  int node, iu, id; 
  err++;
  if (err>128 && 0!=(err&(err-1))) return ~0; /* limit error msgs lg2 */
  printf("\nscfg2idx_err %s not found (mpi_i=%d, numerr=%d)",
          ccfg(cfg), mpi_i, err);
  /* if getl1 was called with index_out_of_range, we have wrong bits */
  iu=dbg_numbits(cfg,1);
  id=dbg_numbits(cfg,2);
  if ((cfg>>(CM(1,2,2)*nn-1))>>1) { /* bit problems */
    printf("\nscfg2idx_err %d: scfg has bad bits above nn=%d", mpi_i, nn);
    ERR(9999);
  }
  if (iu!=nu || id!=nd) {
    printf("\nscfg2idx_err %d: scfg has nu,nd= %d %d which is wrong",
           mpi_i, iu, id);
    ERR(9999);
  }
  node=scfg2node(cfg);
  if (node!=mpi_i) {
    printf("\nscfg2idx_err I am task=%d, but scfg belongs to other task=%d",
              mpi_i, node);
    printf("\nscfg2idx_err mpi_l1[%d]=%s", 0, ccfg(mpi_l1[0]));
    if (node>=1 && node<mpi_n) /* added 2016-03 */
      printf("\nscfg2idx_err mpi_l1[%d]=%s", node-1, ccfg(mpi_l1[node-1]));
    if (node>=0 && node<mpi_n)
      printf("\nscfg2idx_err mpi_l1[%d]=%s", node  , ccfg(mpi_l1[node  ]));
    if (node>=1 && node<mpi_n-1)
      printf("\nscfg2idx_err mpi_l1[%d]=%s", node+1, ccfg(mpi_l1[node+1]));
    printf("\nscfg2idx_err mpi_l1[%d]=%s", mpi_n-1, ccfg(mpi_l1[mpi_n-1]));
    ERR(9999);
  }
  /* recompute the index + checks 2015-03 */
  while ((m2<<1)<node_n1) m2<<=1;	/* Feb00+Nov07 get biggest bit */
  for(j=0l,m=m2;m;m>>=1)if((j|m)<node_n1){
    cfg2=getl1(j|m);
    iu=dbg_numbits(cfg2,1);
    id=dbg_numbits(cfg2,2);
    /* ri2015-10-05 had correct cfg_sm but its not found in table, full dsk */
    minsymcfg_dflt(cfg2,&b); /* ToDo: what is dflt? check it in speed_test! */
    if (cfg2!=b.scfg)  /* bad l1[], cat failed? ToDo: recompute? */
      printf("\n l1[%6lu]= %s but l1.minsym= %s phase= %s %e", j, ccfg(cfg2),
        ccfg(b.scfg), outphase(b.sign), D(b.sign));
    if (iu != nu  ||  id != nd) {
      ULong j2; tbase cfg3;
      /* 2015-03 SC.mkautz.n8 silent single bit flips in l1[] ^_-_^ */
      printf("\nscfg2idx_err %d: iu=%d != nu at l1[%6lu] (bit error?)",
        mpi_i,iu,j|m);
      j^=m;
      if(j-2<node_n1) printf("\n l1[%6lu]= %s",j-2,ccfg(getl1(j-2)));
      if(j-1<node_n1) printf("\n l1[%6lu]= %s",j-1,ccfg(getl1(j-1)));
      if(j  <node_n1) printf("\n l1[%6lu]= %s (bad nu+nd)",j,ccfg(cfg2));
      if(j  <node_n1) printf("\n l1[%6lu]= %s",j  ,ccfg(getl1(j  )));
      if(j+1<node_n1) printf("\n l1[%6lu]= %s",j+1,ccfg(getl1(j+1)));
      if(j+2<node_n1) printf("\n l1[%6lu]= %s",j+2,ccfg(getl1(j+2)));
      printf("\n full scan l1 ...");
      /* flush cpu cache by full scanning */
      cfg3=getl1(0)-1;  /* predecessor, avoid error for j2=0 */
      for(j2=0l;j2<node_n1;j2++){
        cfg2=getl1(j2);
        iu=dbg_numbits(cfg2,1);
        id=dbg_numbits(cfg2,2);
        if (iu != nu  ||  id != nd || cfg2<=cfg3) {
         if(j  <node_n1) printf("\n l1[%6lu]= %s (bad cfg)",j2 ,ccfg(cfg2));
         sleep(1); /* time to ^c on debugging */
        }
        cfg3=cfg2;
      }
      
      j^=m;
      /* ToDo: try to correct table on the fly via next_minsymcfg j-1 ? */
    }
    if( (getl1(j|m)<=cfg) )j|=m;
  }

  if (!(getl1(j)==cfg)) /* ok, its still not ok */
  {			/* not found */
   //b.o=0; 
   fb.o=0;/* gb=getbase may fail on size optimized code without panic */
   allsymcfg(    cfg,&fb);printf("\n     cfg_gb= %s phase= %s %e",
                            ccfg(fb.scfg),outphase(fb.sign),D(fb.sign));
   /* gb=getbase(full) sm=smallest(dflt) ToDo: minsymcfg_loop too? */
   minsymcfg_dflt(cfg,&b);printf("\n     cfg_sm= %s phase= %s %e",
                            ccfg(b.scfg),outphase(b.sign),D(b.sign));
   if (Norm(b.sign)<1e-3) printf("\n err Norm(sign)<1e-3 %s",ccfg(b.scfg));
   if (Norm(b.sign-fb.sign)>1e-9)
      printf("\n err signs of gb=allsymcfg and sm=minsymcfg differ (bug!)");
   if (b.sign==0) printf("\n err sign==0 %s numerr=%d",ccfg(b.scfg),err);
   else { // ToDo: temporary test (remove it)
     if (cfg!=b.scfg) { cfg=b.scfg;
       minsymcfg_dflt(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
     if (cfg!=b.scfg) { cfg=b.scfg;
       minsymcfg_dflt(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
     if (cfg!=b.scfg) { cfg=b.scfg;
       minsymcfg_dflt(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
   }
   cfg=b.scfg;
   for(j=0l,m=m2;m;m>>=1){ if((j|m)<node_n1)if((getl1(j|m)<=cfg))j|=m; }
   if(!(getl1(j)==cfg)){       /* not found again !!! */
     /* if (err<3) */ {
       if(j> (ULong)0) printf("\n l1[%6lu]= %s",0l ,ccfg(getl1( 0l)));
       if(j> (ULong)1) printf("\n l1[%6lu]= %s",1l ,ccfg(getl1( 1l)));
       if(j-2<node_n1) printf("\n l1[%6lu]= %s",j-2,ccfg(getl1(j-2)));
       if(j-1<node_n1) printf("\n l1[%6lu]= %s",j-1,ccfg(getl1(j-1)));
       if(j  <node_n1) printf("\n l1[%6lu]= %s",j  ,ccfg(getl1(j  )));
       if(j+1<node_n1) printf("\n l1[%6lu]= %s",j+1,ccfg(getl1(j+1)));
       j=node_n1-1;
       if(node_n1 > 0) printf("\n l1[%6lu]= %s\n",j,ccfg(getl1(j  )));
       ERR(400); /* not found, try again */
     }
     for(j=0;j<node_n1;j++) if(getl1(j)==cfg) break; /* l1 not sorted? */
     if(j==node_n1)if(err<10){ ERR(410); err_fatal++; }
   }
  }
  if (j>=node_n1) ERR(499); /* 2015-09-08 avoid illegal access l1[] */
//  return 0 is a bad idea, it hides deeper errors
//  if (j>=node_n1) j=0;  /* 2015-12-31 tJ_lc_s1 0+2, set j=0 hope sign=0 */
#ifndef CONFIG_ANSATZ
  if (j>=node_n1 || getl1(j)!=cfg) { /* for j>=node_n1 no getl1()-call */
    if (b.sign==0 && err<5) E0("scfg2idx failed, idx=node_n1 for debugging")
    else  ERR(9999); /* 2012-11-05 this is fatal! report it! */
  }
#endif
  if (j<node_n1 && getl1(j)==cfg && err<5) /* no getl1(j>=node_n1)-call */
    printf("\n node %2d: now scfg2idx() is ok and returns %lu = 0x%08lx",
           mpi_i, (unsigned long)j, (unsigned long)j);
  return (j);
}

/*
  scfg2idx: sym-config to index within node (0..node_n1-1) (old: b2i)
  the routine is short as possible to fit into the cache
   input is spin-configuration (representator)
   output is index to tables
   ToDo:
    - convert more than one index within one call  for better cache performance
    - take 8 highest 8bit as index best_speedup=lg2(n1)/(lg2(n1)-8)=1.36
      no effect for high Sz, because all configs start with 0-bits?
    - l1_rough_idx[0..1023] Oct07 v2.27
      scfg2idx-speedtest getl1+scfg2idx+if nud=33,7  P4-1.3GHz L2=512kB
       search full table           92loops/9s (bits), 94loops/9s (interval)
       search through rough table 147loops/9s
       no search                 4900loops/9s
       lg2(n1=116769)=17bits lg2(Nrough=1024)=10bits speedup=50%
       no speedup for lg2(n1=1e9)=30bits (speedup=30%)
    Idea: reduce rough table for nodes and threads (no mem collisions!)
          as a compromise
    Idea: sort-as-searchtree: 1..10 to { 5,    2, 7,     1,3, 6,8,  ...}
                                         root  2elements 4elements 8elements
           for better caching? not really -> need hardware-search (or hash?)
    Idea: hash table? or IsingErg
      return -i if cfg outside block? -> no, use extra chck+fkt
    ToDo: search thread_l1_block only?, rename to scfg2i
                                
 */
ULong scfg2idx(tbase cfg){ /* (b=base_config=scfg)-to-index < node_n1, rename to scfg2i! */
  ULong j,ja,jb;
  ja=0; j=jb=node_n1-1;
  /* ToDo: block.h b_idx2blk() looks more nice, without +-1 */
  for (;ja<jb;) { j=(ja+jb+1)/2; if (getl1(j)<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(node_n1)*(2cmp+2add+1shft+1mem+2jmp) */
  if (getl1(j)==cfg) return j;  /* getl1 returns the node_index=0..node_n1-1 */
  return scfg2idx_err(cfg); /* slow error-routine outside the cpu-i-cache */
}

void blk_scfg2idx(tbase *cfg, ULong *idx, int n){
 /* (b=base_config=scfg)-to-index < node_n1, rename to scfg2i! */
  ULong j,ja,jb; int ii; 
  for (ii=0;ii<n;ii++) {
    ja=0; j=jb=node_n1-1;
    /* ToDo: block.h b_idx2blk() looks more nice, without +-1 */
    for (;ja<jb;) { j=(ja+jb+1)/2;
      if (getl1(j)<=cfg[ii]) ja=j; else j=jb=j-1; }
    /* price: log2(node_n1)*(2cmp+2add+1shft+1mem+2jmp) */
    if (getl1(j)==cfg[ii]) { idx[ii]=j; }  
    /* getl1 returns the node_index=0..node_n1-1 */
    else idx[ii]= scfg2idx_err(cfg[ii]);
    /* slow error-routine outside the cpu-i-cache */
  }
  return;
}

/* scfg2node: get global mpi node number from sym-config
 */
int scfg2node(tbase cfg){ /* get global mpi node number from sym-config */
  int j,ja,jb;
  ja=0; j=jb=mpi_n-1;
  for (;ja<jb;) { j=(ja+jb+1)/2; if (mpi_l1[j]<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(mpi_n)*(2cmp+2add+1shft+1mem+2jmp) */
  return j;  /* getl1 returns the node_index=0..node_n1-1 */
}
